/* -*- c++ -*- */
/* 
 * Copyright 2018 Ingo D. Rullhusen <d01c@freeshell.de>.
 * 
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gnuradio/io_signature.h>
#include "BBdescrambler_impl.h"


namespace gr {
  namespace DVBparts {

    BBdescrambler::sptr
    BBdescrambler::make()
    {
      return gnuradio::get_initial_sptr
        (new BBdescrambler_impl());
    }

    /*
     * The private constructor
     */
    BBdescrambler_impl::BBdescrambler_impl()
      : gr::sync_block("BBdescrambler",
              gr::io_signature::make(1,1, sizeof(unsigned char)),
              gr::io_signature::make(1,1, sizeof(unsigned char)))
    {
       NBB = 0;

       bzero(&l1plp,sizeof(l1plp));
       setup_L1Pplp_tag( PLPTagValue, l1plp );

       descrambler_offset = 0;
       init_descrambler();

    }

    /*
     * Our virtual destructor.
     */
    BBdescrambler_impl::~BBdescrambler_impl()
    {
    }


    void
    BBdescrambler_impl::init_descrambler(void)
    {
      int sr = 0x4A80;
      for (int i = 0; i <BB_SIZE_MAX; i++) {
        int b = ((sr) ^ (sr >> 1)) & 1;
        descrambler[i] = b;
        sr >>= 1;
        if(b) {
          sr |= 0x4000;
        }
      }
    }


    int
    BBdescrambler_impl::work(int noutput_items,
        gr_vector_const_void_star &input_items,
        gr_vector_void_star &output_items)
    {
       const pmt::pmt_t PLPfecKey( pmt::intern("PLPfec") );
       const unsigned char *in = (const unsigned char *) input_items[0];
       unsigned char *out = (unsigned char *) output_items[0];

       std::vector<tag_t> tags;
       get_tags_in_window( tags, 0, 0, noutput_items, PLPfecKey );

       for ( int n=0; n<noutput_items; ++n )
       {
          uint64_t noff = n + nitems_read(0);
          tag_t *tagPLP = NULL;
          for ( std::vector<tag_t>::iterator t=tags.begin(); t!=tags.end(); ++t )
          { if ( noff == t->offset ) { tagPLP = &(*t); break; } }

          if ( tagPLP )
          {
             pmt::pmt_t PLPTagValue_t = pmt::tuple_ref(tagPLP->value,1);
             if ( ! equal( PLPTagValue, PLPTagValue_t ) )
             {
                PLPTagValue = PLPTagValue_t;
                setup_L1Pplp( l1plp, PLPTagValue_t );

                if ( l1plp.plp_fec_type == 1 ) // 64K
                {
                   switch ( l1plp.plp_cod )
                   {
                      case 0: // 1/2
                      { NBB = 32208; break; }
                      case 1: // 3/5
                      { NBB = 38688; break; }
                      case 2: // 2/3
                      { NBB = 43040; break; }
                      case 3: // 3/4
                      { NBB = 48408; break; }
                      case 4: // 4/5
                      { NBB = 51648; break; }
                      case 5: // 5/6
                      { NBB = 53840; break; }
                   }
                }
                else // 16K
                {
                   switch ( l1plp.plp_cod )
                   {
                      case 0: // 1/2
                      { NBB = 7032; break; }
                      case 1: // 3/5
                      { NBB = 9552; break; }
                      case 2: // 2/3
                      { NBB = 10632; break; }
                      case 3: // 3/4
                      { NBB = 11712; break; }
                      case 4: // 4/5
                      { NBB = 12432; break; }
                      case 5: // 5/6
                      { NBB = 13152; break; }
                   }
                }

             }

             descrambler_offset = 0;
          }

          out[n] = in[n] ^ descrambler[ descrambler_offset ];
          if ( descrambler_offset < NBB ) { ++descrambler_offset;   }
          else                            { descrambler_offset = 0; }
       }

      // Tell runtime system how many output items we produced.
       return noutput_items;
    }

  } /* namespace DVBparts */
} /* namespace gr */

