/* -*- c++ -*- */
/* 
 * Copyright 2019 Ingo D. Rullhusen <d01c@freeshell.de>.
 * 
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <math.h>

#include <gnuradio/io_signature.h>
#include <gnuradio/dtv/dvb_config.h>
#include <gnuradio/dtv/dvbt2_config.h>

#include "crc32.h"
#include "L1-Post_impl.h"


using namespace std;

using namespace gr::dtv;


namespace gr {
  namespace DVBparts {

    L1_Post::sptr
    L1_Post::make()
    {
      return gnuradio::get_initial_sptr
        (new L1_Post_impl());
    }

    /*
     * The private constructor
     */
   L1_Post_impl::L1_Post_impl()
     : gr::block("L1-Post",
       gr::io_signature::make(1,1, sizeof(gr_complex)),
       gr::io_signature::make(1,1, sizeof(gr_complex)))
   {
      d_msg_port = pmt::mp("msg");
      message_port_register_out(d_msg_port);

      bzero(&l1pre,sizeof(l1pre));

      BCHctl = init_bch( /*m*/ 14, /*t*/ 12, /*prim_poly*/ 0x0000402b );

      demod_bits_cnt = 0;
      demod_bits = NULL;

      Kpost_ex_pad = 0;
      L1_post_buffer = NULL;

      init_l1_randomizer();

      set_tag_propagation_policy(TPP_DONT);
      set_output_multiple(32*1024);
      set_fixed_rate(false);
   }

    /*
     * Our virtual destructor.
     */
   L1_Post_impl::~L1_Post_impl()
   {
      delete [] demod_bits;
      delete [] L1_post_buffer;

      free_bch(BCHctl);

   }

    void
    L1_Post_impl::forecast (int noutput_items, gr_vector_int &ninput_items_required)
    {
      /* <+forecast+> e.g. ninput_items_required[0] = noutput_items */
      ninput_items_required[0] = noutput_items;
    }

    void
    L1_Post_impl::init_l1_randomizer(void)
    {
      int sr = 0x4A80;
      for (int i = 0; i < KBCH_1_2; i++) {
        int b = ((sr) ^ (sr >> 1)) & 1;
        l1_randomize[i] = b;
        sr >>= 1;
        if(b) {
          sr |= 0x4000;
        }
      }
    }

    void
    L1_Post_impl::update_parameter(void)
    {
      const int &preamble = l1pre.s1;
      const int &pilotpattern = l1pre.pilot_pattern;
      const int &carriermode = l1pre.bwt_ext;
      const int &paprmode = l1pre.papr;
      const int &guardinterval = l1pre.guard_interval;
      const int &numdatasyms = l1pre.num_data_symbols;

      bool miso;

      if ((preamble == PREAMBLE_T2_SISO) || (preamble == PREAMBLE_T2_LITE_SISO)) {
        miso = false;
        switch (dvbt2_fftsize) {
          case FFTSIZE_1K:
            N_P2 = 16;
            C_P2 = 558;
            break;
          case FFTSIZE_2K:
            N_P2 = 8;
            C_P2 = 1118;
            break;
          case FFTSIZE_4K:
            N_P2 = 4;
            C_P2 = 2236;
            break;
          case FFTSIZE_8K:
          case FFTSIZE_8K_T2GI:
            N_P2 = 2;
            C_P2 = 4472;
            break;
          case FFTSIZE_16K:
          case FFTSIZE_16K_T2GI:
            N_P2 = 1;
            C_P2 = 8944;
            break;
          case FFTSIZE_32K:
          case FFTSIZE_32K_T2GI:
            N_P2 = 1;
            C_P2 = 22432;
            break;
        }
      }
      else {
        miso = true;
        switch (dvbt2_fftsize) {
          case FFTSIZE_1K:
            N_P2 = 16;
            C_P2 = 546;
            break;
          case FFTSIZE_2K:
            N_P2 = 8;
            C_P2 = 1098;
            break;
          case FFTSIZE_4K:
            N_P2 = 4;
            C_P2 = 2198;
            break;
          case FFTSIZE_8K:
          case FFTSIZE_8K_T2GI:
            N_P2 = 2;
            C_P2 = 4398;
            break;
          case FFTSIZE_16K:
          case FFTSIZE_16K_T2GI:
            N_P2 = 1;
            C_P2 = 8814;
            break;
          case FFTSIZE_32K:
          case FFTSIZE_32K_T2GI:
            N_P2 = 1;
            C_P2 = 17612;
            break;
        }
      }
      switch (dvbt2_fftsize) {
        case FFTSIZE_1K:
          C_PS = 853;
          K_EXT = 0;
          K_OFFSET = 0;
          break;
        case FFTSIZE_2K:
          C_PS = 1705;
          K_EXT = 0;
          K_OFFSET = 0;
          break;
        case FFTSIZE_4K:
          C_PS = 3409;
          K_EXT = 0;
          K_OFFSET = 0;
          break;
        case FFTSIZE_8K:
        case FFTSIZE_8K_T2GI:
          if (carriermode == CARRIERS_NORMAL) {
            C_PS = 6817;
            K_EXT = 0;
            K_OFFSET = 48;
          }
          else {
            C_PS = 6913;
            K_EXT = 48;
            K_OFFSET = 0;
          }
          break;
        case FFTSIZE_16K:
        case FFTSIZE_16K_T2GI:
          if (carriermode == CARRIERS_NORMAL) {
            C_PS = 13633;
            K_EXT = 0;
            K_OFFSET = 144;
          }
          else {
            C_PS = 13921;
            K_EXT = 144;
            K_OFFSET = 0;
          }
          break;
        case FFTSIZE_32K:
        case FFTSIZE_32K_T2GI:
          if (carriermode == CARRIERS_NORMAL) {
            C_PS = 27265;
            K_EXT = 0;
            K_OFFSET = 288;
          }
          else {
            C_PS = 27841;
            K_EXT = 288;
            K_OFFSET = 0;
          }
          break;
      }
      Kmin = 0; Kmax = C_PS - 1;
      switch (dvbt2_fftsize) {
        case FFTSIZE_1K:
          switch (pilotpattern) {
            case PILOT_PP1:
              C_DATA = 764;
              N_FC = 568;
              C_FC = 402;
              break;
            case PILOT_PP2:
              C_DATA = 768;
              N_FC = 710;
              C_FC = 654;
              break;
            case PILOT_PP3:
              C_DATA = 798;
              N_FC = 710;
              C_FC = 490;
              break;
            case PILOT_PP4:
              C_DATA = 804;
              N_FC = 780;
              C_FC = 707;
              break;
            case PILOT_PP5:
              C_DATA = 818;
              N_FC = 780;
              C_FC = 544;
              break;
            case PILOT_PP6:
              C_DATA = 0;
              N_FC = 0;
              C_FC = 0;
              break;
            case PILOT_PP7:
              C_DATA = 0;
              N_FC = 0;
              C_FC = 0;
              break;
            case PILOT_PP8:
              C_DATA = 0;
              N_FC = 0;
              C_FC = 0;
              break;
          }
          if (paprmode == PAPR_TR || paprmode == PAPR_BOTH) {
            if (C_DATA != 0) {
              C_DATA -= 10;
            }
            if (N_FC != 0) {
              N_FC -= 10;
            }
            if (C_FC != 0) {
              C_FC -= 10;
            }
          }
          break;
        case FFTSIZE_2K:
          switch (pilotpattern) {
            case PILOT_PP1:
              C_DATA = 1522;
              N_FC = 1136;
              C_FC = 804;
              break;
            case PILOT_PP2:
              C_DATA = 1532;
              N_FC = 1420;
              C_FC = 1309;
              break;
            case PILOT_PP3:
              C_DATA = 1596;
              N_FC = 1420;
              C_FC = 980;
              break;
            case PILOT_PP4:
              C_DATA = 1602;
              N_FC = 1562;
              C_FC = 1415;
              break;
            case PILOT_PP5:
              C_DATA = 1632;
              N_FC = 1562;
              C_FC = 1088;
              break;
            case PILOT_PP6:
              C_DATA = 0;
              N_FC = 0;
              C_FC = 0;
              break;
            case PILOT_PP7:
              C_DATA = 1646;
              N_FC = 1632;
              C_FC = 1396;
              break;
            case PILOT_PP8:
              C_DATA = 0;
              N_FC = 0;
              C_FC = 0;
              break;
          }
          if (paprmode == PAPR_TR || paprmode == PAPR_BOTH) {
            if (C_DATA != 0) {
              C_DATA -= 18;
            }
            if (N_FC != 0) {
              N_FC -= 18;
            }
            if (C_FC != 0) {
              C_FC -= 18;
            }
          }
          break;
        case FFTSIZE_4K:
          switch (pilotpattern) {
            case PILOT_PP1:
              C_DATA = 3084;
              N_FC = 2272;
              C_FC = 1609;
              break;
            case PILOT_PP2:
              C_DATA = 3092;
              N_FC = 2840;
              C_FC = 2619;
              break;
            case PILOT_PP3:
              C_DATA = 3228;
              N_FC = 2840;
              C_FC = 1961;
              break;
            case PILOT_PP4:
              C_DATA = 3234;
              N_FC = 3124;
              C_FC = 2831;
              break;
            case PILOT_PP5:
              C_DATA = 3298;
              N_FC = 3124;
              C_FC = 2177;
              break;
            case PILOT_PP6:
              C_DATA = 0;
              N_FC = 0;
              C_FC = 0;
              break;
            case PILOT_PP7:
              C_DATA = 3328;
              N_FC = 3266;
              C_FC = 2792;
              break;
            case PILOT_PP8:
              C_DATA = 0;
              N_FC = 0;
              C_FC = 0;
              break;
          }
          if (paprmode == PAPR_TR || paprmode == PAPR_BOTH) {
            if (C_DATA != 0) {
              C_DATA -= 36;
            }
            if (N_FC != 0) {
              N_FC -= 36;
            }
            if (C_FC != 0) {
              C_FC -= 36;
            }
          }
          break;
        case FFTSIZE_8K:
        case FFTSIZE_8K_T2GI:
          if (carriermode == CARRIERS_NORMAL) {
            switch (pilotpattern) {
              case PILOT_PP1:
                C_DATA = 6208;
                N_FC = 4544;
                C_FC = 3218;
                break;
              case PILOT_PP2:
                C_DATA = 6214;
                N_FC = 5680;
                C_FC = 5238;
                break;
              case PILOT_PP3:
                C_DATA = 6494;
                N_FC = 5680;
                C_FC = 3922;
                break;
              case PILOT_PP4:
                C_DATA = 6498;
                N_FC = 6248;
                C_FC = 5662;
                break;
              case PILOT_PP5:
                C_DATA = 6634;
                N_FC = 6248;
                C_FC = 4354;
                break;
              case PILOT_PP6:
                C_DATA = 0;
                N_FC = 0;
                C_FC = 0;
                break;
              case PILOT_PP7:
                C_DATA = 6698;
                N_FC = 6532;
                C_FC = 5585;
                break;
              case PILOT_PP8:
                C_DATA = 6698;
                N_FC = 0;
                C_FC = 0;
                break;
            }
          }
          else {
            switch (pilotpattern) {
              case PILOT_PP1:
                C_DATA = 6296;
                N_FC = 4608;
                C_FC = 3264;
                break;
              case PILOT_PP2:
                C_DATA = 6298;
                N_FC = 5760;
                C_FC = 5312;
                break;
              case PILOT_PP3:
                C_DATA = 6584;
                N_FC = 5760;
                C_FC = 3978;
                break;
              case PILOT_PP4:
                C_DATA = 6588;
                N_FC = 6336;
                C_FC = 5742;
                break;
              case PILOT_PP5:
                C_DATA = 6728;
                N_FC = 6336;
                C_FC = 4416;
                break;
              case PILOT_PP6:
                C_DATA = 0;
                N_FC = 0;
                C_FC = 0;
                break;
              case PILOT_PP7:
                C_DATA = 6788;
                N_FC = 6624;
                C_FC = 5664;
                break;
              case PILOT_PP8:
                C_DATA = 6788;
                N_FC = 0;
                C_FC = 0;
                break;
            }
          }
          if (paprmode == PAPR_TR || paprmode == PAPR_BOTH) {
            if (C_DATA != 0) {
              C_DATA -= 72;
            }
            if (N_FC != 0) {
              N_FC -= 72;
            }
            if (C_FC != 0) {
              C_FC -= 72;
            }
          }
          break;
        case FFTSIZE_16K:
        case FFTSIZE_16K_T2GI:
          if (carriermode == CARRIERS_NORMAL) {
            switch (pilotpattern) {
              case PILOT_PP1:
                C_DATA = 12418;
                N_FC = 9088;
                C_FC = 6437;
                break;
              case PILOT_PP2:
                C_DATA = 12436;
                N_FC = 11360;
                C_FC = 10476;
                break;
              case PILOT_PP3:
                C_DATA = 12988;
                N_FC = 11360;
                C_FC = 7845;
                break;
              case PILOT_PP4:
                C_DATA = 13002;
                N_FC = 12496;
                C_FC = 11324;
                break;
              case PILOT_PP5:
                C_DATA = 13272;
                N_FC = 12496;
                C_FC = 8709;
                break;
              case PILOT_PP6:
                C_DATA = 13288;
                N_FC = 13064;
                C_FC = 11801;
                break;
              case PILOT_PP7:
                C_DATA = 13416;
                N_FC = 13064;
                C_FC = 11170;
                break;
              case PILOT_PP8:
                C_DATA = 13406;
                N_FC = 0;
                C_FC = 0;
                break;
            }
          }
          else {
            switch (pilotpattern) {
              case PILOT_PP1:
                C_DATA = 12678;
                N_FC = 9280;
                C_FC = 6573;
                break;
              case PILOT_PP2:
                C_DATA = 12698;
                N_FC = 11600;
                C_FC = 10697;
                break;
              case PILOT_PP3:
                C_DATA = 13262;
                N_FC = 11600;
                C_FC = 8011;
                break;
              case PILOT_PP4:
                C_DATA = 13276;
                N_FC = 12760;
                C_FC = 11563;
                break;
              case PILOT_PP5:
                C_DATA = 13552;
                N_FC = 12760;
                C_FC = 8893;
                break;
              case PILOT_PP6:
                C_DATA = 13568;
                N_FC = 13340;
                C_FC = 12051;
                break;
              case PILOT_PP7:
                C_DATA = 13698;
                N_FC = 13340;
                C_FC = 11406;
                break;
              case PILOT_PP8:
                C_DATA = 13688;
                N_FC = 0;
                C_FC = 0;
                break;
            }
          }
          if (paprmode == PAPR_TR || paprmode == PAPR_BOTH) {
            if (C_DATA != 0) {
              C_DATA -= 144;
            }
            if (N_FC != 0) {
              N_FC -= 144;
            }
            if (C_FC != 0) {
              C_FC -= 144;
            }
          }
          break;
        case FFTSIZE_32K:
        case FFTSIZE_32K_T2GI:
          if (carriermode == CARRIERS_NORMAL) {
            switch (pilotpattern) {
              case PILOT_PP1:
                C_DATA = 0;
                N_FC = 0;
                C_FC = 0;
                break;
              case PILOT_PP2:
                C_DATA = 24886;
                N_FC = 22720;
                C_FC = 20952;
                break;
              case PILOT_PP3:
                C_DATA = 0;
                N_FC = 0;
                C_FC = 0;
                break;
              case PILOT_PP4:
                C_DATA = 26022;
                N_FC = 24992;
                C_FC = 22649;
                break;
              case PILOT_PP5:
                C_DATA = 0;
                N_FC = 0;
                C_FC = 0;
                break;
              case PILOT_PP6:
                C_DATA = 26592;
                N_FC = 26128;
                C_FC = 23603;
                break;
              case PILOT_PP7:
                C_DATA = 26836;
                N_FC = 0;
                C_FC = 0;
                break;
              case PILOT_PP8:
                C_DATA = 26812;
                N_FC = 0;
                C_FC = 0;
                break;
            }
          }
          else {
            switch (pilotpattern) {
              case PILOT_PP1:
                C_DATA = 0;
                N_FC = 0;
                C_FC = 0;
                break;
              case PILOT_PP2:
                C_DATA = 25412;
                N_FC = 23200;
                C_FC = 21395;
                break;
              case PILOT_PP3:
                C_DATA = 0;
                N_FC = 0;
                C_FC = 0;
                break;
              case PILOT_PP4:
                C_DATA = 26572;
                N_FC = 25520;
                C_FC = 23127;
                break;
              case PILOT_PP5:
                C_DATA = 0;
                N_FC = 0;
                C_FC = 0;
                break;
              case PILOT_PP6:
                C_DATA = 27152;
                N_FC = 26680;
                C_FC = 24102;
                break;
              case PILOT_PP7:
                C_DATA = 27404;
                N_FC = 0;
                C_FC = 0;
                break;
              case PILOT_PP8:
                C_DATA = 27376;
                N_FC = 0;
                C_FC = 0;
                break;
            }
          }
          if (paprmode == PAPR_TR || paprmode == PAPR_BOTH) {
            if (C_DATA != 0) {
              C_DATA -= 288;
            }
            if (N_FC != 0) {
              N_FC -= 288;
            }
            if (C_FC != 0) {
              C_FC -= 288;
            }
          }
          break;
      }
      if (miso == false) {
        if (guardinterval == GI_1_128 && pilotpattern == PILOT_PP7) {
          N_FC = 0;
          C_FC = 0;
        }
        if (guardinterval == GI_1_32 && pilotpattern == PILOT_PP4) {
          N_FC = 0;
          C_FC = 0;
        }
        if (guardinterval == GI_1_16 && pilotpattern == PILOT_PP2) {
          N_FC = 0;
          C_FC = 0;
        }
        if (guardinterval == GI_19_256 && pilotpattern == PILOT_PP2) {
          N_FC = 0;
          C_FC = 0;
        }
      }

    }


   void L1_Post_impl::update_PLPslices(void)
   {
      const int NCELLS[2][4] = { {  8100,  4050,  2700, 2025 },
                                 { 32400, 16200, 10800, 8100 }  };
      PLPslice plp_s_t;

      PLPslices.clear();

      plp_s_t.frame_idx = l1post.dyn1.frame_idx;
      plp_s_t.Nsubslices = l1post.conf.sub_slices_per_frame / l1pre.num_rf;

      for ( int n_plp=0; n_plp<l1post.conf.num_plp; ++n_plp )
      {
         const L1Pplp &P = l1post.conf.PLP[n_plp];
         const L1P_d_plp &Pd = l1post.dyn1.PLP[n_plp];

         const int &Ncells = NCELLS[ P.plp_fec_type ]
                                   [ P.plp_mod      ];

         int PI = ( P.time_il_type ? P.time_il_length : 1 );

         plp_s_t.N_BLOCKS_IF = Pd.plp_num_blocks;

         if      ( P.plp_type == 0 ||
                   P.plp_type == 1    )
         {
            plp_s_t.start = Pd.plp_start;
            plp_s_t.length = (Pd.plp_num_blocks*Ncells) / PI;

            plp_s_t.sub_slice_idx = 0;

            plp_s_t.l1plp = P;

            PLPslices.push_back(plp_s_t);
         }
         else if ( P.plp_type == 2    )
         {
            const int &Nsubslices = plp_s_t.Nsubslices;

            plp_s_t.l1plp = P;

            for ( int j=0; j<Nsubslices; ++j )
            {
               plp_s_t.start = Pd.plp_start + j * l1post.dyn1.sub_slice_interval;
               plp_s_t.length = (Pd.plp_num_blocks*Ncells) / (Nsubslices*PI);

               plp_s_t.sub_slice_idx = j;

               PLPslices.push_back(plp_s_t);
            }

         }
         else
         {}
      }

   }


    int
    L1_Post_impl::general_work( int noutput_items,
                       gr_vector_int &ninput_items,
                       gr_vector_const_void_star &input_items,
                       gr_vector_void_star &output_items )
    {
      const gr_complex *in = (const gr_complex *) input_items[0];
      gr_complex *out = (gr_complex *) output_items[0];

      static bool synced = false;
      static int symbol = 0; // symbol #
      static uint64_t framestart = 0;

      static int start_addr;
      static int start_addr_nxt;

// handle frame start tag
      std::vector<tag_t> tags;
      std::vector<tag_t>::iterator t;
      get_tags_in_range( tags, 0,
                         nitems_read(0), nitems_read(0) + ninput_items[0] );

      for ( t = tags.begin(); (!synced) && (t != tags.end()); ++t )
      {
         if ( t->key == pmt::intern("L1-Pre") )
         {
            tag_t &tagL1 = *t;
            synced = true;
            symbol = 0;
            framestart = tagL1.offset;
            start_addr = -1;
            start_addr_nxt = -1;

            pmt::pmt_t L1TagValue_t = tagL1.value;
            if ( L1TagValue != L1TagValue_t )
            {
               L1TagValue = L1TagValue_t;
               setup_L1pre( l1pre, L1TagValue );

///l1pre_out( l1pre );

               dvbt2_fftsize = l1pre.s2>>1;
               switch ( dvbt2_fftsize )
               {
                  case FFTSIZE_2K:  { fftsize= 2*1024; break; }
                  case FFTSIZE_8K:  { fftsize= 8*1024; break; }
                  case FFTSIZE_4K:  { fftsize= 4*1024; break; }
                  case FFTSIZE_1K:  { fftsize= 1*1024; break; }
                  case FFTSIZE_16K: { fftsize=16*1024; break; }
                  case FFTSIZE_32K: { fftsize=32*1024; break; }
                  case FFTSIZE_8K_T2GI:  { fftsize= 8*1024; break; }
                  case FFTSIZE_32K_T2GI: { fftsize=32*1024; break; }
                  case FFTSIZE_16K_T2GI: { fftsize=16*1024; break; }
               }

               update_parameter();

               L1pre_sz = 1840 / N_P2;
               L1post_sz = l1pre.l1_post_size / N_P2;
               L1symbsize = L1pre_sz + L1post_sz;

               demod_bits_cnt = l1pre.l1_post_size;
               switch ( l1pre.l1_mod )
               {
                  case 0: {                      break; } // BPSK
///                  case 1: { demod_bits_cnt *= 2; break; } // QPSK
///                  case 2: { demod_bits_cnt *= 4; break; } // 16-QAM
///                  case 3: { demod_bits_cnt *= 6; break; } // 64-QAM
                  default:
                  {
                     synced = false;
                     demod_bits_cnt = 0;
                     pmt::pmt_t msg;
                     msg = pmt::make_tuple( pmt::mp( "L1-Post" ),
                                            pmt::mp( "modulation unknown" ) );
                     message_port_pub(d_msg_port, msg);
                  }
               }
               delete [] demod_bits;
               demod_bits = new bool[demod_bits_cnt];

               Kpost_ex_pad = l1pre.l1_post_info_size + 32;
               Npost_fec_block = ceil( (double)Kpost_ex_pad / KBCH_1_2 );
               KL1_padding = ceil( Kpost_ex_pad / (double)Npost_fec_block )*Npost_fec_block - Kpost_ex_pad;
               Kpost = Kpost_ex_pad + KL1_padding;
               Ksig = Kpost / Npost_fec_block;

               delete [] L1_post_buffer;
               L1_post_buffer = new unsigned char[Kpost_ex_pad];


               demod_bits_cnt_in_blk = demod_bits_cnt / Npost_fec_block;


               for ( int i=0; i<20; ++i ) { bchcnt_in_blk[i] = 360; }
               bchcnt_in_blk[19] -= 168;
               int Npad = ( Ksig > 360 ? floor( (KBCH_1_2-Ksig)/360.0 )  : 20-1 );
               const int post_padding_bqpsk[20] = // BPSK, QPSK only
               { 18, 17, 16, 15, 14, 13, 12, 11, 4, 10, 9, 8, 3, 2, 7, 6, 5, 1, 19, 0 };
               for ( int n=0; n<Npad; ++n ) { bchcnt_in_blk[ post_padding_bqpsk[n] ] = 0; }
               bchcnt_in_blk[ post_padding_bqpsk[Npad] ]
               -= ( Npad == 19 ? 360-Ksig : KBCH_1_2-Ksig-360*Npad );


            }

            consume_each( framestart - nitems_read(0) );
            return(0);

         }

      }


      // Do <+signal processing+>
      // Tell runtime system how many input items we consumed on
      // each input stream.


// not enough to start frame?
      if ( ! synced )
      {
         consume_each(ninput_items[0]);
         return(0);
      }
      else if ( symbol == 0 )
      {
         int maxNin = ninput_items[0] - fftsize;
         if ( maxNin < 0 )
         {
            consume_each(0);
            return(0);
         }

         uint64_t uld = framestart - nitems_read(0); // modulo difference
         if ( uld > 0 )
         {
            consume_each( uld );
            return(0);
         }
// => framestart = nitems_read(0);
      }


      int nout = 0;

      if ( symbol < N_P2 ) // P2
      {


         for ( int kl=0; kl<L1post_sz; ++kl )
         {
            const gr_complex &c = in[ L1pre_sz + kl ];

// BPSK
            demod_bits[ symbol + N_P2*kl ] = ( c.real() < 0.0 );

         }


         if ( symbol == N_P2-1 )
         {
            unsigned char L1post[16200];

            unsigned char *L1_post_buffer_ = L1_post_buffer;

            int sum_nerrors = 0;
            for ( int fecbn=0; fecbn<Npost_fec_block; ++fecbn )
            {
               bool *bitblock = &demod_bits[fecbn*demod_bits_cnt_in_blk];

// depuncture
//  BCH
               bzero( L1post, 16200 );
               bool *bitblock_ = bitblock;
               unsigned char *L1post_ = L1post;
               for ( int blk=0; blk<20; ++blk )
               {
                  memcpy( L1post_, bitblock_, bchcnt_in_blk[blk] );
                  L1post_ += 360;
                  bitblock_ += bchcnt_in_blk[blk];
               }
               memcpy( &L1post[KBCH_1_2], bitblock_, 168 );

//  LDPC
//  ignored

// decode
//  LDPC ignored

//  BCH
               unsigned int errloc[12];
               uint8_t D[NBCH_1_2/8];

               int bits = 0;
               for ( int b=0; b<NBCH_1_2/8; ++b )
               {
                  uint8_t &B = D[b];
                  B = 0;
                  for ( unsigned int mask=0x80; mask>0; mask>>=1 ) { if(L1post[bits++]) B |= mask; }
               }

               int nerrors = decode_bch( BCHctl, D, 879,
                                         D+879, NULL,
                                         NULL, errloc );

               for (int i = 0; i < nerrors; i++)
               {
                  int b = ((errloc[i]>>3)<<3) + (7 - (errloc[i] % 8));
                  if ( b < KBCH_1_2 ) { L1post[ b ] = ( L1post[ b ] ? 0 : 1 ); }
               }

               if ( sum_nerrors > -1 )
               {
                  if ( nerrors < 0 ) { sum_nerrors  = -1;      }
                  else               { sum_nerrors += nerrors; }
               }

               L1post_ = L1post;

               for ( int blk=0; blk<20; ++blk )
               {
                  memcpy( L1_post_buffer_, L1post_, bchcnt_in_blk[blk] );
                  L1_post_buffer_ += bchcnt_in_blk[blk];
                  L1post_ += 360;
               }


// unscrable
               if ( l1pre.l1_post_scrambled != 0 )
               {
                  unsigned char *L1_post_buffer__ = L1_post_buffer_ - Ksig;
                  for ( int i=0; i<Ksig; ++i )
                  { L1_post_buffer__[i] ^= l1_randomize[i]; }
               }


            }

// L1-Post decode
            bool crc = false;
            if ( (crc=check_crc32( L1_post_buffer, Kpost_ex_pad )) )
            {
               setup_L1post( l1post, l1pre, L1_post_buffer );

///l1post_out(l1pre,l1post);

               update_PLPslices();

            }

            if ( pmt::length( message_ports_in() ) )
            {
               int nofplp = PLPslices.size();
               pmt::pmt_t plpnuml = pmt::make_u32vector(nofplp,0);
               for ( int p=0; p<nofplp; ++p )
               {
                  pmt::u32vector_set( plpnuml, p, PLPslices[p].l1plp.plp_id );
               }

               pmt::pmt_t
               msg = pmt::make_tuple( pmt::mp( "L1-Post" ),

                                      pmt::mp( "PLPs" ),
                                         plpnuml,

                                      pmt::mp( "BCH" ),
                                         pmt::from_long( sum_nerrors ),
                                      pmt::mp( "CRC"),
                                         pmt::from_bool(crc) );

               message_port_pub(d_msg_port, msg);
            }

            nout = (C_P2 - L1symbsize) * N_P2;
            start_addr = 0;
            start_addr_nxt = nout;

         }
         else
         {
            nout = 0;
         }

         consume_each(C_P2);

         memcpy( &out[(C_P2 - L1symbsize)*symbol],
                    &in[L1symbsize],
                       sizeof(gr_complex)*(C_P2 - L1symbsize) );

      }
      else // data
      {
         int nin = C_DATA;
         if ( C_FC>0 && symbol == l1pre.num_data_symbols+N_P2-1 )
         { nin = C_FC; }

         consume_each(nin);

         nout = nin;
         memcpy( out, in, sizeof(gr_complex)*nout );

         start_addr = start_addr_nxt;
         start_addr_nxt += nout;
      }

// tagging
      for ( vector<PLPslice>::iterator plp = PLPslices.begin();
            plp != PLPslices.end();
            ++plp )
      {   
         if ( plp->start >= start_addr && plp->start < start_addr_nxt )
         {
            uint64_t tpos = nitems_written(0) + (plp->start-start_addr);

            pmt::pmt_t plpstag;
            setup_PLPslicetag( plpstag, *plp );

            add_item_tag( 0, tpos, pmt::intern("PLPslice"), plpstag );

         }
      }


      if ( symbol < l1pre.num_data_symbols+N_P2-1 )
      { ++symbol; }
      else
      { synced = false; }

      // Tell runtime system how many output items we produced.
      return(nout);
    }

  } /* namespace DVBparts */
} /* namespace gr */

