/* -*- c++ -*- */
/* 
 * Copyright 2019 Ingo D. Rullhusen <d01c@freeshell.de>.
 * 
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <math.h>

#include <volk/volk.h>

#include <gnuradio/io_signature.h>
#include <gnuradio/expj.h>

#include "crc32.h"
#include "L1-Pre_impl.h"

#include <iostream>

using namespace std;

using namespace gr::dtv;


#define BPSK_SIZE 1840


namespace gr {
  namespace DVBparts {

    L1_Pre::sptr
    L1_Pre::make( int gi, int gi_corr )
    {
      return gnuradio::get_initial_sptr
        (new L1_Pre_impl(gi,gi_corr));
    }

    /*
     * The private constructor
     */
   L1_Pre_impl::L1_Pre_impl( int gi, int gi_corr )
      : gr::block("L1-Pre",
        gr::io_signature::make(1,1, sizeof(gr_complex)),
        gr::io_signature::make2(1, 2, sizeof(gr_complex), sizeof(gr_complex)*BPSK_SIZE) ),
        d_gi_ini(gi), d_gi_corr(gi_corr)
   {
      d_msg_port = pmt::mp("msg");
      message_port_register_out(d_msg_port);

      d_gi_size = 0;

      S1 = S2 = -1;
      Ktotal = Kmin = Kmax = Kext = Koff = N_P2 = C_P2 = 0;

      bzero(&l1pre,sizeof(l1pre));

      fftsize = 1024;
      d_fft_cfw = new fft::fft_complex( fftsize, true,  1 );
      d_fft_cbw = new fft::fft_complex( fftsize, false, 1 );

      update_gi_size( d_gi_ini );

      for ( int i=0; i<fftsize; ++i ) { p2_spec[i] = 0; }

      for ( int i=0; i<fftsize; ++i ) { c_norm[i] = 1.0; }

      init_prbs();

      BCHctl = init_bch( /*m*/ 14, /*t*/ 12, /*prim_poly*/ 0x0000402b );


      set_tag_propagation_policy(TPP_DONT);
      set_output_multiple(32*1024);
      set_fixed_rate(false);
   }

/*
 * Our virtual destructor.
 */
   L1_Pre_impl::~L1_Pre_impl()
   {

      delete d_fft_cfw;
      delete d_fft_cbw;

      free_bch(BCHctl);

   }


   void
   L1_Pre_impl::set_gi( int gi ) { d_gi_ini = gi; }

   void
   L1_Pre_impl::set_gi_corr( int gi_corr ) { d_gi_corr = gi_corr; }


   void
   L1_Pre_impl::update_gi_size( int gi )
   {
      switch ( gi )
      {
         case 0: { d_gi_size = fftsize/32; break; }
         case 1: { d_gi_size = fftsize/16; break; }
         case 2: { d_gi_size = fftsize/8; break; }
         case 3: { d_gi_size = fftsize/4; break; }
         case 4: { d_gi_size = fftsize/128; break; }
         case 5: { d_gi_size = (fftsize*19)/128; break; }
         case 6: { d_gi_size = (fftsize*19)/256; break; }
         default:{ d_gi_size = 0; break; }
      }
   }


   void
   L1_Pre_impl::init_prbs(void)
   {
      int sr = 0x7ff;
      int j = 0;

      for (int i = 0; i < MAX_CARRIERS; i++) {
        int b = ((sr) ^ (sr >> 2)) & 1;
        prbs[i] = sr & 1;
        sr >>= 1;
        if (b) {
          sr |= 0x400;
        }
      }

      for (int i = 0; i < (MAX_CHIPS / 8); i++) {
        for (int k = 7; k >= 0; k--) {
          pn_sequence[j] = (pn_sequence_table[i] >> k) & 0x1;
          j = j + 1;
        }
      }
   }

   void
   L1_Pre_impl::update_cmap_P2(void)
   {
      int step, ki;

      const dvbt2_extended_carrier_t carriermode = CARRIERS_NORMAL; // P2 only
      const dvbt2_misogroup_t miso_group = MISO_TX1;

      for (int i = 0; i < Ktotal; i++) {
        p2_carrier_map[i] = DATA_CARRIER;
      }
      if ((fftsize == 32*1024) && (!miso)) {
        step = 6;
      }
      else {
        step = 3;
      }

      for (int i = 0; i < Ktotal; i += step) {
        if (miso && miso_group == MISO_TX2) {
          if (((i / 3) % 2) && (i % 3 == 0)) {
            p2_carrier_map[i] = P2PILOT_CARRIER_INVERTED;
          }
          else {
            p2_carrier_map[i] = P2PILOT_CARRIER;
          }
        }  
        else {
          p2_carrier_map[i] = P2PILOT_CARRIER;
        }
      }

      if (carriermode == CARRIERS_EXTENDED) {
        for (int i = 0; i < Kext; i++) {
          if (miso && miso_group == MISO_TX2) {
            if (((i / 3) % 2) && (i % 3 == 0)) {
              p2_carrier_map[i] = P2PILOT_CARRIER_INVERTED;
            }
            else {
              p2_carrier_map[i] = P2PILOT_CARRIER;
            }
            if ((((i + (Ktotal - Kext)) / 3) % 2) && ((i + (Ktotal - Kext)) % 3 == 0)) {
              p2_carrier_map[i + (Ktotal - Kext)] = P2PILOT_CARRIER_INVERTED;
            }
            else {
              p2_carrier_map[i + (Ktotal - Kext)] = P2PILOT_CARRIER;
            }
          }  
          else {
            p2_carrier_map[i] = P2PILOT_CARRIER;
            p2_carrier_map[i + (Ktotal - Kext)] = P2PILOT_CARRIER;
          }
        }  
      }

      if (miso) {
        p2_carrier_map[Kext + 1] = P2PILOT_CARRIER;
        p2_carrier_map[Kext + 2] = P2PILOT_CARRIER;
        p2_carrier_map[Ktotal - Kext - 2] = P2PILOT_CARRIER;
        p2_carrier_map[Ktotal - Kext - 3] = P2PILOT_CARRIER;
      }

      switch (fftsize) {
        case 1*1024:
          for (int i = 0; i < 10; i++) {
            p2_carrier_map[p2_papr_map_1k[i]] = P2PAPR_CARRIER;
          }
          if (miso) {
            for (int i = 0; i < 10; i++) {
              ki = p2_papr_map_1k[i] + Kext;
              if (i < 9) {
                if (((ki % 3) == 1) && ((ki + 1) != (p2_papr_map_1k[i + 1] + Kext))) {
                  p2_carrier_map[ki + 1] = P2PILOT_CARRIER;
                }
              }  
              else {
                if ((ki % 3) == 1) {
                  p2_carrier_map[ki + 1] = P2PILOT_CARRIER;
                }
              }  
              if (i > 0) {
                if (((ki % 3) == 2) && ((ki - 1) != (p2_papr_map_1k[i - 1] + Kext))) {
                  p2_carrier_map[ki - 1] = P2PILOT_CARRIER;
                }
              }  
              else {
                if ((ki % 3) == 2) {
                  p2_carrier_map[ki - 1] = P2PILOT_CARRIER;
                }
              }  
            }    
          }      
          break;
        case 2*1024:
          for (int i = 0; i < 18; i++) {
            p2_carrier_map[p2_papr_map_2k[i]] = P2PAPR_CARRIER;
          }
          if (miso) {
            for (int i = 0; i < 18; i++) {
              ki = p2_papr_map_2k[i] + Kext;
              if (i < 17) {
                if (((ki % 3) == 1) && ((ki + 1) != (p2_papr_map_2k[i + 1] + Kext))) {
                  p2_carrier_map[ki + 1] = P2PILOT_CARRIER;
                }
              }  
              else {
                if ((ki % 3) == 1) {
                  p2_carrier_map[ki + 1] = P2PILOT_CARRIER;
                }
              }  
              if (i > 0) {
                if (((ki % 3) == 2) && ((ki - 1) != (p2_papr_map_2k[i - 1] + Kext))) {
                  p2_carrier_map[ki - 1] = P2PILOT_CARRIER;
                }
              }  
              else {
                if ((ki % 3) == 2) {
                  p2_carrier_map[ki - 1] = P2PILOT_CARRIER;
                }
              }  
            }    
          }      
          break;
        case 4*1024:
          for (int i = 0; i < 36; i++) {
            p2_carrier_map[p2_papr_map_4k[i]] = P2PAPR_CARRIER;
          }
          if (miso) {
            for (int i = 0; i < 36; i++) {
              ki = p2_papr_map_4k[i] + Kext;
              if (i < 35) {
                if (((ki % 3) == 1) && ((ki + 1) != (p2_papr_map_4k[i + 1] + Kext))) {
                  p2_carrier_map[ki + 1] = P2PILOT_CARRIER;
                }
              }  
              else {
                if ((ki % 3) == 1) {
                  p2_carrier_map[ki + 1] = P2PILOT_CARRIER;
                }
              }  
              if (i > 0) {
                if (((ki % 3) == 2) && ((ki - 1) != (p2_papr_map_4k[i - 1] + Kext))) {
                  p2_carrier_map[ki - 1] = P2PILOT_CARRIER;
                }
              }  
              else {
                if ((ki % 3) == 2) {
                  p2_carrier_map[ki - 1] = P2PILOT_CARRIER;
                }
              }  
            }    
          }      
          break;
        case 8*1024:
          for (int i = 0; i < 72; i++) {
            p2_carrier_map[p2_papr_map_8k[i] + Kext] = P2PAPR_CARRIER;
          }
          if (miso) {
            for (int i = 0; i < 72; i++) {
              ki = p2_papr_map_8k[i] + Kext;
              if (i < 71) {
                if (((ki % 3) == 1) && ((ki + 1) != (p2_papr_map_8k[i + 1] + Kext))) {
                  p2_carrier_map[ki + 1] = P2PILOT_CARRIER;
                }
              }  
              else {
                if ((ki % 3) == 1) {
                  p2_carrier_map[ki + 1] = P2PILOT_CARRIER;
                }
              }  
              if (i > 0) {
                if (((ki % 3) == 2) && ((ki - 1) != (p2_papr_map_8k[i - 1] + Kext))) {
                  p2_carrier_map[ki - 1] = P2PILOT_CARRIER;
                }
              }  
              else {
                if ((ki % 3) == 2) {
                  p2_carrier_map[ki - 1] = P2PILOT_CARRIER;
                }
              }  
            }    
          }      
          break;
        case 16*1024:
          for (int i = 0; i < 144; i++) {
            p2_carrier_map[p2_papr_map_16k[i] + Kext] = P2PAPR_CARRIER;
          }
          if (miso) {
            for (int i = 0; i < 144; i++) {
              ki = p2_papr_map_16k[i] + Kext;
              if (i < 143) {
                if (((ki % 3) == 1) && ((ki + 1) != (p2_papr_map_16k[i + 1] + Kext))) {
                  p2_carrier_map[ki + 1] = P2PILOT_CARRIER;
                }
              }  
              else {
                if ((ki % 3) == 1) {
                  p2_carrier_map[ki + 1] = P2PILOT_CARRIER;
                }
              }  
              if (i > 0) {
                if (((ki % 3) == 2) && ((ki - 1) != (p2_papr_map_16k[i - 1] + Kext))) {
                  p2_carrier_map[ki - 1] = P2PILOT_CARRIER;
                }
              }  
              else {
                if ((ki % 3) == 2) {
                  p2_carrier_map[ki - 1] = P2PILOT_CARRIER;
                }
              }  
            }    
          }      
          break;
        case 32*1024:
          for (int i = 0; i < 288; i++) {
            p2_carrier_map[p2_papr_map_32k[i] + Kext] = P2PAPR_CARRIER;
          }
          if (miso) {
            for (int i = 0; i < 288; i++) {
              ki = p2_papr_map_32k[i] + Kext;
              if (i < 287) {
                if (((ki % 3) == 1) && ((ki + 1) != (p2_papr_map_32k[i + 1] + Kext))) {
                  p2_carrier_map[ki + 1] = P2PILOT_CARRIER;
                }
              }  
              else {
                if ((ki % 3) == 1) {
                  p2_carrier_map[ki + 1] = P2PILOT_CARRIER;
                }
              }  
              if (i > 0) {
                if (((ki % 3) == 2) && ((ki - 1) != (p2_papr_map_32k[i - 1] + Kext))) {
                  p2_carrier_map[ki - 1] = P2PILOT_CARRIER;
                }
              }  
              else {
                if ((ki % 3) == 2) {
                  p2_carrier_map[ki - 1] = P2PILOT_CARRIER;
                }
              }  
            }    
          }      
          break; 
      }

   }

   void
   L1_Pre_impl::update_H(void)
   {
      int max_states, xor_size, pn_mask, result;
      int q_evenP2 = 0;
      int q_oddP2 = 0;
      int lfsr = 0;
      int logic1k[2] = {0, 4};
      int logic2k[2] = {0, 3};
      int logic4k[2] = {0, 2};
      int logic8k[4] = {0, 1, 4, 6};
      int logic16k[6] = {0, 1, 4, 5, 9, 11};
      int logic32k[4] = {0, 1, 2, 12};
      int *logic;
      const int *bitpermeven, *bitpermodd;
      int pn_degree, even, odd;

      switch (fftsize) {
        case 1*1024:
          pn_degree = 9;
          pn_mask = 0x1ff;
          max_states = 1024;
          logic = &logic1k[0];
          xor_size = 2;
          bitpermeven = &bitperm1keven[0];
          bitpermodd = &bitperm1kodd[0];  
          break;
        case 2*1024:
          pn_degree = 10;
          pn_mask = 0x3ff;
          max_states = 2048;
          logic = &logic2k[0];
          xor_size = 2;
          bitpermeven = &bitperm2keven[0];
          bitpermodd = &bitperm2kodd[0];  
          break;
        case 4*1024:
          pn_degree = 11;
          pn_mask = 0x7ff;
          max_states = 4096;
          logic = &logic4k[0];
          xor_size = 2;  
          bitpermeven = &bitperm4keven[0];
          bitpermodd = &bitperm4kodd[0];  
          break;
        case 8*1024:
          pn_degree = 12;
          pn_mask = 0xfff;
          max_states = 8192;
          logic = &logic8k[0];
          xor_size = 4;
          bitpermeven = &bitperm8keven[0];
          bitpermodd = &bitperm8kodd[0];  
          break;
        case 16*1024:
          pn_degree = 13;
          pn_mask = 0x1fff;
          max_states = 16384;
          logic = &logic16k[0];
          xor_size = 6;
          bitpermeven = &bitperm16keven[0];
          bitpermodd = &bitperm16kodd[0];  
          break;
        case 32*1024:
          pn_degree = 14;      
          pn_mask = 0x3fff;    
          max_states = 32768;  
          logic = &logic32k[0];
          xor_size = 4;
          bitpermeven = &bitperm32k[0];
          bitpermodd = &bitperm32k[0];
          break;
        default:
          pn_degree = 0; 
          pn_mask = 0;   
          max_states = 0;
          logic = &logic1k[0];
          xor_size = 0;
          break;
      }

      for (int i = 0; i < max_states; i++) {
        if (i == 0 || i == 1) {
          lfsr = 0;
        }
        else if (i == 2) {
          lfsr = 1;
        }
        else {
          result = 0;
          for (int k = 0; k < xor_size; k++) {
            result ^= (lfsr >> logic[k]) & 1; 
          }
          lfsr &= pn_mask;
          lfsr >>= 1;
          lfsr |= result << (pn_degree - 1);
        }

        even = 0;
        odd = 0; 
        for (int n = 0; n < pn_degree; n++) {
          even |= ((lfsr >> n) & 0x1) << bitpermeven[n];
        }
        for (int n = 0; n < pn_degree; n++) {
          odd |= ((lfsr >> n) & 0x1) << bitpermodd[n];
        }
        even = even + ((i % 2) * (max_states / 2));
        odd = odd + ((i % 2) * (max_states / 2));  
        if (even < C_P2) {
          HevenP2[q_evenP2++] = even;
        }
        if (odd < C_P2) {
          HoddP2[q_oddP2++] = odd;
        }
      }  

      if (fftsize == 32*1024) {
        for (int j = 0; j < q_oddP2; j++) {
          int a;
          a = HoddP2[j];
          HevenP2[a] = j;
        }
      }

   }



    void
    L1_Pre_impl::forecast (int noutput_items, gr_vector_int &ninput_items_required)
    {
      ninput_items_required[0] = (int)ceil( noutput_items * (double)(fftsize+d_gi_size)/(double)(fftsize) );
    }

    int
    L1_Pre_impl::general_work (int noutput_items,
                       gr_vector_int &ninput_items,
                       gr_vector_const_void_star &input_items,
                       gr_vector_void_star &output_items)
    {
      const pmt::pmt_t P2startKey( pmt::intern("P2_start") );
      const gr_complex *in = (const gr_complex *) input_items[0];
      gr_complex *out_ifft = (gr_complex *) output_items[0];
      gr_complex *out_bpsk = (gr_complex *) output_items[1];

      int maxNin = ninput_items[0] - fftsize - d_gi_size;

      static bool synced = false;
      static uint64_t symbstart = 0;

      static int symbol = 0; // symbol #
      static unsigned char bpsk[BPSK_SIZE];
      static unsigned char L1pre_t[200];
      static pmt::pmt_t L1TagValue;

      static int t_shift = 0;
      static double t_shift_f = 0.0;

      if ( maxNin < 0 )
      {
         consume_each(0);
         return(0);
      }

// handle frame start tag
      std::vector<tag_t> tags;
      get_tags_in_range( tags, 0,
                         nitems_read(0), nitems_read(0) + ninput_items[0],
                         P2startKey );

      if ( (!synced) && tags.size() > 0 )
      {
         tag_t &tag = tags.front();

         unsigned int s1, s2;
         s1 = pmt::to_long(pmt::tuple_ref(tag.value,0));
         s2 = pmt::to_long(pmt::tuple_ref(tag.value,1));

         symbstart = tag.offset + d_gi_corr;
         synced = true;
         symbol = 0;

// change?
         if ( s1 != S1 || s2 != S2 )
         {
            miso = ( s1&0x1 || s1&0x100 );

            if ( ((s1&0x6) == 0) || (s1 == 3) || (s1 == 4) )
            {
               switch ( s2>>1 )
               {
                  case 0: { fftsize=2*1024; break; }
                  case 1: { fftsize=8*1024; break; }
                  case 2: { fftsize=4*1024; break; }
                  case 3: { fftsize=1*1024; break; }
                  case 4: { fftsize=16*1024; break; }
                  case 5: { fftsize=32*1024; break; }
                  case 6: { fftsize=8*1024; break; }
                  case 7: { fftsize=32*1024; break; }
               }

               delete d_fft_cfw;
               d_fft_cfw = new fft::fft_complex( fftsize, true,  1 );
               delete d_fft_cbw;
               d_fft_cbw = new fft::fft_complex( fftsize, false, 1 );

               update_gi_size( d_gi_ini );

               for ( int i=0; i<fftsize; ++i ) { c_norm[i] = 1.0; }

               switch( fftsize )
               {
                  case 1*1024:
                  { Kmin = 0; Kmax = 852; Kext = 0; N_P2 = 16; C_P2 = (miso? 546 : 558 ); break; }
                  case 2*1024:
                  { Kmin = 0; Kmax = 1704; Kext = 0; N_P2 = 8; C_P2 = (miso? 1098 : 1118 ); break; }
                  case 4*1024:
                  { Kmin = 0; Kmax = 3408; Kext = 0; N_P2 = 4; C_P2 = (miso? 2198 : 2236 ); break; }
                  case 8*1024:
                  { Kmin = 0; Kmax = 6816; Kext = 48; N_P2 = 2; C_P2 = (miso? 4398 : 4472 ); break; }
                  case 16*1024:
                  { Kmin = 0; Kmax = 13632; Kext = 144; N_P2 = 1; C_P2 = (miso? 8814 : 8944 ); break; }
                  case 32*1024:
                  { Kmin = 0; Kmax = 27264; Kext = 288; N_P2 = 1; C_P2 = (miso? 17612 : 22432 ); break; }
               }

               Koff = Kext; Kext = 0; // no ext mode
               Ktotal = Kmax + 1;


               update_H();
               update_cmap_P2();

               if ( (s1==0) && (fftsize==32*1024) ) // 32 SISO
               { A_P2 = sqrt(37.0)/5.0; }
               else
               { A_P2 = sqrt(31.0)/5.0; }

               for ( int i=0; i<fftsize; ++i ) { p2_spec[i] = 0; }
               for ( int k=0; k<Ktotal; ++k )
               {
                  int ks = k - (Kmax-Kmin)/2;
                  int idx = (ks < 0 ? ks+fftsize : ks );

                  if ( p2_carrier_map[k] == P2PILOT_CARRIER )
                  {
                     if ( prbs[k + Koff ] ^ pn_sequence[0] )
                     { p2_spec[idx] = -A_P2; }
                     else
                     { p2_spec[idx] = +A_P2; }
                  }
                  else if ( p2_carrier_map[k] == P2PILOT_CARRIER_INVERTED )
                  {
                     if ( prbs[k + Koff ] ^ pn_sequence[0] )
                     { p2_spec[idx] = +A_P2; }
                     else
                     { p2_spec[idx] = -A_P2; }
                  }

               }


            }

            S1 = s1;
            S2 = s2;
         }

      }


      // Do <+signal processing+>
      // Tell runtime system how many input items we consumed on
      // each input stream.


// not enough to start frame?
      if ( ! synced )
      {
         consume_each(ninput_items[0]);
         return(0);
      }
      else if ( symbol == 0 )
      {
         uint64_t uld = symbstart - nitems_read(0); // modulo difference
         if ( uld > maxNin )
         {
            consume_each( maxNin );
            return(0);
         }
         else if ( uld > 0 )
         {
            consume_each( uld );
            return(0);
         }
// => symbstart = nitems_read(0);
      }


      gr_complex *fftbuff1_c, *fftbuff2_c;
      fftbuff1_c = d_fft_cfw->get_inbuf();
      memcpy( fftbuff1_c, in, sizeof(gr_complex)*fftsize );

      d_fft_cfw->execute();

      fftbuff2_c = d_fft_cfw->get_outbuf();


      if ( symbol < N_P2 ) // P2
      {
         int out_offset = symbol * fftsize;

         if ( symbol==0 )
         {
// time shift
            gr_complex *fftcorr;
            fftcorr = d_fft_cbw->get_inbuf();
            volk_32fc_32f_multiply_32fc(
               fftcorr,
                  fftbuff2_c,
                  p2_spec,
            fftsize);

            d_fft_cbw->execute();
            fftcorr = d_fft_cbw->get_outbuf();


            float max_corr = 0.0, y1 = 0, y3 = 0;
            gr_complex max_c_v = 0.0;
            for ( int i=0; i<500; ++i )
            {
               float av = abs( fftcorr[i] );
               if ( av > max_corr )
               {
                  t_shift = i;
                  max_corr = av;
                  y3 = abs( fftcorr[i+1] );
                  y1 = abs( fftcorr[ ( i==0 ? fftsize-1 : i-1 ) ] );
                  max_c_v = fftcorr[i];
               }
            }
            for ( int i=fftsize-500; i<fftsize; ++i )
            {
               float av = abs( fftcorr[i] );
               if ( av > max_corr )
               {
                  t_shift = i - fftsize;
                  max_corr = av;
                  y1 = abs( fftcorr[i-1] );
                  y3 = abs( fftcorr[ ( i<fftsize-1 ? i+1 : 0 ) ] );
                  max_c_v = fftcorr[i];
               }
            }
            t_shift_f = (y1-y3)/( 2.0 * (y3-2.0*max_corr+y1) );

            gr_complex tcfac = gr_expj(+2.0*M_PI * ( t_shift + t_shift_f ) / ( (double)fftsize ) );
            c_norm[0] = 1.0;
            for ( int i=1; i<fftsize/2; ++i )
            {
               c_norm[i] = c_norm[i-1]*tcfac;
            }
            tcfac = conj(tcfac);
            c_norm[fftsize-1] = tcfac;
            for ( int i=fftsize-2; i>=fftsize/2; --i )
            {
               c_norm[i] = c_norm[i+1]*tcfac; 
            }

         }

// constellation reference phase from normal part
         gr_complex c0phase = 0.0;
         int c0cnt = 0;

         for ( int k=0; k<Ktotal; ++k )
         {
            int ks = k - (Kmax-Kmin)/2;
            int idx = (ks < 0 ? ks+fftsize : ks );

            switch ( p2_carrier_map[k] )
            {
               case P2PILOT_CARRIER:
               {
                  const gr_complex cv = fftbuff2_c[idx] * c_norm[idx];
                  if ( prbs[k + Koff ] ^ pn_sequence[symbol] )
                  { c0phase -= cv; }
                  else
                  { c0phase += cv; }

                  ++c0cnt;
                  break;  
               }
               case P2PILOT_CARRIER_INVERTED:
               {
                  const gr_complex cv = fftbuff2_c[idx] * c_norm[idx];
                  if ( prbs[k + Koff ] ^ pn_sequence[symbol] )
                  { c0phase += cv; }
                  else
                  { c0phase -= cv; }

                  ++c0cnt;
                  break;  
               }
            }

         }

         c0phase = (float)c0cnt/c0phase;


         volk_32fc_x2_multiply_32fc(
            &out_ifft[out_offset],
               fftbuff2_c,
               c_norm,
         fftsize);


// freq deinterleave
         int bits = 0;
         int *H;
         if ( (symbol % 2) == 0 ) { H = HevenP2; }
         else                     { H = HoddP2;  }

         for ( int k=0; (k<Ktotal) && (bits<C_P2); ++k )
         {
            int ks = k - (Kmax-Kmin)/2;
            int idx = (ks < 0 ? ks+fftsize : ks );

            if ( p2_carrier_map[k] == DATA_CARRIER )
            {
               gr_complex cv = fftbuff2_c[idx] * c_norm[idx] * c0phase;

               if ( H[bits] < 1840/N_P2 )
               {
                  out_bpsk[ symbol + N_P2*H[bits] ] =   cv;
                  bpsk    [ symbol + N_P2*H[bits] ] = ( cv.real() < 0 ? 1 : 0 );
               }

               ++bits;
            }

         }


         if ( symbol == N_P2-1 )
         {
            unsigned char L1pre[16200];

// depuncture
            for ( int i=0; i<200; ++i ) { L1pre[i] = bpsk[i]; }
            for ( int i=200; i<3072; ++i ) { L1pre[i] = 0; }
            for ( int i=3072; i<3240; ++i ) { L1pre[i] = bpsk[ i - 2872 ]; }

//  LDPC ignored
//  BCH
            unsigned int errloc[12];
            uint8_t D[3240/8];

            bits = 0;
            for ( int b=0; b<3240/8; ++b )
            {
               uint8_t &B = D[b];
               B = 0;
               for ( unsigned int mask=0x80; mask>0; mask>>=1 ) { if(L1pre[bits++]) B |= mask; }
            }

            int nerrors = decode_bch( BCHctl, D, 384,
                                      D+384, NULL,
                                      NULL, errloc );

            for (int i = 0; i < nerrors; i++)
            {
               int b = ((errloc[i]>>3)<<3) + (7 - (errloc[i] % 8));
               if ( b < 200 ) { L1pre[ b ] = ( L1pre[ b ] ? 0 : 1 ); }
            }


            bool crc = false;
            if ( (crc=check_crc32( L1pre, 200 )) )
            {
               if ( memcmp((const char*)L1pre, (const char*)L1pre_t, 200 ) != 0 )
               {
                  setup_L1pre( l1pre, L1pre );
                  memcpy( L1pre_t, L1pre, 200 );

                  update_gi_size( l1pre.guard_interval );

                  setup_L1tag( L1TagValue, l1pre );

               }

///l1pre_out(l1pre);
               uint64_t tpos = nitems_written(0);
               add_item_tag( 0, tpos, pmt::intern("L1-Pre"), L1TagValue );

            }

            if ( pmt::length( message_ports_in() ) )
            {
               pmt::pmt_t
               msg = pmt::make_tuple( pmt::mp( "L1-Pre" ),
                                      pmt::mp( "t_shift" ),
                                      pmt::from_long( t_shift ),
                                      pmt::mp( "t_shift_f" ),
                                      pmt::from_float( t_shift_f ),
                                      pmt::mp( "BCH" ),
                                      pmt::from_long( nerrors ),
                                      pmt::mp( "CRC"),
                                      pmt::from_bool(crc) );

               message_port_pub(d_msg_port, msg);
            }

            produce( 0, N_P2 * fftsize );
            produce( 1, 1 );
         }
         else
         {
            produce( 0, 0 );
            produce( 1, 0 );
         }


      }
      else // data
      {
         volk_32fc_x2_multiply_32fc(
            out_ifft,
               fftbuff2_c,
               c_norm,
         fftsize);

         produce( 0, fftsize );
         produce( 1, 0 );
      }

      if ( symbol < l1pre.num_data_symbols+N_P2-1 )
      { ++symbol; }
      else
      { synced = false; }

      consume_each( fftsize + d_gi_size );

      // Tell runtime system how many output items we produced.
      return( WORK_CALLED_PRODUCE );
    }

  } /* namespace DVBparts */
} /* namespace gr */

