/* -*- c++ -*- */
/* 
 * Copyright 2017 Ingo D. Rullhusen <d01c@freeshell.de>.
 * 
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <float.h>

#include <volk/volk.h>

#include <gnuradio/io_signature.h>
#include "P1_Detect_impl.h"

namespace gr {
  namespace DVBparts {

    P1_Detect::sptr
    P1_Detect::make(int TA, int TB, int TC)
    {
      return gnuradio::get_initial_sptr
        (new P1_Detect_impl(TA, TB, TC));
    }

    /*
     * The private constructor
     */
   P1_Detect_impl::P1_Detect_impl(int TA, int TB, int TC)
   : gr::sync_block("P1_Detect",
        gr::io_signature::make(1,1, sizeof(gr_complex)),
        gr::io_signature::make(1,1, sizeof(float)) ),
     d_TA(TA), d_TB(TB), d_TC(TC)
   {
      d_TR = d_TA;

      accDTR1 = accDTR2 = 0;

      d_proc_size = d_TA * 1024;

      unsigned int alignment = volk_get_alignment();

      EXPO = (gr_complex*)volk_malloc(sizeof(gr_complex)*d_proc_size, alignment);
      EXPO_TC = (gr_complex*)volk_malloc(sizeof(gr_complex)*d_proc_size, alignment);
      for ( int i=0; i<d_proc_size; ++i )
      {
         EXPO[i] = exp( gr_complex( 0.0, -2.0*M_PI*i/(double)d_TR ) );
         EXPO_TC[i] = exp( gr_complex( 0.0, -2.0*M_PI*(i-d_TC-d_TA)/(double)d_TR ) );
      }

      CTMP1 = (gr_complex*)volk_malloc(sizeof(gr_complex)*d_proc_size, alignment);
      CTMP2 = (gr_complex*)volk_malloc(sizeof(gr_complex)*d_proc_size, alignment);
      CTMP3 = (gr_complex*)volk_malloc(sizeof(gr_complex)*d_proc_size, alignment);

      C1t = (gr_complex*)volk_malloc(sizeof(gr_complex)*(d_TR+d_proc_size), alignment);
      CORR_1 = &C1t[d_TR];
      C2t = (gr_complex*)volk_malloc(sizeof(gr_complex)*(d_TR+d_proc_size), alignment);
      CORR_2 = &C2t[d_TR];
      for ( int i=0; i<d_TR; ++i )
      {
         C1t[i] = 0;
         C2t[i] = 0;
      }

      FTMP1 = (float*)volk_malloc(sizeof(float)*d_proc_size, alignment);
      FTMP2 = (float*)volk_malloc(sizeof(float)*d_proc_size, alignment);

      invNq = 1.0 / ( (double)d_TR * (double)d_TR );
      rms = 0.0;
      norm = 1.0;

      set_output_multiple( d_proc_size );
      set_history( d_TC + d_TA + 1 );

    }

/*
 * Our virtual destructor.
 */
   P1_Detect_impl::~P1_Detect_impl()
   {
      volk_free(EXPO);
      volk_free(EXPO_TC);
      volk_free(CTMP1);
      volk_free(CTMP2);
      volk_free(CTMP3);

      volk_free(C1t);
      volk_free(C2t);

      volk_free(FTMP1);
      volk_free(FTMP2);

   }

   int
   P1_Detect_impl::work(int noutput_items,
       gr_vector_const_void_star &input_items,
       gr_vector_void_star &output_items)
   {
      const gr_complex *in = (const gr_complex *) input_items[0];
      float *trigger = (float *) output_items[0];

      float rms_t;
      volk_32fc_magnitude_squared_32f(
         FTMP1,
            in,
      noutput_items );
      volk_32f_accumulator_s32f(
         &rms_t,
            FTMP1,
      noutput_items );
      if ( rms_t < FLT_MAX )
      { rms = 0.4*rms + 0.6*rms_t/(double)noutput_items; }

      norm = invNq / (rms*rms);

      float re, im;

      volk_32fc_deinterleave_real_32f(
         FTMP1,
            C1t,
      d_TR );
      volk_32f_accumulator_s32f(
         &re,
            FTMP1,
      d_TR );
      volk_32fc_deinterleave_imag_32f(
         FTMP1,
            C1t,
      d_TR ); 
      volk_32f_accumulator_s32f(
         &im,
            FTMP1,
      d_TR );
      gr_complex accDTR1(re,im);

      volk_32fc_deinterleave_real_32f(
         FTMP2,
            C2t,
      d_TR );
      volk_32f_accumulator_s32f(
         &re,
            FTMP2,
      d_TR );
      volk_32fc_deinterleave_imag_32f(
         FTMP2,
            C2t,
      d_TR ); 
      volk_32f_accumulator_s32f(
         &im,
            FTMP2,
      d_TR );
      gr_complex accDTR2(re,im);

      int nofb = noutput_items / d_proc_size;
      for ( int b=0; b<nofb; ++b )
      {
         const gr_complex *IN_b = &in[ (b*d_proc_size) ];

         volk_32fc_x2_multiply_32fc(
            CTMP1,
               IN_b,
               EXPO_TC,
         d_proc_size);

         volk_32fc_x2_multiply_conjugate_32fc(
            CORR_1,
               IN_b + d_TC,
               CTMP1,
         d_proc_size);

         volk_32fc_x2_multiply_32fc(
            CTMP1,
               IN_b + (d_TA+d_TC),
               EXPO,
         d_proc_size);

         volk_32fc_x2_multiply_conjugate_32fc(
            CORR_2,
               CTMP1,
               IN_b + (d_TA+d_TC - d_TB),
         d_proc_size);

         volk_32f_x2_subtract_32f(
            (float*)CTMP1,
               (float*)CORR_1,
               (float*)C1t,
         2*d_proc_size);
         volk_32f_x2_subtract_32f(
            (float*)CTMP2,
               (float*)CORR_2,
               (float*)C2t,
         2*d_proc_size);
         for ( int n=0; n<d_proc_size; ++n )
         {
            accDTR1 = ( CTMP1[ n ] += accDTR1 );
            accDTR2 = ( CTMP2[ n ] += accDTR2 );
         }
         memcpy(C1t, &CORR_1[d_proc_size-d_TR], sizeof(gr_complex)*d_TR);
         memcpy(C2t, &CORR_2[d_proc_size-d_TR], sizeof(gr_complex)*d_TR);

         volk_32fc_x2_multiply_conjugate_32fc(
            CTMP3,
               CTMP1,
               CTMP1,
         d_proc_size);
         volk_32fc_deinterleave_real_32f(
            FTMP1,
               CTMP3,
         d_proc_size);

         volk_32fc_x2_multiply_conjugate_32fc(
            CTMP3,
               CTMP2,
               CTMP2,
         d_proc_size);
         volk_32fc_deinterleave_real_32f(
            FTMP2,
               CTMP3,
         d_proc_size);

         float *TRIGGER = &trigger[ (b*d_proc_size) ];
         volk_32f_x2_multiply_32f(
            TRIGGER,
               FTMP1,
               FTMP2,
         d_proc_size);
         volk_32f_sqrt_32f(
            TRIGGER,
               TRIGGER,
         d_proc_size);
         volk_32f_s32f_multiply_32f(
            TRIGGER,
               TRIGGER,
               norm,
         d_proc_size);

      }

// Tell runtime system how many output items we produced.
      return noutput_items;
   }

  } /* namespace DVBparts */
} /* namespace gr */

