/* -*- c++ -*- */
/* 
 * Copyright 2018 Ingo D. Rullhusen <d01c@freeshell.de>.
 * 
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gnuradio/io_signature.h>
#include "PLPselector_impl.h"

using namespace std;

namespace gr {
  namespace DVBparts {

    PLPselector::sptr
    PLPselector::make( int id )
    {
      return gnuradio::get_initial_sptr
        (new PLPselector_impl(id));
    }

    /*
     * The private constructor
     */
    PLPselector_impl::PLPselector_impl( int id )
      : gr::block("PLPselector",
        gr::io_signature::make(1,1, sizeof(gr_complex)),
        gr::io_signature::make(1,1, sizeof(gr_complex))),
        d_plp_id(id)
    {
       set_tag_propagation_policy(TPP_DONT);
       set_fixed_rate(false);
    }

    /*
     * Our virtual destructor.
     */
    PLPselector_impl::~PLPselector_impl()
    {
    }


   void PLPselector_impl::set_plp_id( int id ) { d_plp_id = id; }

   void PLPselector_impl::forecast( int noutput_items, gr_vector_int &ninput_items_required )
   { ninput_items_required[0] = noutput_items; }

    int
    PLPselector_impl::general_work (int noutput_items,
                       gr_vector_int &ninput_items,
                       gr_vector_const_void_star &input_items,
                       gr_vector_void_star &output_items)
    {
      const gr_complex *in = (const gr_complex *) input_items[0];
      gr_complex *out = (gr_complex *) output_items[0];

      static bool synced = false;
      static int items_to_read = 0;
      static uint64_t framestart = 0;

// handle frame start tag
      std::vector<tag_t> tags;
      std::vector<tag_t>::iterator t;
      get_tags_in_range( tags, 0,
                         nitems_read(0), nitems_read(0) + ninput_items[0] );

      for ( t = tags.begin(); (!synced) && (t != tags.end()); ++t )
      {
         if ( t->key == pmt::intern("PLPslice") )
         {
            tag_t &PLPstag = *t;

            const pmt::pmt_t &plptag = pmt::tuple_ref(PLPstag.value,5);
            const unsigned int plp_id = pmt::u32vector_ref(plptag,0);

            if ( d_plp_id == plp_id )
            {
               framestart = PLPstag.offset;

               items_to_read = pmt::to_long(pmt::tuple_ref(PLPstag.value,0));

               consume_each( framestart - nitems_read(0) );

               if ( items_to_read > 0 )
               {
                  synced = true;

                  uint64_t tpos = nitems_written(0);
                  add_item_tag( 0, tpos, pmt::intern("PLPslice"), PLPstag.value );

               }

               return(0);
            }

         }

      }


      if ( ! synced )
      {
         consume_each(ninput_items[0]);
         return(0);
      }


      int nout = noutput_items;
      if ( nout > ninput_items[0] ) { nout = ninput_items[0]; }
      if ( nout > items_to_read   ) { nout = items_to_read;   }

      consume_each( nout );
      memcpy( out, in, sizeof(gr_complex)*nout );
      items_to_read -= nout;

      if ( items_to_read < 1 ) { synced = false; }

      // Tell runtime system how many output items we produced.
      return(nout);
    }

  } /* namespace DVBparts */
} /* namespace gr */

