/* -*- c++ -*- */
/* 
 * Copyright 2018 Ingo D. Rullhusen <d01c@freeshell.de>.
 * 
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gnuradio/io_signature.h>
#include "bch_decoder_impl.h"


using namespace std;


namespace gr {
  namespace DVBparts {

    bch_decoder::sptr
    bch_decoder::make()
    {
      return gnuradio::get_initial_sptr
        (new bch_decoder_impl());
    }

    /*
     * The private constructor
     */
    bch_decoder_impl::bch_decoder_impl()
      : gr::block("bch_decoder",
              gr::io_signature::make(1,1, sizeof(unsigned char)),
              gr::io_signature::make(1,1, sizeof(unsigned char)))
    {
       forecast_f = 1.0;
       NBCH = 0;
       KBCH = 0;

       bzero(&l1plp,sizeof(l1plp));
       setup_L1Pplp_tag( PLPTagValue, l1plp );

       BCHctl = NULL;

       set_tag_propagation_policy(TPP_DONT);
    }

    /*
     * Our virtual destructor.
     */
    bch_decoder_impl::~bch_decoder_impl()
    {
       free_bch(BCHctl);
    }

    void
    bch_decoder_impl::forecast (int noutput_items, gr_vector_int &ninput_items_required)
    { ninput_items_required[0] = ceil( forecast_f * noutput_items + 1 ); }

    void bch_decoder_impl::decode()
    {
//  BCH
       unsigned int errloc[12];
       uint8_t D[ BCH_SIZE_MAX /8];

       int bits = 0;
       for ( int b=0; b<NBCH/8; ++b )
       {
          uint8_t &B = D[b];
          B = 0;
          for ( unsigned int mask=0x80; mask>0; mask>>=1 ) { if(fec_buffer[bits++]) B |= mask; }
       }

       int nerrors = decode_bch( BCHctl, D, KBCH/8,
                                 D+(KBCH/8), NULL,
                                 NULL, errloc );

       for (int i = 0; i < nerrors; i++)
       {
          int b = ((errloc[i]>>3)<<3) + (7 - (errloc[i] % 8));
          if ( b < KBCH ) { fec_buffer[ b ] = ( fec_buffer[ b ] ? 0 : 1 ); }
       }

//clog << "BCH " << nerrors << endl;
//if ( nerrors > -1 ) { clog << "BCH ok " << nerrors << endl; }

    }

    int
    bch_decoder_impl::general_work (int noutput_items,
                       gr_vector_int &ninput_items,
                       gr_vector_const_void_star &input_items,
                       gr_vector_void_star &output_items)
    {
       const pmt::pmt_t PLPfecKey( pmt::intern("PLPfec") );
       const unsigned char *in = (const unsigned char *) input_items[0];
       unsigned char *out = (unsigned char *) output_items[0];

       static enum { Wait, Read, Write } state = Wait;

       static int q = 0;
       if ( state == Write )
       {
          int nout = noutput_items;
          if ( nout >= KBCH-q ) { nout = KBCH-q; state = Wait; }
          memcpy( out, &fec_buffer[q], nout*sizeof(unsigned char) );
          q += nout;

          consume_each(0);
          return(nout);
       }

       static int items_to_read = 0;
       if ( state == Read )
       {
          int nin = items_to_read;
          if ( nin > ninput_items[0] ) { nin = ninput_items[0]; }
          memcpy( &fec_buffer[fec_buffer_offset], in, nin*sizeof(unsigned char) );
          fec_buffer_offset += nin;
          items_to_read -= nin;

          if ( items_to_read < 1 ) { decode(); q = 0; state = Write; }

          consume_each(nin);
          return(0);
       }


// handle frame start tag
       uint64_t uld = ninput_items[0];

       std::vector<tag_t> tags;
       std::vector<tag_t>::iterator t;

       get_tags_in_range( tags, 0,
                          nitems_read(0), nitems_read(0) + uld,
                          PLPfecKey );

       if ( tags.size() > 0 )
       {
          tag_t &tagPLP = tags[0];
          int fec_size = pmt::to_long(pmt::tuple_ref(tagPLP.value,0));
          pmt::pmt_t PLPTagValue_t = pmt::tuple_ref(tagPLP.value,1);
          if ( ! equal( PLPTagValue, PLPTagValue_t ) )
          {
             PLPTagValue = PLPTagValue_t;
             setup_L1Pplp( l1plp, PLPTagValue_t );

             free_bch(BCHctl);

             if ( l1plp.plp_fec_type == 1 ) // 64K
             {
                int t;
                switch ( l1plp.plp_cod )
                {
                   case 0: // 1/2
                   { KBCH = 32208; NBCH = 32400; t = 12; break; }
                   case 1: // 3/5
                   { KBCH = 38688; NBCH = 38880; t = 12; break; }
                   case 2: // 2/3
                   { KBCH = 43040; NBCH = 43200; t = 10; break; }
                   case 3: // 3/4
                   { KBCH = 48408; NBCH = 48600; t = 12; break; }
                   case 4: // 4/5
                   { KBCH = 51648; NBCH = 51840; t = 12; break; }
                   case 5: // 5/6
                   { KBCH = 53840; NBCH = 54000; t = 10; break; }
                }

                BCHctl = init_bch( /*m*/ 16, /*t*/ t, /*prim_poly*/ 0x0001002d );
             }
             else // 16K
             {
                switch ( l1plp.plp_cod )
                {
                   case 0: // 1/2
                   { KBCH = 7032; NBCH = 7200; break; }
                   case 1: // 3/5
                   { KBCH = 9552; NBCH = 9720; break; }
                   case 2: // 2/3
                   { KBCH = 10632; NBCH = 10800; break; }
                   case 3: // 3/4
                   { KBCH = 11712; NBCH = 11880; break; }
                   case 4: // 4/5
                   { KBCH = 12432; NBCH = 12600; break; }
                   case 5: // 5/6
                   { KBCH = 13152; NBCH = 13320; break; }
                }

                BCHctl = init_bch( /*m*/ 14, /*t*/ 12, /*prim_poly*/ 0x0000402b );
             }
if ( BCHctl==NULL ) { clog << "BCH FAILED!" << endl; }

             forecast_f = NBCH / (double)KBCH;
          }

          uint64_t framestart = tagPLP.offset;
          uld = framestart - nitems_read(0); // modulo difference

          uint64_t tpos = nitems_written(0);
          add_item_tag( 0, tpos, pmt::intern("PLPfec"),
                           pmt::make_tuple(
                              pmt::from_long(KBCH),
                              PLPTagValue
                                          )
                      );

          items_to_read = NBCH;
          fec_buffer_offset = 0;
          state = Read;
       }

       consume_each(uld);
       return(0);
    }

  } /* namespace DVBparts */
} /* namespace gr */

