/* -*- c++ -*- */
/* 
 * Copyright 2018 Ingo D. Rullhusen <d01c@freeshell.de>.
 * 
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gnuradio/io_signature.h>
#include "bit_deinterleaver_impl.h"

using namespace std;


namespace gr {
  namespace DVBparts {

    bit_deinterleaver::sptr
    bit_deinterleaver::make()
    {
      return gnuradio::get_initial_sptr
        (new bit_deinterleaver_impl());
    }

    /*
     * The private constructor
     */
    bit_deinterleaver_impl::bit_deinterleaver_impl()
      : gr::sync_block("bit_deinterleaver",
              gr::io_signature::make(1,1, sizeof(unsigned char)),
              gr::io_signature::make(1,1, sizeof(unsigned char)))
    {
       fec_size = 0;

       bzero(&l1plp,sizeof(l1plp));
       setup_L1Pplp_tag( PLPTagValue, l1plp );

       set_output_multiple(64800);
       set_history(16+1);
    }

    /*
     * Our virtual destructor.
     */
    bit_deinterleaver_impl::~bit_deinterleaver_impl()
    {
    }


    const int perm_qpsk_cnt = 2;
    const int perm_qpsk[] = { 0, 1 };

// 1/2 3/4 4/5 5/6
    const int perm_16qam_a_cnt = 8;
    const int perm_16qam_a[] = { 7, 1, 4, 2, 5, 3, 6, 0 };

    const int perm_64qam_a_cnt = 12;
    const int perm_64qam_a[] = { 11, 7, 3, 10, 6, 2, 9, 5, 1, 8, 4, 0 };

    const int perm_256qam1_a_cnt = 16;
    const int perm_256qam1_a[] = { 15, 1, 13, 3, 8, 11, 9, 5, 10, 6, 4, 7, 12, 2, 14, 0 };

    const int perm_256qam2_a_cnt = 8;
    const int perm_256qam2_a[] = { 7, 3, 1, 5, 2, 6, 4, 0 };

// 3/5
    const int perm_16qam1_b_cnt = 8;
    const int perm_16qam1_b[] = { 0, 5, 1, 2, 4, 7, 3, 6 };

    const int perm_16qam2_b_cnt = 8;
    const int perm_16qam2_b[] = { 7, 1, 4, 2, 5, 3, 6, 0 };

    const int perm_64qam1_b_cnt = 12;
    const int perm_64qam1_b[] = { 2, 7, 6, 9, 0, 3, 1, 8, 4, 11, 5, 10 };

    const int perm_64qam2_b_cnt = 12;
    const int perm_64qam2_b[] = { 11, 7, 3, 10, 6, 2, 9, 5, 1, 8, 4, 0 };

    const int perm_256qam1_b_cnt = 16;
    const int perm_256qam1_b[] = { 2, 11, 3, 4, 0, 9, 1, 8, 10, 13, 7, 14, 6, 15, 5, 12 };

    const int perm_256qam2_b_cnt = 8;
    const int perm_256qam2_b[] = { 7, 3, 1, 5, 2, 6, 4, 0 };

// 2/3
    const int perm_16qam_c_cnt = 8;
    const int perm_16qam_c[] = { 7, 1, 4, 2, 5, 3, 6, 0 };

    const int perm_64qam_c_cnt = 12;
    const int perm_64qam_c[] = { 11, 7, 3, 10, 6, 2, 9, 5, 1, 8, 4, 0 };

    const int perm_256qam1_c_cnt = 16;
    const int perm_256qam1_c[] = { 7, 2, 9, 0, 4, 6, 13, 3, 14, 10, 15, 5, 8, 12, 11, 1 };

    const int perm_256qam2_c_cnt = 8;
    const int perm_256qam2_c[] = { 7, 3, 1, 5, 2, 6, 4, 0 };


    int
    bit_deinterleaver_impl::work(int noutput_items,
        gr_vector_const_void_star &input_items,
        gr_vector_void_star &output_items)
    {
       const pmt::pmt_t PLPfecKey( pmt::intern("PLPfec") );
       const unsigned char *in = (const unsigned char *) input_items[0];
       unsigned char *out = (unsigned char *) output_items[0];

       std::vector<tag_t> tags;
       get_tags_in_window( tags, 0, 0, noutput_items, PLPfecKey );

       static int e_cnt = perm_qpsk_cnt;
       static const int *permute = perm_qpsk;
       static int e = 0;

       in += 16;

       for ( int q=0; q<noutput_items; ++q )
       {
          uint64_t qoff = q + nitems_read(0);
          tag_t *tagPLP = NULL;
          for ( std::vector<tag_t>::iterator t=tags.begin(); t!=tags.end(); ++t )
          { if ( qoff == t->offset ) { tagPLP = &(*t); break; } }
          if ( tagPLP )
          {
             fec_size = pmt::to_long(pmt::tuple_ref(tagPLP->value,0));
             pmt::pmt_t PLPTagValue_t = pmt::tuple_ref(tagPLP->value,1);
             if ( ! equal( PLPTagValue, PLPTagValue_t ) )
             {
                PLPTagValue = PLPTagValue_t;
                setup_L1Pplp( l1plp, PLPTagValue_t );
                switch ( l1plp.plp_mod )
                {

                   case 0: // QPSK
                   { e_cnt = perm_qpsk_cnt; permute = perm_qpsk; break; }

                   case 1: // 16-QAM
                   {
                      switch ( l1plp.plp_cod )
                      {
                         case 0: // 1/2
                         case 3: // 3/4
                         case 4: // 4/5
                         case 5: // 5/6
                         {
                            e_cnt = perm_16qam_a_cnt;
                            permute = perm_16qam_a;
                            break;
                         }
                         case 1: // 3/5
                         {
                            if ( l1plp.plp_fec_type == 1 ) // 64K
                            {
                               e_cnt = perm_16qam1_b_cnt;
                               permute = perm_16qam1_b;
                            }
                            else // 16K
                            {
                               e_cnt = perm_16qam2_b_cnt;
                               permute = perm_16qam2_b;
                            }
                            break;
                         }
                         case 2: // 2/3
                         {
                            e_cnt = perm_16qam_c_cnt;
                            permute = perm_16qam_c;
                            break;
                         }
                      }

                      break;
                   }

                   case 2: // 64-QAM
                   {
                      switch ( l1plp.plp_cod )
                      {
                         case 0: // 1/2
                         case 3: // 3/4
                         case 4: // 4/5
                         case 5: // 5/6
                         {
                            e_cnt = perm_64qam_a_cnt;
                            permute = perm_64qam_a;
                            break;
                         }
                         case 1: // 3/5
                         {
                            if ( l1plp.plp_fec_type == 1 ) // 64K
                            {
                               e_cnt = perm_64qam1_b_cnt;
                               permute = perm_64qam1_b;
                            }
                            else // 16K
                            {
                               e_cnt = perm_64qam2_b_cnt;
                               permute = perm_64qam2_b;
                            }
                            break;
                         }
                         case 2: // 2/3
                         {
                            e_cnt = perm_64qam_c_cnt;
                            permute = perm_64qam_c;
                            break;
                         }
                      }

                      break;
                   }

                   case 3: // 256-QAM
                   {

                      switch ( l1plp.plp_cod )
                      {
                         case 0: // 1/2
                         case 3: // 3/4
                         case 4: // 4/5
                         case 5: // 5/6
                         {
                            if ( l1plp.plp_fec_type == 1 ) // 64K
                            {
                               e_cnt = perm_256qam1_a_cnt;
                               permute = perm_256qam1_a;
                            }
                            else // 16K
                            {
                               e_cnt = perm_256qam2_a_cnt;
                               permute = perm_256qam2_a;
                            }
                            break;
                         }
                         case 1: // 3/5
                         {
                            if ( l1plp.plp_fec_type == 1 ) // 64K
                            {
                               e_cnt = perm_256qam1_b_cnt;
                               permute = perm_256qam1_b;
                            }
                            else // 16K
                            {
                               e_cnt = perm_256qam2_b_cnt;
                               permute = perm_256qam2_b;
                            }
                            break;
                         }
                         case 2: // 2/3
                         {
                            if ( l1plp.plp_fec_type == 1 ) // 64K
                            {
                               e_cnt = perm_256qam1_c_cnt;
                               permute = perm_256qam1_c;
                            }
                            else // 16K
                            {
                               e_cnt = perm_256qam2_c_cnt;
                               permute = perm_256qam2_c;
                            }
                            break;
                         }
                      }

                      break;
                   }

                }

             }

             e = 0;
          }

          out[q] = in[ (q-e) + permute[e] ];

          if ( ++e >= e_cnt ) { e = 0; }

       }

      return(noutput_items);
    }

  } /* namespace DVBparts */
} /* namespace gr */

