/* -*- c++ -*- */
/* 
 * Copyright 2018 Ingo D. Rullhusen <d01c@freeshell.de>.
 * 
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gnuradio/io_signature.h>
#include "cell_deinterleaver_impl.h"


using namespace std;

using namespace gr::dtv;


namespace gr {
  namespace DVBparts {

    cell_deinterleaver::sptr
    cell_deinterleaver::make()
    {
      return gnuradio::get_initial_sptr
        (new cell_deinterleaver_impl());
    }

    /*
     * The private constructor
     */
    cell_deinterleaver_impl::cell_deinterleaver_impl()
      : gr::block("cell_deinterleaver",
              gr::io_signature::make(1,1, sizeof(gr_complex)),
              gr::io_signature::make(1,1, sizeof(gr_complex)))
    {
       ti_blocks = -1;
       fec_blocks = -1;

       bzero(&l1plp,sizeof(l1plp));
       setup_L1Pplp_tag( PLPTagValue, l1plp );

       set_tag_propagation_policy(TPP_DONT);
       set_output_multiple(FRAME_SIZE_NORMAL);
    }

    /*
     * Our virtual destructor.
     */
    cell_deinterleaver_impl::~cell_deinterleaver_impl()
    {}


   void cell_deinterleaver_impl::setup_permutations()
   {
      const dvb_framesize_t &framesize = (dvb_framesize_t)l1plp.plp_fec_type;
      const dvb_constellation_t &constellation = (dvb_constellation_t)l1plp.plp_mod;

      int max_states, xor_size, pn_mask, result, q = 0;
      int lfsr = 0;
      int logic11[2] = {0, 3};
      int logic12[2] = {0, 2};
      int logic13[4] = {0, 1, 4, 6};
      int logic14[6] = {0, 1, 4, 5, 9, 11};
      int logic15[4] = {0, 1, 2, 12};
      int *logic;
      if (framesize == FECFRAME_NORMAL) {
        switch (constellation) {
          case MOD_QPSK:
            cell_size = 32400;
            pn_degree = 15;
            pn_mask = 0x3fff;
            max_states = 32768;
            logic = &logic15[0];
            xor_size = 4;
            break;
          case MOD_16QAM:
            cell_size = 16200;
            pn_degree = 14;
            pn_mask = 0x1fff;
            max_states = 16384;
            logic = &logic14[0];
            xor_size = 6;
            break;
          case MOD_64QAM:
            cell_size = 10800;
            pn_degree = 14;
            pn_mask = 0x1fff;
            max_states = 16384;
            logic = &logic14[0];
            xor_size = 6;
            break;
          case MOD_256QAM:
            cell_size = 8100;
            pn_degree = 13;
            pn_mask = 0xfff;
            max_states = 8192;
            logic = &logic13[0];
            xor_size = 4;
            break;
          default:
            cell_size = 32400;
            pn_degree = 15;
            pn_mask = 0x3fff;
            max_states = 32768;
            logic = &logic15[0];
            xor_size = 4;
            break;
        }
      }
      else {
        switch (constellation) {
          case MOD_QPSK:
            cell_size = 8100;
            pn_degree = 13;
            pn_mask = 0xfff;
            max_states = 8192;
            logic = &logic13[0];
            xor_size = 4;
            break;
          case MOD_16QAM:
            cell_size = 4050;
            pn_degree = 12;
            pn_mask = 0x7ff;
            max_states = 4096;
            logic = &logic12[0];
            xor_size = 2;
            break;
          case MOD_64QAM:
            cell_size = 2700;
            pn_degree = 12;
            pn_mask = 0x7ff;
            max_states = 4096;
            logic = &logic12[0];
            xor_size = 2;
            break;
          case MOD_256QAM:
            cell_size = 2025;
            pn_degree = 11;
            pn_mask = 0x3ff;
            max_states = 2048;
            logic = &logic11[0];
            xor_size = 2;
            break;
          default:
            cell_size = 8100;
            pn_degree = 13;
            pn_mask = 0xfff;
            max_states = 8192;
            logic = &logic13[0];
            xor_size = 4;
            break;
        }
      }
      for (int i = 0; i < max_states; i++) {
        if (i == 0 || i == 1) {
          lfsr = 0;
        }
        else if (i == 2) {
          lfsr = 1;
        }
        else {
          result = 0;
          for (int k = 0; k < xor_size; k++) {
            result ^= (lfsr >> logic[k]) & 1;
          }
          lfsr &= pn_mask;
          lfsr >>= 1;
          lfsr |= result << (pn_degree - 2);
        }
        lfsr |= (i % 2) << (pn_degree - 1);
        if (lfsr < cell_size) {
          permutations[q++] = lfsr;
        }
      }

   }


    void
    cell_deinterleaver_impl::forecast (int noutput_items, gr_vector_int &ninput_items_required)
    { ninput_items_required[0] = noutput_items; }

    int
    cell_deinterleaver_impl::general_work (int noutput_items,
                       gr_vector_int &ninput_items,
                       gr_vector_const_void_star &input_items,
                       gr_vector_void_star &output_items)
    {
       const gr_complex *in = (const gr_complex *) input_items[0];
       gr_complex *out = (gr_complex *) output_items[0];

       static enum { Wait, Read, Write } state = Wait;


       static int n = 0;
       static int r = 0;
       static int s = 0;


       static int shift = 0;
       static int q = 0;
       if ( state == Write )
       {
          int nout = 0;
          for ( ; q<cell_size; ++q )
          {
             if ( nout < noutput_items )
             {
                out[nout++] = ci_buffer[ ((permutations[q] + shift) % cell_size) ];
/////                out[nout++] = ci_buffer[ q ];
             }
             else 
             { goto qExit; }
          }
          state = Wait;
qExit:
          consume_each(0);
          return(nout);
       }


       static int items_to_read = 0;
       if ( state == Read )
       {
          int nin = items_to_read;
          if ( nin > ninput_items[0] ) { nin = ninput_items[0]; }
          memcpy( &ci_buffer[ci_buffer_offset], in, nin*sizeof(gr_complex) );
          ci_buffer_offset += nin;
          items_to_read -= nin;

          if ( items_to_read < 1 ) { state = Write; }

          consume_each(nin);
          return(0);
       }


// next FEC block
       if ( r < fec_blocks )
       {
          int temp;

          shift = cell_size;
          while (shift >= cell_size) {
            temp = n;
            shift = 0;
            for (int p = 0; p < pn_degree; p++) {
              shift |= temp & 1;
              shift <<= 1;
              temp >>= 1; 
            }
            n++;
          }

          uint64_t tpos = nitems_written(0);
          add_item_tag( 0, tpos, pmt::intern("PLPfec"),
                           pmt::make_tuple(
                              pmt::from_long(cell_size),
                              PLPTagValue
                                          )
                      );

          q = 0;

          state = Read;
          ci_buffer_offset = 0;
          items_to_read = cell_size;
          ++r;

          consume_each(0);
          return(0);
       }


// handle frame start tag
       uint64_t uld = ninput_items[0];

       std::vector<tag_t> tags;
       get_tags_in_range( tags, 0,
                          nitems_read(0), nitems_read(0) + uld,
                          pmt::intern("PLPti") );

       if ( tags.size() > 0 )
       {
          tag_t &tagTI = tags[0];

          items_to_read = pmt::to_long(pmt::tuple_ref(tagTI.value,0));
          fec_blocks = pmt::to_long(pmt::tuple_ref(tagTI.value,1));
          s = pmt::to_long(pmt::tuple_ref(tagTI.value,2));

          const pmt::pmt_t &plp_tag = pmt::tuple_ref(tagTI.value,3);

          n = 0;
          r = 0;

          if ( ! equal( PLPTagValue, plp_tag ) )
          {
             PLPTagValue = plp_tag;
             setup_L1Pplp( l1plp, plp_tag );

             setup_permutations();

             ti_blocks = ( l1plp.time_il_type ? 1 : l1plp.time_il_length );

          }

          uint64_t framestart = tagTI.offset;
          uld = framestart - nitems_read(0); // modulo difference

       }


       consume_each(uld);
       return(0);
    }

  } /* namespace DVBparts */
} /* namespace gr */

