/* -*- c++ -*- */
/* 
 * Copyright 2018 Ingo D. Rullhusen <d01c@freeshell.de>.
 * 
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <volk/volk.h>

#include <gnuradio/io_signature.h>
#include "demodulator_impl.h"

using namespace std;

namespace gr {
  namespace DVBparts {

    demodulator::sptr
    demodulator::make()
    {
      return gnuradio::get_initial_sptr
        (new demodulator_impl());
    }

    /*
     * The private constructor
     */
    demodulator_impl::demodulator_impl()
      : gr::block("demodulator",
              gr::io_signature::make(1,1, sizeof(gr_complex)),
              gr::io_signature::make(1,1, sizeof(unsigned char)))
    {
       forecast_f = 1.0;

       bzero(&l1plp,sizeof(l1plp));
       setup_L1Pplp_tag( PLPTagValue, l1plp );

       cell_size = 0;
       eta_mod = 1;

       unsigned int alignment = volk_get_alignment();
       c_re = (float*)volk_malloc(sizeof(float)*FRAME_SIZE_NORMAL, alignment);
       c_im = (float*)volk_malloc(sizeof(float)*FRAME_SIZE_NORMAL, alignment);
       c_t  = (float*)volk_malloc(sizeof(float)*FRAME_SIZE_NORMAL, alignment);

       set_tag_propagation_policy(TPP_DONT);
       set_output_multiple(FRAME_SIZE_NORMAL);
    }


    const unsigned char symbits_qam[4][16][4]
       = { { {1},  {0} },
           { {1,0}, {1,1},  {0,1}, {0,0} },
           { {1,0,0}, {1,0,1}, {1,1,1}, {1,1,0},  {0,1,0}, {0,1,1}, {0,0,1}, {0,0,0} },
           { {1,0,0,0}, {1,0,0,1}, {1,0,1,1}, {1,0,1,0}, {1,1,1,0}, {1,1,1,1}, {1,1,0,1}, {1,1,0,0},  {0,1,0,0}, {0,1,0,1}, {0,1,1,1}, {0,1,1,0}, {0,0,1,0}, {0,0,1,1}, {0,0,0,1}, {0,0,0,0} }
         };

    int demodulator_impl::get_symbits( unsigned char *symbits,
                                       double i_re, double i_im )
    {
       int re0 = round( i_re-0.5 );
       if      ( re0 < 0       ) { re0 = 0;       }
       else if ( re0 > q_steps ) { re0 = q_steps; }

       int im0 = round( i_im-0.5 );
       if      ( im0 < 0       ) { im0 = 0;       }
       else if ( im0 > q_steps ) { im0 = q_steps; }

       const unsigned char (&symbits_qam_t)[16][4] = symbits_qam[ l1plp.plp_mod ];
       unsigned char *symbits_t = symbits;
       for ( int b1=0; b1<eta_mod_2; ++b1 )
       {
          *(symbits_t++) = symbits_qam_t[re0][b1];
          *(symbits_t++) = symbits_qam_t[im0][b1];
       }

       return(eta_mod);
    }


    /*
     * Our virtual destructor.
     */
    demodulator_impl::~demodulator_impl()
    {
       volk_free(c_re);
       volk_free(c_im);
       volk_free(c_t);
    }

    void
    demodulator_impl::forecast (int noutput_items, gr_vector_int &ninput_items_required)
    { ninput_items_required[0] = ceil( forecast_f * noutput_items + 1 ); }

    int
    demodulator_impl::general_work (int noutput_items,
                       gr_vector_int &ninput_items,
                       gr_vector_const_void_star &input_items,
                       gr_vector_void_star &output_items)
    {
       const pmt::pmt_t PLPfecKey( pmt::intern("PLPfec") );
       const gr_complex *in = (const gr_complex *) input_items[0];
       unsigned char *out = (unsigned char *) output_items[0];

       static bool synced = false;
       static uint64_t fecstart = 0;
       static float f = 1.0;
       static float d = 0.0;

// handle frame start tag
       std::vector<tag_t> tags;
       get_tags_in_range( tags, 0,
                          nitems_read(0), nitems_read(0) + ninput_items[0],
                          PLPfecKey );

       if ( tags.size() > 0 && (!synced) )
       {
          tag_t &tagPLP = tags[0];
          synced = true;
          fecstart = tagPLP.offset;

          cell_size = pmt::to_long(pmt::tuple_ref(tagPLP.value,0));
          pmt::pmt_t PLPTagValue_t = pmt::tuple_ref(tagPLP.value,1);
          if ( ! equal( PLPTagValue, PLPTagValue_t ) )
          {
             PLPTagValue = PLPTagValue_t;
             setup_L1Pplp( l1plp, PLPTagValue_t );

             switch ( l1plp.plp_mod )
             {
                case 0: { q_steps =  1; eta_mod = 2; f = sqrt(  2.0)*0.5; d = 1.0; break; } // QPSK
                case 1: { q_steps =  3; eta_mod = 4; f = sqrt( 10.0)*0.5; d = 2.0; break; } // 16-QAM
                case 2: { q_steps =  7; eta_mod = 6; f = sqrt( 42.0)*0.5; d = 4.0; break; } // 64-QAM
                case 3: { q_steps = 15; eta_mod = 8; f = sqrt(170.0)*0.5; d = 8.0; break; } // 256-QAM
                default:
                {
                   synced = false;
                   cerr << "unsupported modulation" << endl;
                   break;
                }
             }
             eta_mod_2 = eta_mod / 2;
             fec_size = cell_size * eta_mod;

             if ( l1plp.plp_rotation )
             {
                synced = false;
                cerr << "rotation not supported" << endl;
             }

             forecast_f = 1.0 / (float)eta_mod;

          }

       }


// not enough to start frame?
      if ( ! synced )
      {
         consume_each(ninput_items[0]);
         return(0);
      }
      else
      {
         int maxNin = ninput_items[0] - cell_size;
         if ( maxNin < 0 )
         {
            consume_each(0);
            return(0);
         }

         uint64_t uld = fecstart - nitems_read(0); // modulo difference
         if ( uld > 0 )
         {
            consume_each( uld );
            return(0);
         }
// => framestart = nitems_read(0);
      }


//
      uint64_t tpos = nitems_written(0);
      add_item_tag( 0, tpos, PLPfecKey,
                       pmt::make_tuple(
                          pmt::from_long(fec_size),
                          PLPTagValue
                                      )
                  );

      volk_32fc_deinterleave_32f_x2(c_re,c_im, in, cell_size);

      volk_32f_s32f_multiply_32f(c_t, c_re, f, cell_size);
      for ( unsigned int ii = 0; ii < cell_size; ++ii )
      { c_re[ii] = c_t[ii] + d; }

      volk_32f_s32f_multiply_32f(c_t, c_im, f, cell_size);  
      for ( unsigned int ii = 0; ii < cell_size; ++ii )
      { c_im[ii] = c_t[ii] + d; }

      int nout = 0;
      for ( int q=0; q<cell_size; ++q )
      {
         nout += get_symbits( &out[nout], c_re[q], c_im[q] );
      }
///if ( nout > FRAME_SIZE_NORMAL ) { clog << "too much output" << endl; }

      synced = false;

      consume_each(cell_size);

      return(nout);
    }

  } /* namespace DVBparts */
} /* namespace gr */

