/* -*- c++ -*- */
/* 
 * Copyright 2021 Ingo D. Rullhusen <d01c@freeshell.de>.
 * 
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gnuradio/io_signature.h>
#include "ldpc_decoder_impl.h"


#ifdef _OPENMP
#include <omp.h>
#else
#warning OpenMP is NOT used
#endif 


using namespace std;


namespace gr {
  namespace DVBparts {


    template <typename TABLE>
    void SparseHMatrix::update_sparse_matrix()
    {
       delete [] row;
       delete [] value;
       delete [] bits_in_syndrome;

       N = TABLE::N;
       K = TABLE::K;
       Q = ( TABLE::N - TABLE::K ) / TABLE::M;

       row = new unsigned int[ K + 1 ];

       value = new unsigned int[ TABLE::LINKS_TOTAL ];

       bits_in_syndrome = new std::set<int>[ N - K ];

       row[0] = 0;
       if ( N == 0 ) { return; } // init only

       for ( int m=0; m<K; ++m )
       {
          int Hrow = floor( m / TABLE::M );
          int Hoffset = 0;
          if ( Hrow < TABLE::LEN[0] )
          {
             Hoffset = Hrow * TABLE::DEG[0];
             row[ m+1 ] = row[ m ] + TABLE::DEG[0];
          }
          else
          {
             Hoffset = TABLE::LEN[0]*TABLE::DEG[0] + ( Hrow - TABLE::LEN[0] )*TABLE::DEG[1];
             row[ m+1 ] = row[ m ] + TABLE::DEG[1];
          }

          for ( int p_idx=row[m]; p_idx<row[m+1]; ++p_idx )
          {
             int p = p_idx - row[m];
             value[ p_idx ] = ( TABLE::POS[ p+Hoffset ] + ( m % TABLE::M )*Q ) % ( N - K );

             bits_in_syndrome[ value[ p_idx ] ].insert(m);

          }

       }

       bits_in_syndrome[0].insert( K );
       for ( int p=1; p<N-K; ++p )
       {
          bits_in_syndrome[ p ].insert( K+p-1 );
          bits_in_syndrome[ p ].insert( K+p   );
       }

    }


    ldpc_decoder::sptr
    ldpc_decoder::make()
    {
      return gnuradio::get_initial_sptr
        (new ldpc_decoder_impl());
    }

    /*
     * The private constructor
     */
    ldpc_decoder_impl::ldpc_decoder_impl()
      : gr::block("ldpc_decoder",
              gr::io_signature::make(1,1, sizeof(unsigned char)),
              gr::io_signature::make(1,1, sizeof(unsigned char)))
    {
       fec_buffer_offset = 0;
       forecast_f = 1.0;

       bzero(&l1plp,sizeof(l1plp));
       setup_L1Pplp_tag( PLPTagValue, l1plp );

       set_output_multiple( UFRAME_MAX_SIZE_NORMAL * MAX_CHUNK_SIZE );
       set_tag_propagation_policy(TPP_DONT);
    }

    /*
     * Our virtual destructor.
     */
    ldpc_decoder_impl::~ldpc_decoder_impl()
    {
    }


    void ldpc_decoder_impl::update_tables()
    {
       int im = 0;
       int index = 0;

       if ( l1plp.plp_fec_type == 1 ) // 64K
       {
          switch ( l1plp.plp_cod )
          {
             case 0: // 1/2
             { H.update_sparse_matrix<DVB_S2_TABLE_B4>(); break; }
             case 1: // 3/5
             { H.update_sparse_matrix<DVB_S2_TABLE_B5>(); break; }
             case 2: // 2/3
             { H.update_sparse_matrix<DVB_T2_TABLE_A3>(); break; }
             case 3: // 3/4
             { H.update_sparse_matrix<DVB_S2_TABLE_B7>(); break; }
             case 4: // 4/5
             { H.update_sparse_matrix<DVB_S2_TABLE_B8>(); break; }
             case 5: // 5/6
             { H.update_sparse_matrix<DVB_S2_TABLE_B9>(); break; }
          }
       }
       else // 16K
       {
          switch ( l1plp.plp_cod )
          {
             case 0: // 1/2
             { H.update_sparse_matrix<DVB_T2X_TABLE_NULL>(); cerr<<"TAB fehlt"<<endl;exit(1); break; }
             case 1: // 3/5
             { H.update_sparse_matrix<DVB_T2_TABLE_B3>(); break; }
             case 2: // 2/3
             { H.update_sparse_matrix<DVB_T2X_TABLE_NULL>(); cerr<<"TAB fehlt"<<endl;exit(1); break; }
             case 3: // 3/4
             { H.update_sparse_matrix<DVB_T2X_TABLE_NULL>(); cerr<<"TAB fehlt"<<endl;exit(1); break; }
             case 4: // 4/5
             { H.update_sparse_matrix<DVB_T2X_TABLE_NULL>(); cerr<<"TAB fehlt"<<endl;exit(1); break; }
             case 5: // 5/6
             { H.update_sparse_matrix<DVB_T2X_TABLE_NULL>(); cerr<<"TAB fehlt"<<endl;exit(1); break; }
          }
       }

    }

    int ldpc_decoder_impl::fec_decoding( int thread )
    {
       bool *fec_buffer = &fec_buffer_n[ thread * H.N ]; // FRAME_SIZE_NORMAL ];
       bool *syndrome_d = &syndrome_d_n[ thread * MAX_SYN_SIZE ];
       bool *syndrome_p = &syndrome_p_n[ thread * MAX_SYN_SIZE ];
       uint8_t *error_counter = &error_counter_n[ thread * FRAME_SIZE_NORMAL ];

       bzero( syndrome_d, (H.N-H.K)*sizeof(bool) );

       for ( int m=0; m<H.K; ++m )
       {
          for ( int idx=H.row[m]; idx<H.row[m+1]; ++idx )
          {
             syndrome_d[ H.value[idx] ] ^= fec_buffer[m];
          }
       }

       bzero( syndrome_p, (H.N-H.K)*sizeof(bool) );

       syndrome_p[0] = fec_buffer[ H.K ];
       for ( int p=1; p<H.N-H.K; ++p )
       {
          syndrome_p[ p ] = fec_buffer[H.K+p-1] ^ fec_buffer[H.K+p];
       }

       bzero( error_counter, H.N*sizeof(uint8_t) );
       for ( int p=0; p<H.N-H.K; ++p )
       {
          if ( syndrome_p[ p ] != syndrome_d[ p ] )
          {
             for ( std::set<int>::iterator it_b = H.bits_in_syndrome[p].begin();
                   it_b != H.bits_in_syndrome[p].end();
                   ++it_b )
             {
                ++(error_counter[ *it_b ]);
             }
          }
       }

       int max_ec = 0;
       for ( int m=0; m<H.N; ++m )
       {
          if ( max_ec < error_counter[m] ) { max_ec = error_counter[m]; }
       }

       if ( max_ec == 0 ) { return(0); }

       for ( int m=0; m<H.N; ++m )
       {
          if ( max_ec == error_counter[m] )
          {
             fec_buffer[ m ] = ! fec_buffer[ m ];
          }
       }

       return(max_ec);
    }

    int ldpc_decoder_impl::decode( int max_iter, int thread )
    {
       int iter = 0, max_ec = 1;

       while ( iter < max_iter && max_ec > 0 )
       {
          max_ec = fec_decoding( thread );
          ++iter;
       }
       return( iter );
    }


    void
    ldpc_decoder_impl::forecast (int noutput_items, gr_vector_int &ninput_items_required)
    { ninput_items_required[0] = round( forecast_f * noutput_items ); }

    int
    ldpc_decoder_impl::general_work (int noutput_items,
                       gr_vector_int &ninput_items,
                       gr_vector_const_void_star &input_items,
                       gr_vector_void_star &output_items)
    {
       const pmt::pmt_t PLPfecKey( pmt::intern("PLPfec") );
       const unsigned char *in = (const unsigned char *) input_items[0];
       unsigned char *out = (unsigned char *) output_items[0];

       static int in_chunk_size = 0;
       static enum { Wait, Read, Write } state = Wait;

       static int items_to_read = 0;

       int nin = 0;

       if ( state == Read )
       {
          nin = items_to_read;
          if ( nin > ninput_items[0] ) { nin = ninput_items[0]; }
          for ( int i = 0; i<nin; ++i )
          {
             fec_buffer_n[fec_buffer_offset+i] = ( in[i]>0 );
          }
          fec_buffer_offset += nin;
          items_to_read -= nin;

          if ( items_to_read < 1 )
          {
#pragma omp parallel for
             for ( int n=0; n<in_chunk_size; ++n )
             {
                int count = 0;

                count = decode( MAX_ITER_COUNT, n );
/***
                if (count >= MAX_ITER_COUNT )
                {
#pragma omp critical
                   clog << "TN=" << omp_get_thread_num() << ": LDPC decoder failed at converging to a code word." << endl;
                }
                else
                {
#pragma omp critical
                   clog << "TN=" << omp_get_thread_num() << ": LDPC trial count = " << count << endl;
                }
***/
             }

             fec_buffer_offset = 0;
             state = Write;

             forecast_f = 0.0;
          }

          if ( state == Read ) { consume_each(nin); return(0); }
       }


       if ( state == Write )
       {
          int out_chunk_size = noutput_items / H.K;
          if ( out_chunk_size > in_chunk_size ) { out_chunk_size = in_chunk_size; }
          int nout = out_chunk_size * H.K;

          int out_offset = 0;
          for ( int b=0; b<out_chunk_size; ++b )
          {
             uint64_t tpos = nitems_written(0) + b*H.K;
             add_item_tag( 0, tpos, pmt::intern("PLPfec"),
                              pmt::make_tuple(
                                 pmt::from_long(H.K),
                                 PLPTagValue
                                             )
                         );

             for ( int i = 0; i<H.K; ++i )
             {
                out[ out_offset + i ] = ( fec_buffer_n[ fec_buffer_offset + i ] ? 1 : 0 );
             }

             out_offset += H.K;
             fec_buffer_offset += H.N;

          }

          if ( fec_buffer_offset >= in_chunk_size*H.N )
          {
             state = Wait;
             forecast_f = H.N/(double)H.K;
          }

          consume_each(nin);
          return(nout);
       }


// handle frame start tag
       uint64_t uld = ninput_items[0];

       std::vector<tag_t> tags;

       get_tags_in_range( tags, 0,
                          nitems_read(0), nitems_read(0) + uld,
                          PLPfecKey );

       if ( tags.size() > 0 )
       {
          tag_t &tagPLP = tags[0];
          int fec_size = pmt::to_long(pmt::tuple_ref(tagPLP.value,0));
          pmt::pmt_t PLPTagValue_t = pmt::tuple_ref(tagPLP.value,1);

          if ( ! equal( PLPTagValue, PLPTagValue_t ) )
          {
             PLPTagValue = PLPTagValue_t;
             setup_L1Pplp( l1plp, PLPTagValue_t );

             update_tables();

             forecast_f = H.N / (double)H.K;

          }

          in_chunk_size = 0;
          for ( int tagidx=0; tagidx < ( tags.size() > MAX_CHUNK_SIZE ? MAX_CHUNK_SIZE : tags.size() ); ++tagidx )
          {
             tag_t &tagPLP_t = tags[ tagidx ];
             pmt::pmt_t PLPTagValue_t = pmt::tuple_ref(tagPLP_t.value,1);
             if ( equal( PLPTagValue, PLPTagValue_t ) )
             {
                ++in_chunk_size;
             }
             else
             {
                break;
             }
          }

          uint64_t framestart = tagPLP.offset;
          uld = framestart - nitems_read(0); // modulo difference

          items_to_read = in_chunk_size * H.N;
          fec_buffer_offset = 0;
          state = Read;

       }

       consume_each(uld);
       return(0);
    }

  } /* namespace DVBparts */
} /* namespace gr */

