/* -*- c++ -*- */
/* 
 * Copyright 2018 Ingo D. Rullhusen <d01c@freeshell.de>.
 * 
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gnuradio/io_signature.h>
#include "ldpc_decoder_soft_impl.h"


using namespace std;


#define FACTOR 2


namespace gr {
  namespace DVBparts {

    ldpc_decoder_soft::sptr
    ldpc_decoder_soft::make()
    {
      return gnuradio::get_initial_sptr
        (new ldpc_decoder_soft_impl());
    }

    /*
     * The private constructor
     */
    ldpc_decoder_soft_impl::ldpc_decoder_soft_impl()
      : gr::block("ldpc_decoder_soft",
              gr::io_signature::make(1,1, sizeof(unsigned char)),
              gr::io_signature::make(1,1, sizeof(unsigned char)))
    {
       forecast_f = 1.0;
       NBCH = 0;
       NLDPC = 0;
       plen = 0;

       bzero(&l1plp,sizeof(l1plp));
       setup_L1Pplp_tag( PLPTagValue, l1plp );

       ldpc = new LDPC<DVB_T2X_TABLE_NULL>();
       aligned_buffer = aligned_alloc(sizeof(simd_type), sizeof(simd_type) * FRAME_SIZE_NORMAL);

       set_tag_propagation_policy(TPP_DONT);

       set_output_multiple( FRAME_SIZE_NORMAL * SIZEOF_SIMD );
    }

    /*
     * Our virtual destructor.
     */
    ldpc_decoder_soft_impl::~ldpc_decoder_soft_impl()
    {
       free(aligned_buffer);
       delete ldpc;
    }


    void ldpc_decoder_soft_impl::update_tables()
    {
       int im = 0;
       int index = 0;
       const int &dbits = NBCH;
       const int &pbits = plen;
       const int &q = q_val;


static int dbits_t = 0;
static int pbits_t = 0;
static int q_t = 0;
if ( dbits_t==dbits && pbits_t==pbits && q_t==q ) return;
dbits_t = dbits;
pbits_t = pbits;
q_t = q;


clog << "UPDATE TABLE" << endl;
       delete ldpc;

       if ( l1plp.plp_fec_type == 1 ) // 64K/normal
       {
          switch ( l1plp.plp_cod )
          {
             case 0: // 1/2
             { ldpc = new LDPC<DVB_S2_TABLE_B4>(); break; }
             case 1: // 3/5
             { ldpc = new LDPC<DVB_S2_TABLE_B5>(); break; }
             case 2: // 2/3
             { ldpc = new LDPC<DVB_T2_TABLE_A3>(); break; }
             case 3: // 3/4
             { ldpc = new LDPC<DVB_S2_TABLE_B7>(); break; }
             case 4: // 4/5
             { ldpc = new LDPC<DVB_S2_TABLE_B8>(); break; }
             case 5: // 5/6
             { ldpc = new LDPC<DVB_S2_TABLE_B9>(); break; }
          }
       }
       else // 16K/medium
       {
          switch ( l1plp.plp_cod )
          {
             case 0: // 1/2
             { ldpc=NULL; cerr<<"TAB fehlt"<<endl;exit(1); break; }
             case 1: // 3/5
             { ldpc = new LDPC<DVB_T2_TABLE_B3>(); break; }
             case 2: // 2/3
             { ldpc=NULL; cerr<<"TAB fehlt"<<endl;exit(1); break; }
             case 3: // 3/4
             { ldpc=NULL; cerr<<"TAB fehlt"<<endl;exit(1); break; }
             case 4: // 4/5
             { ldpc=NULL; cerr<<"TAB fehlt"<<endl;exit(1); break; }
             case 5: // 5/6
             { ldpc=NULL; cerr<<"TAB fehlt"<<endl;exit(1); break; }
          }
       }


///       bp.set_L1Pplp_sigma( l1plp, 0.3 );

       free(aligned_buffer);
       aligned_buffer = aligned_alloc(sizeof(simd_type), sizeof(simd_type) * ldpc->code_len());

       decode.init(ldpc);

clog << "UPDATED" << endl;
    }


    void
    ldpc_decoder_soft_impl::forecast (int noutput_items, gr_vector_int &ninput_items_required)
    { ninput_items_required[0] = ceil( forecast_f * noutput_items ); }

    int
    ldpc_decoder_soft_impl::general_work (int noutput_items,
                       gr_vector_int &ninput_items,
                       gr_vector_const_void_star &input_items,
                       gr_vector_void_star &output_items)
    {
       const pmt::pmt_t PLPfecKey( pmt::intern("PLPfec") );
       const unsigned char *in = (const unsigned char *) input_items[0];
       unsigned char *out = (unsigned char *) output_items[0];

       static enum { Wait, Read, Write } state = Wait;

       const int CODE_LEN = ldpc->code_len();
       const int DATA_LEN = ldpc->data_len();
       simd_type *simd = reinterpret_cast<simd_type *>(aligned_buffer);

       static int q = 0;
       if ( state == Write )
       {
          int nout = noutput_items;
          if ( nout >= SIZEOF_SIMD*NBCH-q ) { nout = SIZEOF_SIMD*NBCH-q; state = Wait; }

          for ( int i = 0; i<nout; ++i )
          {
             int n = (q+i) / NBCH;
////             out[i] = ( fec_buffer[q+(n*plen)+i]>=0 ? 0 : 1 );

             int j = (q+(n*plen)+i) % NLDPC;
             out[i] = ( reinterpret_cast<code_type *>(simd + j)[n]>=0 ? 0 : 1 );

          }
          q += nout;

          consume_each(0);
          return(nout);
       }

       static int items_to_read = 0;
       if ( state == Read )
       {
          int nin = items_to_read;
          if ( nin > ninput_items[0] ) { nin = ninput_items[0]; }

          for ( int i = 0; i<nin; ++i ) /// slow
          {

///          fec_buffer[fec_buffer_offset+i] = (int8_t)((uint8_t)127 - in[i]);

          int n = (fec_buffer_offset+i) / CODE_LEN;
          int j = (fec_buffer_offset+i) % CODE_LEN;
          
          reinterpret_cast<code_type *>(simd + j)[n] = (int8_t)((uint8_t)127 - in[i]);
///       reinterpret_cast<code_type *>(simd + j)[n] = (   (int8_t)((uint8_t)127 - in[i])  >=0 ? 127 : -128 );

          }

          fec_buffer_offset += nin;
          items_to_read -= nin;

          if ( items_to_read < 1 ) {


       int trials = 5;

/**
clog <<"pre1"<<endl;
        for (int n = 0; n < SIZEOF_SIMD; n++) {
          for (int j = 0; j < CODE_LEN; j++) {
            reinterpret_cast<code_type *>(simd + j)[n] = fec_buffer[(n * CODE_LEN) + j];
          }
        }
clog <<"pre2 "<<DATA_LEN << " " << trials << " " << SIZEOF_SIMD <<endl;
**/
        int count = decode(simd, simd + DATA_LEN, trials, SIZEOF_SIMD);

/***
clog <<"pre3 "<<trials<<endl;
        for (int n = 0; n < SIZEOF_SIMD; n++) {
          for (int j = 0; j < CODE_LEN; j++) {
            fec_buffer[(n * CODE_LEN) + j] = reinterpret_cast<code_type *>(simd + j)[n];
          }
        }
***/

        if (count < 0) {
          printf("LDPC decoder failed at converging to a code word.\n");
        }
        else {
          printf("LDPC max trials = %d\n", trials - count);
        }

          q = 0;
          state = Write;

          }

          consume_each(nin);
          return(0);
       }


// handle frame start tag
       uint64_t uld = ninput_items[0];

       std::vector<tag_t> tags;
       std::vector<tag_t>::iterator t;

       get_tags_in_range( tags, 0,
                          nitems_read(0), nitems_read(0) + uld,
                          PLPfecKey );

       if ( tags.size() > 0 )
       {
          tag_t &tagPLP = tags[0];
          int fec_size = pmt::to_long(pmt::tuple_ref(tagPLP.value,0));
          pmt::pmt_t PLPTagValue_t = pmt::tuple_ref(tagPLP.value,1);
/**
clog << "PMT  :" << PLPTagValue << endl;
clog << "PMT_t:" << PLPTagValue_t << endl;
**/

          if ( ! equal( PLPTagValue, PLPTagValue_t ) )
          {
             PLPTagValue = PLPTagValue_t;
             setup_L1Pplp( l1plp, PLPTagValue_t );

             if ( l1plp.plp_fec_type == 1 ) // 64K
             {
                NLDPC = 64800;
                switch ( l1plp.plp_cod )
                {
                   case 0: // 1/2
                   { NBCH = 32400; q_val = 90; break; }
                   case 1: // 3/5
                   { NBCH = 38880; q_val = 72; break; }
                   case 2: // 2/3
                   { NBCH = 43200; q_val = 60; break; }
                   case 3: // 3/4
                   { NBCH = 48600; q_val = 45; break; }
                   case 4: // 4/5
                   { NBCH = 51840; q_val = 36; break; }
                   case 5: // 5/6
                   { NBCH = 54000; q_val = 30; break; }
                }
             }
             else // 16K
             {
                NLDPC = 16200;
                switch ( l1plp.plp_cod )
                {
                   case 0: // 1/2
                   { NBCH = 7200; q_val = 25; break; }
                   case 1: // 3/5
                   { NBCH = 9720; q_val = 18; break; }
                   case 2: // 2/3
                   { NBCH = 10800; q_val = 15; break; }
                   case 3: // 3/4
                   { NBCH = 11880; q_val = 12; break; }
                   case 4: // 4/5
                   { NBCH = 12600; q_val = 10; break; }
                   case 5: // 5/6
                   { NBCH = 13320; q_val = 8; break; }
                }
             }

             plen = NLDPC - NBCH;    //number of parity bits

             forecast_f = NLDPC / (double)NBCH;

             update_tables();
clog << "UPDATE " << NBCH << " " << q_val << endl;
          }

          uint64_t framestart = tagPLP.offset;
          uld = framestart - nitems_read(0); // modulo difference

          uint64_t tpos = nitems_written(0);
          for ( int n = 0; n < SIZEOF_SIMD; n++, tpos+=NBCH )
          {
             add_item_tag( 0, tpos, pmt::intern("PLPfec"),
                              pmt::make_tuple(
                                 pmt::from_long(NBCH),
                                 PLPTagValue
                                             )
                         );
          }


          items_to_read = NLDPC * SIZEOF_SIMD;
          fec_buffer_offset = 0;
          state = Read;
       }
clog << "NLDPC " << NLDPC << " NBCH " << NBCH << " uld " << uld << endl;

       consume_each(uld);
       return(0);
    }

  } /* namespace DVBparts */
} /* namespace gr */

