/* -*- c++ -*- */
/* 
 * Copyright 2018 Ingo D. Rullhusen <d01c@freeshell.de>.
 * 
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gnuradio/io_signature.h>
#include "parity_deinterleaver_impl.h"


namespace gr {
  namespace DVBparts {

    parity_deinterleaver::sptr
    parity_deinterleaver::make()
    {
      return gnuradio::get_initial_sptr
        (new parity_deinterleaver_impl());
    }

    /*
     * The private constructor
     */
    parity_deinterleaver_impl::parity_deinterleaver_impl()
      : gr::block("parity_deinterleaver",
              gr::io_signature::make(1,1, sizeof(unsigned char)),
              gr::io_signature::make(1,1, sizeof(unsigned char)))
    {
       NLDPC = 0;
       QLDPC = 0;

       bzero(&l1plp,sizeof(l1plp));
       setup_L1Pplp_tag( PLPTagValue, l1plp );

///       set_output_multiple(64800);
set_output_multiple( 64800 * 8 );
       set_tag_propagation_policy(TPP_DONT);
    }

    /*
     * Our virtual destructor.
     */
    parity_deinterleaver_impl::~parity_deinterleaver_impl()
    {
    }

    void
    parity_deinterleaver_impl::forecast (int noutput_items, gr_vector_int &ninput_items_required)
    { ninput_items_required[0] = noutput_items; }


    const int TC_0[] = { 0 };

    const int TC_16_64[] = { 0, 0, 2, 4, 4, 5, 7, 7 };
    const int TC_16_16[] = { 0, 0, 0, 1, 7, 20, 20, 21 };

    const int TC_64_64[] = { 0, 0, 2, 2, 3, 4, 4, 5, 5, 7, 8, 9 };
    const int TC_64_16[] = { 0, 0, 0, 2, 2, 2, 3, 3, 3, 6, 7, 7 };

    const int TC_256_64[] = { 0, 2, 2, 2, 2, 3, 7, 15, 16, 20, 22, 22, 27, 27, 28, 32 };
    const int TC_256_16[] = { 0, 0, 0, 1, 7, 20, 20, 21 };

    int
    parity_deinterleaver_impl::general_work (int noutput_items,
                       gr_vector_int &ninput_items,
                       gr_vector_const_void_star &input_items,
                       gr_vector_void_star &output_items)
    {
       const pmt::pmt_t PLPfecKey( pmt::intern("PLPfec") );
       const unsigned char *in = (const unsigned char *) input_items[0];
       unsigned char *out = (unsigned char *) output_items[0];

       static enum { Wait, Read, Write } state = Wait;


       static int q = 0;
       static const int *tc;
       static int Nr, Nc;
       if ( state == Write )
       {
          int nout = noutput_items;
          if ( nout >= NLDPC-q ) { nout = NLDPC-q; state = Wait; }

          for ( int k=0; k<nout; ++k )
          {
             int i = q;
             if ( i >= KLDPC )
             {
                int di = i - KLDPC;
                int t = di % QLDPC;
                int s = di / QLDPC;

                i = KLDPC + 360*t + s;
             }
             int ci = i / Nr;
             int ri = ( i + tc[ci] ) % Nr;

             out[k] = fec_buffer[ ci + Nc*ri ];

             ++q;
          }

          consume_each(0);
          return(nout);
       }

       static int items_to_read = 0;
       if ( state == Read )
       {
          int nin = items_to_read;
          if ( nin > ninput_items[0] ) { nin = ninput_items[0]; }
          memcpy( &fec_buffer[fec_buffer_offset], in, nin*sizeof(unsigned char) );
          fec_buffer_offset += nin;
          items_to_read -= nin;

          if ( items_to_read < 1 ) { q = 0; state = Write; }

          consume_each(nin);
          return(0);
       }


// handle frame start tag
       uint64_t uld = ninput_items[0];

       std::vector<tag_t> tags;
       std::vector<tag_t>::iterator t;

       get_tags_in_range( tags, 0,
                          nitems_read(0), nitems_read(0) + uld,
                          PLPfecKey );

       if ( tags.size() > 0 )
       {
          tag_t &tagPLP = tags[0];
          int fec_size = pmt::to_long(pmt::tuple_ref(tagPLP.value,0));
          pmt::pmt_t PLPTagValue_t = pmt::tuple_ref(tagPLP.value,1);
          if ( ! equal( PLPTagValue, PLPTagValue_t ) )
          {
             PLPTagValue = PLPTagValue_t;
             setup_L1Pplp( l1plp, PLPTagValue_t );

             if ( l1plp.plp_fec_type == 1 ) // 64K
             {
                NLDPC = 64800;

                switch ( l1plp.plp_cod )
                {
                   case 0: // 1/2
                   { KLDPC = 32400; QLDPC = 90;  break; }
                   case 1: // 3/5
                   { KLDPC = 38880; QLDPC = 72; break; }
                   case 2: // 2/3
                   { KLDPC = 43200; QLDPC = 60; break; }
                   case 3: // 3/4
                   { KLDPC = 48600; QLDPC = 45; break; }
                   case 4: // 4/5
                   { KLDPC = 51840; QLDPC = 36; break; }
                   case 5: // 5/6
                   { KLDPC = 54000; QLDPC = 30; break; }
                }

                switch ( l1plp.plp_mod )
                {
                   case 1: { Nr = 8100; Nc = 8; tc = TC_16_64; break; } // 16-QAM
                   case 2: { Nr = 5400; Nc = 12; tc = TC_64_64; break; } // 64-QAM
                   case 3: { Nr = 4050; Nc = 16; tc = TC_256_64; break; } // 256-QAM
                   default:
                   {
                      Nr = NLDPC; Nc = 1; tc = TC_0;
                      KLDPC = NLDPC;
                      break;
                   }
                }

             }
             else // 16K
             {
                NLDPC = 16200;

                switch ( l1plp.plp_cod )
                {
                   case 0: // 1/2
                   { KLDPC = 7200; QLDPC = 25; break; }
                   case 1: // 3/5
                   { KLDPC = 9720; QLDPC = 18; break; }
                   case 2: // 2/3
                   { KLDPC = 10800; QLDPC = 15; break; }
                   case 3: // 3/4
                   { KLDPC = 11880; QLDPC = 12; break; }
                   case 4: // 4/5
                   { KLDPC = 12600; QLDPC = 10; break; }
                   case 5: // 5/6
                   { KLDPC = 13320; QLDPC = 8; break; }
                }

                switch ( l1plp.plp_mod )
                {
                   case 1: { Nr = 2025; Nc = 8; tc = TC_16_16; break; } // 16-QAM
                   case 2: { Nr = 1350; Nc = 12; tc = TC_64_16; break; } // 64-QAM
                   case 3: { Nr = 2025; Nc = 8; tc = TC_256_16; break; } // 256-QAM
                   default:
                   {
                      Nr = NLDPC; Nc = 1; tc = TC_0;
                      KLDPC = NLDPC;
                      break;
                   }
                }

             }

          }

          uint64_t framestart = tagPLP.offset;
          uld = framestart - nitems_read(0); // modulo difference

          uint64_t tpos = nitems_written(0);
          add_item_tag( 0, tpos, PLPfecKey,
                           pmt::make_tuple(
                              pmt::from_long(NLDPC),
                              PLPTagValue
                                          )
                      );

          items_to_read = NLDPC;
          fec_buffer_offset = 0;
          state = Read;
       }

       consume_each(uld);
       return(0);
    }

  } /* namespace DVBparts */
} /* namespace gr */

