/* -*- c++ -*- */
/* 
 * Copyright 2018 Ingo D. Rullhusen <d01c@freeshell.de>.
 * 
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <volk/volk.h>

#include <gnuradio/io_signature.h>
#include <gnuradio/dtv/dvbt2_config.h>
#include "pilot_remove_renorm_impl.h"

#include <iostream>

using namespace std;

using namespace gr::dtv;


namespace gr {
  namespace DVBparts {

    pilot_remove_renorm::sptr
    pilot_remove_renorm::make()
    {
      return gnuradio::get_initial_sptr
        (new pilot_remove_renorm_impl());
    }

    /*
     * The private constructor
     */
   pilot_remove_renorm_impl::pilot_remove_renorm_impl()
      : gr::block("pilot_remove_renorm",
        gr::io_signature::make(1,1, sizeof(gr_complex)),
        gr::io_signature::make(1,1, sizeof(gr_complex)))
   {
      memset(&L1TagValue,0,sizeof(L1TagValue));
      memset(&l1pre,0,sizeof(l1pre));

      for ( int k=0; k<MAX_FFT_SIZE; ++k )
      { c_norm[k] = 0; }

      forecast_f = 1.0;

      init_prbs();

      set_tag_propagation_policy(TPP_DONT);
      set_output_multiple( 32*1024 );
      set_relative_rate(0.5);
      set_fixed_rate(false);
   }


    /*
     * Our virtual destructor.
     */
    pilot_remove_renorm_impl::~pilot_remove_renorm_impl()
    {
    }


   void
   pilot_remove_renorm_impl::init_prbs(void)
   {
      int sr = 0x7ff;
      int j = 0;

      for (int i = 0; i < MAX_CARRIERS; i++) {
        int b = ((sr) ^ (sr >> 2)) & 1;
        prbs[i] = sr & 1;
        sr >>= 1;
        if (b) {
          sr |= 0x400;
        }
      }

      for (int i = 0; i < (MAX_CHIPS / 8); i++) {
        for (int k = 7; k >= 0; k--) {
          pn_sequence[j] = (pn_sequence_table[i] >> k) & 0x1;
          j = j + 1;
        }
      }
   }

   void
   pilot_remove_renorm_impl::init_pilots(unsigned char *data_carrier_map, int symbol)
   {
      const int &pilot_pattern = l1pre.pilot_pattern;
      const int &carrier_mode = l1pre.bwt_ext;
      const int &papr_mode = l1pre.papr;

      int remainder, shift;
      for (int i = 0; i < C_PS; i++) {
        data_carrier_map[i] = DATA_CARRIER;
      }
      switch (dvbt2_fftsize) {
        case FFTSIZE_1K:
          switch (pilot_pattern) {
            case PILOT_PP1:
              for (int i = 0; i < 20; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp1_cp1[i] % 1632) / dx)) % 2 && (((pp1_cp1[i] % 1632) % dx) == 0)) {
                    data_carrier_map[pp1_cp1[i] % 1632] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp1_cp1[i] % 1632] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp1_cp1[i] % 1632] = CONTINUAL_CARRIER;
                }
              }
              break;
            case PILOT_PP2:
              for (int i = 0; i < 20; i++) {
                data_carrier_map[pp2_cp1[i] % 1632] = CONTINUAL_CARRIER;
              }
              break;
            case PILOT_PP3:
              for (int i = 0; i < 22; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp3_cp1[i] % 1632) / dx)) % 2 && (((pp3_cp1[i] % 1632) % dx) == 0)) {
                    data_carrier_map[pp3_cp1[i] % 1632] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp3_cp1[i] % 1632] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp3_cp1[i] % 1632] = CONTINUAL_CARRIER;
                }
              }
              break;
            case PILOT_PP4:
              for (int i = 0; i < 20; i++) {
                data_carrier_map[pp4_cp1[i] % 1632] = CONTINUAL_CARRIER;
              }
              break;
            case PILOT_PP5:
              for (int i = 0; i < 19; i++) {
                data_carrier_map[pp5_cp1[i] % 1632] = CONTINUAL_CARRIER;
              }
              break;
            case PILOT_PP6:
              break;
            case PILOT_PP7:
              for (int i = 0; i < 15; i++) {
                data_carrier_map[pp7_cp1[i] % 1632] = CONTINUAL_CARRIER;
              }
              break;
            case PILOT_PP8:
              break;
          }
          break;
        case FFTSIZE_2K:
          switch (pilot_pattern) {
            case PILOT_PP1:
              for (int i = 0; i < 20; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp1_cp1[i] % 1632) / dx)) % 2 && (((pp1_cp1[i] % 1632) % dx) == 0)) {
                    data_carrier_map[pp1_cp1[i] % 1632] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp1_cp1[i] % 1632] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp1_cp1[i] % 1632] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 25; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp1_cp2[i] % 1632) / dx)) % 2 && (((pp1_cp2[i] % 1632) % dx) == 0)) {
                    data_carrier_map[pp1_cp2[i] % 1632] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp1_cp2[i] % 1632] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp1_cp2[i] % 1632] = CONTINUAL_CARRIER;
                }
              }
              break;
            case PILOT_PP2:
              for (int i = 0; i < 20; i++) {
                data_carrier_map[pp2_cp1[i] % 1632] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 22; i++) {
                data_carrier_map[pp2_cp2[i] % 1632] = CONTINUAL_CARRIER;
              }
              break;
            case PILOT_PP3:
              for (int i = 0; i < 22; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp3_cp1[i] % 1632) / dx)) % 2 && (((pp3_cp1[i] % 1632) % dx) == 0)) {
                    data_carrier_map[pp3_cp1[i] % 1632] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp3_cp1[i] % 1632] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp3_cp1[i] % 1632] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 20; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp3_cp2[i] % 1632) / dx)) % 2 && (((pp3_cp2[i] % 1632) % dx) == 0)) {
                    data_carrier_map[pp3_cp2[i] % 1632] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp3_cp2[i] % 1632] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp3_cp2[i] % 1632] = CONTINUAL_CARRIER;
                }
              }
              break;
            case PILOT_PP4:
              for (int i = 0; i < 20; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp4_cp1[i] % 1632) / dx)) % 2 && (((pp4_cp1[i] % 1632) % dx) == 0)) {
                    data_carrier_map[pp4_cp1[i] % 1632] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp4_cp1[i] % 1632] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp4_cp1[i] % 1632] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 23; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp4_cp2[i] % 1632) / dx)) % 2 && (((pp4_cp2[i] % 1632) % dx) == 0)) {
                    data_carrier_map[pp4_cp2[i] % 1632] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp4_cp2[i] % 1632] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp4_cp2[i] % 1632] = CONTINUAL_CARRIER;
                }
              }
              break;
            case PILOT_PP5:
              for (int i = 0; i < 19; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp5_cp1[i] % 1632) / dx)) % 2 && (((pp5_cp1[i] % 1632) % dx) == 0)) {
                    data_carrier_map[pp5_cp1[i] % 1632] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp5_cp1[i] % 1632] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp5_cp1[i] % 1632] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 23; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp5_cp2[i] % 1632) / dx)) % 2 && (((pp5_cp2[i] % 1632) % dx) == 0)) {
                    data_carrier_map[pp5_cp2[i] % 1632] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp5_cp2[i] % 1632] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp5_cp2[i] % 1632] = CONTINUAL_CARRIER;
                }
              }
              break;
            case PILOT_PP6:
              break;
            case PILOT_PP7:
              for (int i = 0; i < 15; i++) {
                data_carrier_map[pp7_cp1[i] % 1632] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 30; i++) {
                data_carrier_map[pp7_cp2[i] % 1632] = CONTINUAL_CARRIER;
              }
              break;
            case PILOT_PP8:
              break;
          }
          break;
        case FFTSIZE_4K:
          switch (pilot_pattern) {
            case PILOT_PP1:
              for (int i = 0; i < 20; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp1_cp1[i] % 3264) / dx)) % 2 && (((pp1_cp1[i] % 3264) % dx) == 0)) {
                    data_carrier_map[pp1_cp1[i] % 3264] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp1_cp1[i] % 3264] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp1_cp1[i] % 3264] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 25; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp1_cp2[i] % 3264) / dx)) % 2 && (((pp1_cp2[i] % 3264) % dx) == 0)) {
                    data_carrier_map[pp1_cp2[i] % 3264] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp1_cp2[i] % 3264] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp1_cp2[i] % 3264] = CONTINUAL_CARRIER;
                }
              }
              break;
            case PILOT_PP2:
              for (int i = 0; i < 20; i++) {
                data_carrier_map[pp2_cp1[i] % 3264] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 22; i++) {
                data_carrier_map[pp2_cp2[i] % 3264] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 2; i++) {
                data_carrier_map[pp2_cp3[i] % 3264] = CONTINUAL_CARRIER;
              }
              break;
            case PILOT_PP3:
              for (int i = 0; i < 22; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp3_cp1[i] % 3264) / dx)) % 2 && (((pp3_cp1[i] % 3264) % dx) == 0)) {
                    data_carrier_map[pp3_cp1[i] % 3264] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp3_cp1[i] % 3264] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp3_cp1[i] % 3264] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 20; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp3_cp2[i] % 3264) / dx)) % 2 && (((pp3_cp2[i] % 3264) % dx) == 0)) {
                    data_carrier_map[pp3_cp2[i] % 3264] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp3_cp2[i] % 3264] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp3_cp2[i] % 3264] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 1; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp3_cp3[i] % 3264) / dx)) % 2 && (((pp3_cp3[i] % 3264) % dx) == 0)) {
                    data_carrier_map[pp3_cp3[i] % 3264] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp3_cp3[i] % 3264] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp3_cp3[i] % 3264] = CONTINUAL_CARRIER;
                }
              }
              break;
            case PILOT_PP4:
              for (int i = 0; i < 20; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp4_cp1[i] % 3264) / dx)) % 2 && (((pp4_cp1[i] % 3264) % dx) == 0)) {
                    data_carrier_map[pp4_cp1[i] % 3264] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp4_cp1[i] % 3264] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp4_cp1[i] % 3264] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 23; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp4_cp2[i] % 3264) / dx)) % 2 && (((pp4_cp2[i] % 3264) % dx) == 0)) {
                    data_carrier_map[pp4_cp2[i] % 3264] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp4_cp2[i] % 3264] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp4_cp2[i] % 3264] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 1; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp4_cp3[i] % 3264) / dx)) % 2 && (((pp4_cp3[i] % 3264) % dx) == 0)) {
                    data_carrier_map[pp4_cp3[i] % 3264] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp4_cp3[i] % 3264] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp4_cp3[i] % 3264] = CONTINUAL_CARRIER;
                }
              }
              break;
            case PILOT_PP5:
              for (int i = 0; i < 19; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp5_cp1[i] % 3264) / dx)) % 2 && (((pp5_cp1[i] % 3264) % dx) == 0)) {
                    data_carrier_map[pp5_cp1[i] % 3264] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp5_cp1[i] % 3264] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp5_cp1[i] % 3264] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 23; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp5_cp2[i] % 3264) / dx)) % 2 && (((pp5_cp2[i] % 3264) % dx) == 0)) {
                    data_carrier_map[pp5_cp2[i] % 3264] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp5_cp2[i] % 3264] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp5_cp2[i] % 3264] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 3; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp5_cp3[i] % 3264) / dx)) % 2 && (((pp5_cp3[i] % 3264) % dx) == 0)) {
                    data_carrier_map[pp5_cp3[i] % 3264] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp5_cp3[i] % 3264] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp5_cp3[i] % 3264] = CONTINUAL_CARRIER;
                }
              }
              break;
            case PILOT_PP6:
              break;
            case PILOT_PP7:
              for (int i = 0; i < 15; i++) {
                data_carrier_map[pp7_cp1[i] % 3264] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 30; i++) {
                data_carrier_map[pp7_cp2[i] % 3264] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 5; i++) {
                data_carrier_map[pp7_cp3[i] % 3264] = CONTINUAL_CARRIER;
              }
              break;
            case PILOT_PP8:
              break;
          }
          break;
        case FFTSIZE_8K:
        case FFTSIZE_8K_T2GI:
          switch (pilot_pattern) {
            case PILOT_PP1:
              for (int i = 0; i < 20; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp1_cp1[i] % 6528) / dx)) % 2 && (((pp1_cp1[i] % 6528) % dx) == 0)) {
                    data_carrier_map[pp1_cp1[i] % 6528] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp1_cp1[i] % 6528] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp1_cp1[i] % 6528] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 25; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp1_cp2[i] % 6528) / dx)) % 2 && (((pp1_cp2[i] % 6528) % dx) == 0)) {
                    data_carrier_map[pp1_cp2[i] % 6528] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp1_cp2[i] % 6528] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp1_cp2[i] % 6528] = CONTINUAL_CARRIER;
                }
              }
              break;
            case PILOT_PP2:
              for (int i = 0; i < 20; i++) {
                data_carrier_map[pp2_cp1[i] % 6528] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 22; i++) {
                data_carrier_map[pp2_cp2[i] % 6528] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 2; i++) {
                data_carrier_map[pp2_cp3[i] % 6528] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 2; i++) {
                data_carrier_map[pp2_cp4[i] % 6528] = CONTINUAL_CARRIER;
              }
              if (carrier_mode == CARRIERS_EXTENDED) {
                for (int i = 0; i < 4; i++) {
                  data_carrier_map[pp2_8k[i]] = CONTINUAL_CARRIER;
                }
              }
              break;
            case PILOT_PP3:
              for (int i = 0; i < 22; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp3_cp1[i] % 6528) / dx)) % 2 && (((pp3_cp1[i] % 6528) % dx) == 0)) {
                    data_carrier_map[pp3_cp1[i] % 6528] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp3_cp1[i] % 6528] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp3_cp1[i] % 6528] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 20; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp3_cp2[i] % 6528) / dx)) % 2 && (((pp3_cp2[i] % 6528) % dx) == 0)) {
                    data_carrier_map[pp3_cp2[i] % 6528] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp3_cp2[i] % 6528] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp3_cp2[i] % 6528] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 1; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp3_cp3[i] % 6528) / dx)) % 2 && (((pp3_cp3[i] % 6528) % dx) == 0)) {
                    data_carrier_map[pp3_cp3[i] % 6528] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp3_cp3[i] % 6528] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp3_cp3[i] % 6528] = CONTINUAL_CARRIER;
                }
              }
              if (carrier_mode == CARRIERS_EXTENDED) {
                for (int i = 0; i < 2; i++) {
                  if (miso == true && miso_group == MISO_TX2) {
                    if (((pp3_8k[i] / dx)) % 2 && ((pp3_8k[i] % dx) == 0)) {
                      data_carrier_map[pp3_8k[i]] = CONTINUAL_CARRIER_INVERTED;
                    }
                    else {
                      data_carrier_map[pp3_8k[i]] = CONTINUAL_CARRIER;
                    }
                  }
                  else {
                    data_carrier_map[pp3_8k[i]] = CONTINUAL_CARRIER;
                  }
                }
              }
              break;
            case PILOT_PP4:
              for (int i = 0; i < 20; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp4_cp1[i] % 6528) / dx)) % 2 && (((pp4_cp1[i] % 6528) % dx) == 0)) {
                    data_carrier_map[pp4_cp1[i] % 6528] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp4_cp1[i] % 6528] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp4_cp1[i] % 6528] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 23; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp4_cp2[i] % 6528) / dx)) % 2 && (((pp4_cp2[i] % 6528) % dx) == 0)) {
                    data_carrier_map[pp4_cp2[i] % 6528] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp4_cp2[i] % 6528] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp4_cp2[i] % 6528] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 1; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp4_cp3[i] % 6528) / dx)) % 2 && (((pp4_cp3[i] % 6528) % dx) == 0)) {
                    data_carrier_map[pp4_cp3[i] % 6528] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp4_cp3[i] % 6528] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp4_cp3[i] % 6528] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 2; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp4_cp4[i] % 6528) / dx)) % 2 && (((pp4_cp4[i] % 6528) % dx) == 0)) {
                    data_carrier_map[pp4_cp4[i] % 6528] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp4_cp4[i] % 6528] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp4_cp4[i] % 6528] = CONTINUAL_CARRIER;
                }
              }
              if (carrier_mode == CARRIERS_EXTENDED) {
                for (int i = 0; i < 2; i++) {
                  if (miso == true && miso_group == MISO_TX2) {
                    if (((pp4_8k[i] / dx)) % 2 && ((pp4_8k[i] % dx) == 0)) {
                      data_carrier_map[pp4_8k[i]] = CONTINUAL_CARRIER_INVERTED;
                    }
                    else {
                      data_carrier_map[pp4_8k[i]] = CONTINUAL_CARRIER;
                    }
                  }
                  else {
                    data_carrier_map[pp4_8k[i]] = CONTINUAL_CARRIER;
                  }
                }
              }
              break;
            case PILOT_PP5:
              for (int i = 0; i < 19; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp5_cp1[i] % 6528) / dx)) % 2 && (((pp5_cp1[i] % 6528) % dx) == 0)) {
                    data_carrier_map[pp5_cp1[i] % 6528] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp5_cp1[i] % 6528] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp5_cp1[i] % 6528] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 23; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp5_cp2[i] % 6528) / dx)) % 2 && (((pp5_cp2[i] % 6528) % dx) == 0)) {
                    data_carrier_map[pp5_cp2[i] % 6528] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp5_cp2[i] % 6528] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp5_cp2[i] % 6528] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 3; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp5_cp3[i] % 6528) / dx)) % 2 && (((pp5_cp3[i] % 6528) % dx) == 0)) {
                    data_carrier_map[pp5_cp3[i] % 6528] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp5_cp3[i] % 6528] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp5_cp3[i] % 6528] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 1; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp5_cp4[i] % 6528) / dx)) % 2 && (((pp5_cp4[i] % 6528) % dx) == 0)) {
                    data_carrier_map[pp5_cp4[i] % 6528] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp5_cp4[i] % 6528] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp5_cp4[i] % 6528] = CONTINUAL_CARRIER;
                }
              }
              break;
            case PILOT_PP6:
              break;
            case PILOT_PP7:
              for (int i = 0; i < 15; i++) {
                data_carrier_map[pp7_cp1[i] % 6528] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 30; i++) {
                data_carrier_map[pp7_cp2[i] % 6528] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 5; i++) {
                data_carrier_map[pp7_cp3[i] % 6528] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 3; i++) {
                data_carrier_map[pp7_cp4[i] % 6528] = CONTINUAL_CARRIER;
              }
              if (carrier_mode == CARRIERS_EXTENDED) {
                for (int i = 0; i < 5; i++) {
                  data_carrier_map[pp7_8k[i]] = CONTINUAL_CARRIER;
                }
              }
              break;
            case PILOT_PP8:
              for (int i = 0; i < 47; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp8_cp4[i] % 6528) / dx)) % 2 && (((pp8_cp4[i] % 6528) % dx) == 0)) {
                    data_carrier_map[pp8_cp4[i] % 6528] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp8_cp4[i] % 6528] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp8_cp4[i] % 6528] = CONTINUAL_CARRIER;
                }
              }
              if (carrier_mode == CARRIERS_EXTENDED) {
                for (int i = 0; i < 5; i++) {
                  if (miso == true && miso_group == MISO_TX2) {
                    if (((pp8_8k[i] / dx)) % 2 && ((pp8_8k[i] % dx) == 0)) {
                      data_carrier_map[pp8_8k[i]] = CONTINUAL_CARRIER_INVERTED;
                    }
                    else {
                      data_carrier_map[pp8_8k[i]] = CONTINUAL_CARRIER;
                    }
                  }
                  else {
                    data_carrier_map[pp8_8k[i]] = CONTINUAL_CARRIER;
                  }
                }
              }
              break;
          }
          break;
        case FFTSIZE_16K:
        case FFTSIZE_16K_T2GI:
          switch (pilot_pattern) {
            case PILOT_PP1:
              for (int i = 0; i < 20; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp1_cp1[i] % 13056) / dx)) % 2 && (((pp1_cp1[i] % 13056) % dx) == 0)) {
                    data_carrier_map[pp1_cp1[i] % 13056] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp1_cp1[i] % 13056] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp1_cp1[i] % 13056] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 25; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp1_cp2[i] % 13056) / dx)) % 2 && (((pp1_cp2[i] % 13056) % dx) == 0)) {
                    data_carrier_map[pp1_cp2[i] % 13056] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp1_cp2[i] % 13056] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp1_cp2[i] % 13056] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 44; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp1_cp5[i] % 13056) / dx)) % 2 && (((pp1_cp5[i] % 13056) % dx) == 0)) {
                    data_carrier_map[pp1_cp5[i] % 13056] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp1_cp5[i] % 13056] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp1_cp5[i] % 13056] = CONTINUAL_CARRIER;
                }
              }
              if (carrier_mode == CARRIERS_EXTENDED) {
                for (int i = 0; i < 4; i++) {
                  if (miso == true && miso_group == MISO_TX2) {
                    if (((pp1_16k[i] / dx)) % 2 && ((pp1_16k[i] % dx) == 0)) {
                      data_carrier_map[pp1_16k[i]] = CONTINUAL_CARRIER_INVERTED;
                    }
                    else {
                      data_carrier_map[pp1_16k[i]] = CONTINUAL_CARRIER;
                    }
                  }
                  else {
                    data_carrier_map[pp1_16k[i]] = CONTINUAL_CARRIER;
                  }
                }
              }
              break;
            case PILOT_PP2:
              for (int i = 0; i < 20; i++) {
                data_carrier_map[pp2_cp1[i] % 13056] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 22; i++) {
                data_carrier_map[pp2_cp2[i] % 13056] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 2; i++) {
                data_carrier_map[pp2_cp3[i] % 13056] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 2; i++) {
                data_carrier_map[pp2_cp4[i] % 13056] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 41; i++) {
                data_carrier_map[pp2_cp5[i] % 13056] = CONTINUAL_CARRIER;
              }
              if (carrier_mode == CARRIERS_EXTENDED) {
                for (int i = 0; i < 2; i++) {
                  data_carrier_map[pp2_16k[i]] = CONTINUAL_CARRIER;
                }
              }
              break;
            case PILOT_PP3:
              for (int i = 0; i < 22; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp3_cp1[i] % 13056) / dx)) % 2 && (((pp3_cp1[i] % 13056) % dx) == 0)) {
                    data_carrier_map[pp3_cp1[i] % 13056] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp3_cp1[i] % 13056] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp3_cp1[i] % 13056] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 20; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp3_cp2[i] % 13056) / dx)) % 2 && (((pp3_cp2[i] % 13056) % dx) == 0)) {
                    data_carrier_map[pp3_cp2[i] % 13056] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp3_cp2[i] % 13056] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp3_cp2[i] % 13056] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 1; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp3_cp3[i] % 13056) / dx)) % 2 && (((pp3_cp3[i] % 13056) % dx) == 0)) {
                    data_carrier_map[pp3_cp3[i] % 13056] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp3_cp3[i] % 13056] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp3_cp3[i] % 13056] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 44; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp3_cp5[i] % 13056) / dx)) % 2 && (((pp3_cp5[i] % 13056) % dx) == 0)) {
                    data_carrier_map[pp3_cp5[i] % 13056] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp3_cp5[i] % 13056] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp3_cp5[i] % 13056] = CONTINUAL_CARRIER;
                }
              }
              if (carrier_mode == CARRIERS_EXTENDED) {
                for (int i = 0; i < 2; i++) {
                  if (miso == true && miso_group == MISO_TX2) {
                    if (((pp3_16k[i] / dx)) % 2 && ((pp3_16k[i] % dx) == 0)) {
                      data_carrier_map[pp3_16k[i]] = CONTINUAL_CARRIER_INVERTED;
                    }
                    else {
                      data_carrier_map[pp3_16k[i]] = CONTINUAL_CARRIER;
                    }
                  }
                  else {
                    data_carrier_map[pp3_16k[i]] = CONTINUAL_CARRIER;
                  }
                }
              }
              break;
            case PILOT_PP4:
              for (int i = 0; i < 20; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp4_cp1[i] % 13056) / dx)) % 2 && (((pp4_cp1[i] % 13056) % dx) == 0)) {
                    data_carrier_map[pp4_cp1[i] % 13056] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp4_cp1[i] % 13056] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp4_cp1[i] % 13056] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 23; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp4_cp2[i] % 13056) / dx)) % 2 && (((pp4_cp2[i] % 13056) % dx) == 0)) {
                    data_carrier_map[pp4_cp2[i] % 13056] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp4_cp2[i] % 13056] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp4_cp2[i] % 13056] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 1; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp4_cp3[i] % 13056) / dx)) % 2 && (((pp4_cp3[i] % 13056) % dx) == 0)) {
                    data_carrier_map[pp4_cp3[i] % 13056] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp4_cp3[i] % 13056] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp4_cp3[i] % 13056] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 2; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp4_cp4[i] % 13056) / dx)) % 2 && (((pp4_cp4[i] % 13056) % dx) == 0)) {
                    data_carrier_map[pp4_cp4[i] % 13056] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp4_cp4[i] % 13056] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp4_cp4[i] % 13056] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 44; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp4_cp5[i] % 13056) / dx)) % 2 && (((pp4_cp5[i] % 13056) % dx) == 0)) {
                    data_carrier_map[pp4_cp5[i] % 13056] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp4_cp5[i] % 13056] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp4_cp5[i] % 13056] = CONTINUAL_CARRIER;
                }
              }
              if (carrier_mode == CARRIERS_EXTENDED) {
                for (int i = 0; i < 2; i++) {
                  if (miso == true && miso_group == MISO_TX2) {
                    if (((pp4_16k[i] / dx)) % 2 && ((pp4_16k[i] % dx) == 0)) {
                      data_carrier_map[pp4_16k[i]] = CONTINUAL_CARRIER_INVERTED;
                    }
                    else {
                      data_carrier_map[pp4_16k[i]] = CONTINUAL_CARRIER;
                    }
                  }
                  else {
                    data_carrier_map[pp4_16k[i]] = CONTINUAL_CARRIER;
                  }
                }
              }
              break;
            case PILOT_PP5:
              for (int i = 0; i < 19; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp5_cp1[i] % 13056) / dx)) % 2 && (((pp5_cp1[i] % 13056) % dx) == 0)) {
                    data_carrier_map[pp5_cp1[i] % 13056] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp5_cp1[i] % 13056] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp5_cp1[i] % 13056] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 23; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp5_cp2[i] % 13056) / dx)) % 2 && (((pp5_cp2[i] % 13056) % dx) == 0)) {
                    data_carrier_map[pp5_cp2[i] % 13056] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp5_cp2[i] % 13056] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp5_cp2[i] % 13056] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 3; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp5_cp3[i] % 13056) / dx)) % 2 && (((pp5_cp3[i] % 13056) % dx) == 0)) {
                    data_carrier_map[pp5_cp3[i] % 13056] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp5_cp3[i] % 13056] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp5_cp3[i] % 13056] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 1; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp5_cp4[i] % 13056) / dx)) % 2 && (((pp5_cp4[i] % 13056) % dx) == 0)) {
                    data_carrier_map[pp5_cp4[i] % 13056] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp5_cp4[i] % 13056] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp5_cp4[i] % 13056] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 44; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp5_cp5[i] % 13056) / dx)) % 2 && (((pp5_cp5[i] % 13056) % dx) == 0)) {
                    data_carrier_map[pp5_cp5[i] % 13056] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp5_cp5[i] % 13056] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp5_cp5[i] % 13056] = CONTINUAL_CARRIER;
                }
              }
              if (carrier_mode == CARRIERS_EXTENDED) {
                for (int i = 0; i < 2; i++) {
                  if (miso == true && miso_group == MISO_TX2) {
                    if (((pp5_16k[i] / dx)) % 2 && ((pp5_16k[i] % dx) == 0)) {
                      data_carrier_map[pp5_16k[i]] = CONTINUAL_CARRIER_INVERTED;
                    }
                    else {
                      data_carrier_map[pp5_16k[i]] = CONTINUAL_CARRIER;
                    }
                  }
                  else {
                    data_carrier_map[pp5_16k[i]] = CONTINUAL_CARRIER;
                  }
                }
              }
              break;
            case PILOT_PP6:
              for (int i = 0; i < 88; i++) {
                data_carrier_map[pp6_cp5[i] % 13056] = CONTINUAL_CARRIER;
              }
              if (carrier_mode == CARRIERS_EXTENDED) {
                for (int i = 0; i < 2; i++) {
                  data_carrier_map[pp6_16k[i]] = CONTINUAL_CARRIER;
                }
              }
              break;
            case PILOT_PP7:
              for (int i = 0; i < 15; i++) {
                data_carrier_map[pp7_cp1[i] % 13056] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 30; i++) {
                data_carrier_map[pp7_cp2[i] % 13056] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 5; i++) {
                data_carrier_map[pp7_cp3[i] % 13056] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 3; i++) {
                data_carrier_map[pp7_cp4[i] % 13056] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 35; i++) {
                data_carrier_map[pp7_cp5[i] % 13056] = CONTINUAL_CARRIER;
              }
              if (carrier_mode == CARRIERS_EXTENDED) {
                for (int i = 0; i < 3; i++) {
                  data_carrier_map[pp7_16k[i]] = CONTINUAL_CARRIER;
                }
              }
              break;
            case PILOT_PP8:
              for (int i = 0; i < 47; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp8_cp4[i] % 13056) / dx)) % 2 && (((pp8_cp4[i] % 13056) % dx) == 0)) {
                    data_carrier_map[pp8_cp4[i] % 13056] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp8_cp4[i] % 13056] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp8_cp4[i] % 13056] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 39; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if ((((pp8_cp5[i] % 13056) / dx)) % 2 && (((pp8_cp5[i] % 13056) % dx) == 0)) {
                    data_carrier_map[pp8_cp5[i] % 13056] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp8_cp5[i] % 13056] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp8_cp5[i] % 13056] = CONTINUAL_CARRIER;
                }
              }
              if (carrier_mode == CARRIERS_EXTENDED) {
                for (int i = 0; i < 3; i++) {
                  if (miso == true && miso_group == MISO_TX2) {
                    if (((pp8_16k[i] / dx)) % 2 && ((pp8_16k[i] % dx) == 0)) {
                      data_carrier_map[pp8_16k[i]] = CONTINUAL_CARRIER_INVERTED;
                    }
                    else {
                      data_carrier_map[pp8_16k[i]] = CONTINUAL_CARRIER;
                    }
                  }
                  else {
                    data_carrier_map[pp8_16k[i]] = CONTINUAL_CARRIER;
                  }
                }
              }
              break;
          }
          break;
        case FFTSIZE_32K:
        case FFTSIZE_32K_T2GI:
          switch (pilot_pattern) {
            case PILOT_PP1:
              for (int i = 0; i < 20; i++) {
                data_carrier_map[pp1_cp1[i]] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 25; i++) {
                data_carrier_map[pp1_cp2[i]] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 44; i++) {
                data_carrier_map[pp1_cp5[i]] = CONTINUAL_CARRIER;
              }
              break;
            case PILOT_PP2:
              for (int i = 0; i < 20; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if (((pp2_cp1[i] / dx)) % 2 && ((pp2_cp1[i] % dx) == 0)) {
                    data_carrier_map[pp2_cp1[i]] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp2_cp1[i]] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp2_cp1[i]] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 22; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if (((pp2_cp2[i] / dx)) % 2 && ((pp2_cp2[i] % dx) == 0)) {
                    data_carrier_map[pp2_cp2[i]] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp2_cp2[i]] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp2_cp2[i]] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 2; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if (((pp2_cp3[i] / dx)) % 2 && ((pp2_cp3[i] % dx) == 0)) {
                    data_carrier_map[pp2_cp3[i]] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp2_cp3[i]] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp2_cp3[i]] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 2; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if (((pp2_cp4[i] / dx)) % 2 && ((pp2_cp4[i] % dx) == 0)) {
                    data_carrier_map[pp2_cp4[i]] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp2_cp4[i]] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp2_cp4[i]] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 41; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if (((pp2_cp5[i] / dx)) % 2 && ((pp2_cp5[i] % dx) == 0)) {
                    data_carrier_map[pp2_cp5[i]] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp2_cp5[i]] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp2_cp5[i]] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 88; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if (((pp2_cp6[i] / dx)) % 2 && ((pp2_cp6[i] % dx) == 0)) {
                    data_carrier_map[pp2_cp6[i]] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp2_cp6[i]] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp2_cp6[i]] = CONTINUAL_CARRIER;
                }
              }
              if (carrier_mode == CARRIERS_EXTENDED) {
                for (int i = 0; i < 2; i++) {
                  if (miso == true && miso_group == MISO_TX2) {
                    if (((pp2_32k[i] / dx)) % 2 && ((pp2_32k[i] % dx) == 0)) {
                      data_carrier_map[pp2_32k[i]] = CONTINUAL_CARRIER_INVERTED;
                    }
                    else {
                      data_carrier_map[pp2_32k[i]] = CONTINUAL_CARRIER;
                    }
                  }
                  else {
                    data_carrier_map[pp2_32k[i]] = CONTINUAL_CARRIER;
                  }
                }
              }
              break;
            case PILOT_PP3:
              for (int i = 0; i < 22; i++) {
                data_carrier_map[pp3_cp1[i]] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 20; i++) {
                data_carrier_map[pp3_cp2[i]] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 1; i++) {
                data_carrier_map[pp3_cp3[i]] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 44; i++) {
                data_carrier_map[pp3_cp5[i]] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 49; i++) {
                data_carrier_map[pp3_cp6[i]] = CONTINUAL_CARRIER;
              }
              break;
            case PILOT_PP4:
              for (int i = 0; i < 20; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if (((pp4_cp1[i] / dx)) % 2 && ((pp4_cp1[i] % dx) == 0)) {
                    data_carrier_map[pp4_cp1[i]] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp4_cp1[i]] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp4_cp1[i]] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 23; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if (((pp4_cp2[i] / dx)) % 2 && ((pp4_cp2[i] % dx) == 0)) {
                    data_carrier_map[pp4_cp2[i]] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp4_cp2[i]] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp4_cp2[i]] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 1; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if (((pp4_cp3[i] / dx)) % 2 && ((pp4_cp3[i] % dx) == 0)) {
                    data_carrier_map[pp4_cp3[i]] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp4_cp3[i]] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp4_cp3[i]] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 2; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if (((pp4_cp4[i] / dx)) % 2 && ((pp4_cp4[i] % dx) == 0)) {
                    data_carrier_map[pp4_cp4[i]] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp4_cp4[i]] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp4_cp4[i]] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 44; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if (((pp4_cp5[i] / dx)) % 2 && ((pp4_cp5[i] % dx) == 0)) {
                    data_carrier_map[pp4_cp5[i]] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp4_cp5[i]] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp4_cp5[i]] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 86; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if (((pp4_cp6[i] / dx)) % 2 && ((pp4_cp6[i] % dx) == 0)) {
                    data_carrier_map[pp4_cp6[i]] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp4_cp6[i]] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp4_cp6[i]] = CONTINUAL_CARRIER;
                }
              }
              if (carrier_mode == CARRIERS_EXTENDED) {
                for (int i = 0; i < 2; i++) {
                  if (miso == true && miso_group == MISO_TX2) {
                    if (((pp4_32k[i] / dx)) % 2 && ((pp4_32k[i] % dx) == 0)) {
                      data_carrier_map[pp4_32k[i]] = CONTINUAL_CARRIER_INVERTED;
                    }
                    else {
                      data_carrier_map[pp4_32k[i]] = CONTINUAL_CARRIER;
                    }
                  }
                  else {
                    data_carrier_map[pp4_32k[i]] = CONTINUAL_CARRIER;
                  }
                }
              }
              break;
            case PILOT_PP5:
              for (int i = 0; i < 19; i++) {
                data_carrier_map[pp5_cp1[i]] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 23; i++) {
                data_carrier_map[pp5_cp2[i]] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 3; i++) {
                data_carrier_map[pp5_cp3[i]] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 1; i++) {
                data_carrier_map[pp5_cp4[i]] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 44; i++) {
                data_carrier_map[pp5_cp5[i]] = CONTINUAL_CARRIER;
              }
              break;
            case PILOT_PP6:
              for (int i = 0; i < 88; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if (((pp6_cp5[i] / dx)) % 2 && ((pp6_cp5[i] % dx) == 0)) {
                    data_carrier_map[pp6_cp5[i]] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp6_cp5[i]] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp6_cp5[i]] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 88; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if (((pp6_cp6[i] / dx)) % 2 && ((pp6_cp6[i] % dx) == 0)) {
                    data_carrier_map[pp6_cp6[i]] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp6_cp6[i]] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp6_cp6[i]] = CONTINUAL_CARRIER;
                }
              }
              if (carrier_mode == CARRIERS_EXTENDED) {
                for (int i = 0; i < 4; i++) {
                  if (miso == true && miso_group == MISO_TX2) {
                    if (((pp6_32k[i] / dx)) % 2 && ((pp6_32k[i] % dx) == 0)) {
                      data_carrier_map[pp6_32k[i]] = CONTINUAL_CARRIER_INVERTED;
                    }
                    else {
                      data_carrier_map[pp6_32k[i]] = CONTINUAL_CARRIER;
                    }
                  }
                  else {
                    data_carrier_map[pp6_32k[i]] = CONTINUAL_CARRIER;
                  }
                }
              }
              break;
            case PILOT_PP7:
              for (int i = 0; i < 15; i++) {
                data_carrier_map[pp7_cp1[i]] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 30; i++) {
                data_carrier_map[pp7_cp2[i]] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 5; i++) {
                data_carrier_map[pp7_cp3[i]] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 3; i++) {
                data_carrier_map[pp7_cp4[i]] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 35; i++) {
                data_carrier_map[pp7_cp5[i]] = CONTINUAL_CARRIER;
              }
              for (int i = 0; i < 92; i++) {
                data_carrier_map[pp7_cp6[i]] = CONTINUAL_CARRIER;
              }
              if (carrier_mode == CARRIERS_EXTENDED) {
                for (int i = 0; i < 2; i++) {
                  data_carrier_map[pp7_32k[i]] = CONTINUAL_CARRIER;
                }
              }
              break;
            case PILOT_PP8:
              for (int i = 0; i < 47; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if (((pp8_cp4[i] / dx)) % 2 && ((pp8_cp4[i] % dx) == 0)) {
                    data_carrier_map[pp8_cp4[i]] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp8_cp4[i]] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp8_cp4[i]] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 39; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if (((pp8_cp5[i] / dx)) % 2 && ((pp8_cp5[i] % dx) == 0)) {
                    data_carrier_map[pp8_cp5[i]] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp8_cp5[i]] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp8_cp5[i]] = CONTINUAL_CARRIER;
                }
              }
              for (int i = 0; i < 89; i++) {
                if (miso == true && miso_group == MISO_TX2) {
                  if (((pp8_cp6[i] / dx)) % 2 && ((pp8_cp6[i] % dx) == 0)) {
                    data_carrier_map[pp8_cp6[i]] = CONTINUAL_CARRIER_INVERTED;
                  }
                  else {
                    data_carrier_map[pp8_cp6[i]] = CONTINUAL_CARRIER;
                  }
                }
                else {
                  data_carrier_map[pp8_cp6[i]] = CONTINUAL_CARRIER;
                }
              }
              if (carrier_mode == CARRIERS_EXTENDED) {
                for (int i = 0; i < 6; i++) {
                  if (miso == true && miso_group == MISO_TX2) {
                    if (((pp8_32k[i] / dx)) % 2 && ((pp8_32k[i] % dx) == 0)) {
                      data_carrier_map[pp8_32k[i]] = CONTINUAL_CARRIER_INVERTED;
                    }
                    else {
                      data_carrier_map[pp8_32k[i]] = CONTINUAL_CARRIER;
                    }
                  }
                  else {
                    data_carrier_map[pp8_32k[i]] = CONTINUAL_CARRIER;
                  }
                }
              }
              break;
          }
          break;
      }
      for (int i = 0; i < C_PS; i++) {
        remainder = (i - K_EXT) % (dx * dy);
        if (remainder < 0) {
          remainder += (dx * dy);
        }
        if (remainder == (dx * (symbol % dy))) {
          if (miso == true && miso_group == MISO_TX2) {
            if ((i / dx) % 2) {
              data_carrier_map[i] = SCATTERED_CARRIER_INVERTED;
            }
            else {
              data_carrier_map[i] = SCATTERED_CARRIER;
            }
          }
          else {
            data_carrier_map[i] = SCATTERED_CARRIER;
          }
        }
      }
      if (miso == true && miso_group == MISO_TX2) {
        if (symbol % 2) {
          data_carrier_map[0] = SCATTERED_CARRIER_INVERTED;
          data_carrier_map[C_PS - 1] = SCATTERED_CARRIER_INVERTED;
        }
        else {
          data_carrier_map[0] = SCATTERED_CARRIER;
          data_carrier_map[C_PS - 1] = SCATTERED_CARRIER;
        }
      }
      else {
        data_carrier_map[0] = SCATTERED_CARRIER;
        data_carrier_map[C_PS - 1] = SCATTERED_CARRIER;
      }
      if (papr_mode == PAPR_TR || papr_mode == PAPR_BOTH) {
        if (carrier_mode == CARRIERS_NORMAL) {
          shift = dx * (symbol % dy);
        }
        else {
          shift = dx * ((symbol + (K_EXT / dx)) % dy);
        }
        switch (dvbt2_fftsize) {
          case FFTSIZE_1K:
            for (int i = 0; i < 10; i++) {
              data_carrier_map[tr_papr_map_1k[i] + shift] = TRPAPR_CARRIER;
            }
            break;
          case FFTSIZE_2K:
            for (int i = 0; i < 18; i++) {
              data_carrier_map[tr_papr_map_2k[i] + shift] = TRPAPR_CARRIER;
            }
            break;
          case FFTSIZE_4K:
            for (int i = 0; i < 36; i++) {
              data_carrier_map[tr_papr_map_4k[i] + shift] = TRPAPR_CARRIER;
            }
            break;
          case FFTSIZE_8K:
          case FFTSIZE_8K_T2GI:
            for (int i = 0; i < 72; i++) {
              data_carrier_map[tr_papr_map_8k[i] + shift] = TRPAPR_CARRIER;
            }
            break;
          case FFTSIZE_16K:
          case FFTSIZE_16K_T2GI:
            for (int i = 0; i < 144; i++) {
              data_carrier_map[tr_papr_map_16k[i] + shift] = TRPAPR_CARRIER;
            }
            break;
          case FFTSIZE_32K:
          case FFTSIZE_32K_T2GI:
            for (int i = 0; i < 288; i++) {
              data_carrier_map[tr_papr_map_32k[i] + shift] = TRPAPR_CARRIER;
            }
            break;
        }
      }
   }

   void
   pilot_remove_renorm_impl::update_cmap(void)
   {
      const int &preamble = l1pre.s1;
      const int &pilotpattern = l1pre.pilot_pattern;
      const int &carriermode = l1pre.bwt_ext;
      const int &paprmode = l1pre.papr;
      const int &guardinterval = l1pre.guard_interval;
      const int &numdatasyms = l1pre.num_data_symbols;

      int p2_carrier_map[MAX_CARRIERS];
      int data_carrier_map[MAX_CARRIERS];
      int fc_carrier_map[MAX_CARRIERS];

      int step, ki;

      miso_group = MISO_TX1;
      if ((preamble == PREAMBLE_T2_SISO) || (preamble == PREAMBLE_T2_LITE_SISO)) {
        miso = false;
        switch (dvbt2_fftsize) {
          case FFTSIZE_1K:
            N_P2 = 16;
            C_P2 = 558;
            break;
          case FFTSIZE_2K:
            N_P2 = 8;
            C_P2 = 1118;
            break;
          case FFTSIZE_4K:
            N_P2 = 4;
            C_P2 = 2236;
            break;
          case FFTSIZE_8K:
          case FFTSIZE_8K_T2GI:
            N_P2 = 2;
            C_P2 = 4472;
            break;
          case FFTSIZE_16K:
          case FFTSIZE_16K_T2GI:
            N_P2 = 1;
            C_P2 = 8944;
            break;
          case FFTSIZE_32K:
          case FFTSIZE_32K_T2GI:
            N_P2 = 1;
            C_P2 = 22432;
            break;
        }
      }
      else {
        miso = true;
        switch (dvbt2_fftsize) {
          case FFTSIZE_1K:
            N_P2 = 16;
            C_P2 = 546;
            break;
          case FFTSIZE_2K:
            N_P2 = 8;
            C_P2 = 1098;
            break;
          case FFTSIZE_4K:
            N_P2 = 4;
            C_P2 = 2198;
            break;
          case FFTSIZE_8K:
          case FFTSIZE_8K_T2GI:
            N_P2 = 2;
            C_P2 = 4398;
            break;
          case FFTSIZE_16K:
          case FFTSIZE_16K_T2GI:
            N_P2 = 1;
            C_P2 = 8814;
            break;
          case FFTSIZE_32K:
          case FFTSIZE_32K_T2GI:
            N_P2 = 1;
            C_P2 = 17612;
            break;
        }
      }
      switch (dvbt2_fftsize) {
        case FFTSIZE_1K:
          C_PS = 853;
          K_EXT = 0;
          K_OFFSET = 0;
          break;
        case FFTSIZE_2K:
          C_PS = 1705;
          K_EXT = 0;
          K_OFFSET = 0;
          break;
        case FFTSIZE_4K:
          C_PS = 3409;
          K_EXT = 0;
          K_OFFSET = 0;
          break;
        case FFTSIZE_8K:
        case FFTSIZE_8K_T2GI:
          if (carriermode == CARRIERS_NORMAL) {
            C_PS = 6817;
            K_EXT = 0;
            K_OFFSET = 48;
          }
          else {
            C_PS = 6913;
            K_EXT = 48;
            K_OFFSET = 0;
          }
          break;
        case FFTSIZE_16K:
        case FFTSIZE_16K_T2GI:
          if (carriermode == CARRIERS_NORMAL) {
            C_PS = 13633;
            K_EXT = 0;
            K_OFFSET = 144;
          }
          else {
            C_PS = 13921;
            K_EXT = 144;
            K_OFFSET = 0;
          }
          break;
        case FFTSIZE_32K:
        case FFTSIZE_32K_T2GI:
          if (carriermode == CARRIERS_NORMAL) {
            C_PS = 27265;
            K_EXT = 0;
            K_OFFSET = 288;
          }
          else {
            C_PS = 27841;
            K_EXT = 288;
            K_OFFSET = 0;
          }
          break;
      }
      Kmin = 0; Kmax = C_PS - 1;
      switch (dvbt2_fftsize) {
        case FFTSIZE_1K:
          switch (pilotpattern) {
            case PILOT_PP1:
              C_DATA = 764;
              N_FC = 568;
              C_FC = 402;
              break;
            case PILOT_PP2:
              C_DATA = 768;
              N_FC = 710;
              C_FC = 654;
              break;
            case PILOT_PP3:
              C_DATA = 798;
              N_FC = 710;
              C_FC = 490;
              break;
            case PILOT_PP4:
              C_DATA = 804;
              N_FC = 780;
              C_FC = 707;
              break;
            case PILOT_PP5:
              C_DATA = 818;
              N_FC = 780;
              C_FC = 544;
              break;
            case PILOT_PP6:
              C_DATA = 0;
              N_FC = 0;
              C_FC = 0;
              break;
            case PILOT_PP7:
              C_DATA = 0;
              N_FC = 0;
              C_FC = 0;
              break;
            case PILOT_PP8:
              C_DATA = 0;
              N_FC = 0;
              C_FC = 0;
              break;
          }
          if (paprmode == PAPR_TR || paprmode == PAPR_BOTH) {
            if (C_DATA != 0) {
              C_DATA -= 10;
            }
            if (N_FC != 0) {
              N_FC -= 10;
            }
            if (C_FC != 0) {
              C_FC -= 10;
            }
          }
          break;
        case FFTSIZE_2K:
          switch (pilotpattern) {
            case PILOT_PP1:
              C_DATA = 1522;
              N_FC = 1136;
              C_FC = 804;
              break;
            case PILOT_PP2:
              C_DATA = 1532;
              N_FC = 1420;
              C_FC = 1309;
              break;
            case PILOT_PP3:
              C_DATA = 1596;
              N_FC = 1420;
              C_FC = 980;
              break;
            case PILOT_PP4:
              C_DATA = 1602;
              N_FC = 1562;
              C_FC = 1415;
              break;
            case PILOT_PP5:
              C_DATA = 1632;
              N_FC = 1562;
              C_FC = 1088;
              break;
            case PILOT_PP6:
              C_DATA = 0;
              N_FC = 0;
              C_FC = 0;
              break;
            case PILOT_PP7:
              C_DATA = 1646;
              N_FC = 1632;
              C_FC = 1396;
              break;
            case PILOT_PP8:
              C_DATA = 0;
              N_FC = 0;
              C_FC = 0;
              break;
          }
          if (paprmode == PAPR_TR || paprmode == PAPR_BOTH) {
            if (C_DATA != 0) {
              C_DATA -= 18;
            }
            if (N_FC != 0) {
              N_FC -= 18;
            }
            if (C_FC != 0) {
              C_FC -= 18;
            }
          }
          break;
        case FFTSIZE_4K:
          switch (pilotpattern) {
            case PILOT_PP1:
              C_DATA = 3084;
              N_FC = 2272;
              C_FC = 1609;
              break;
            case PILOT_PP2:
              C_DATA = 3092;
              N_FC = 2840;
              C_FC = 2619;
              break;
            case PILOT_PP3:
              C_DATA = 3228;
              N_FC = 2840;
              C_FC = 1961;
              break;
            case PILOT_PP4:
              C_DATA = 3234;
              N_FC = 3124;
              C_FC = 2831;
              break;
            case PILOT_PP5:
              C_DATA = 3298;
              N_FC = 3124;
              C_FC = 2177;
              break;
            case PILOT_PP6:
              C_DATA = 0;
              N_FC = 0;
              C_FC = 0;
              break;
            case PILOT_PP7:
              C_DATA = 3328;
              N_FC = 3266;
              C_FC = 2792;
              break;
            case PILOT_PP8:
              C_DATA = 0;
              N_FC = 0;
              C_FC = 0;
              break;
          }
          if (paprmode == PAPR_TR || paprmode == PAPR_BOTH) {
            if (C_DATA != 0) {
              C_DATA -= 36;
            }
            if (N_FC != 0) {
              N_FC -= 36;
            }
            if (C_FC != 0) {
              C_FC -= 36;
            }
          }
          break;
        case FFTSIZE_8K:
        case FFTSIZE_8K_T2GI:
          if (carriermode == CARRIERS_NORMAL) {
            switch (pilotpattern) {
              case PILOT_PP1:
                C_DATA = 6208;
                N_FC = 4544;
                C_FC = 3218;
                break;
              case PILOT_PP2:
                C_DATA = 6214;
                N_FC = 5680;
                C_FC = 5238;
                break;
              case PILOT_PP3:
                C_DATA = 6494;
                N_FC = 5680;
                C_FC = 3922;
                break;
              case PILOT_PP4:
                C_DATA = 6498;
                N_FC = 6248;
                C_FC = 5662;
                break;
              case PILOT_PP5:
                C_DATA = 6634;
                N_FC = 6248;
                C_FC = 4354;
                break;
              case PILOT_PP6:
                C_DATA = 0;
                N_FC = 0;
                C_FC = 0;
                break;
              case PILOT_PP7:
                C_DATA = 6698;
                N_FC = 6532;
                C_FC = 5585;
                break;
              case PILOT_PP8:
                C_DATA = 6698;
                N_FC = 0;
                C_FC = 0;
                break;
            }
          }
          else {
            switch (pilotpattern) {
              case PILOT_PP1:
                C_DATA = 6296;
                N_FC = 4608;
                C_FC = 3264;
                break;
              case PILOT_PP2:
                C_DATA = 6298;
                N_FC = 5760;
                C_FC = 5312;
                break;
              case PILOT_PP3:
                C_DATA = 6584;
                N_FC = 5760;
                C_FC = 3978;
                break;
              case PILOT_PP4:
                C_DATA = 6588;
                N_FC = 6336;
                C_FC = 5742;
                break;
              case PILOT_PP5:
                C_DATA = 6728;
                N_FC = 6336;
                C_FC = 4416;
                break;
              case PILOT_PP6:
                C_DATA = 0;
                N_FC = 0;
                C_FC = 0;
                break;
              case PILOT_PP7:
                C_DATA = 6788;
                N_FC = 6624;
                C_FC = 5664;
                break;
              case PILOT_PP8:
                C_DATA = 6788;
                N_FC = 0;
                C_FC = 0;
                break;
            }
          }
          if (paprmode == PAPR_TR || paprmode == PAPR_BOTH) {
            if (C_DATA != 0) {
              C_DATA -= 72;
            }
            if (N_FC != 0) {
              N_FC -= 72;
            }
            if (C_FC != 0) {
              C_FC -= 72;
            }
          }
          break;
        case FFTSIZE_16K:
        case FFTSIZE_16K_T2GI:
          if (carriermode == CARRIERS_NORMAL) {
            switch (pilotpattern) {
              case PILOT_PP1:
                C_DATA = 12418;
                N_FC = 9088;
                C_FC = 6437;
                break;
              case PILOT_PP2:
                C_DATA = 12436;
                N_FC = 11360;
                C_FC = 10476;
                break;
              case PILOT_PP3:
                C_DATA = 12988;
                N_FC = 11360;
                C_FC = 7845;
                break;
              case PILOT_PP4:
                C_DATA = 13002;
                N_FC = 12496;
                C_FC = 11324;
                break;
              case PILOT_PP5:
                C_DATA = 13272;
                N_FC = 12496;
                C_FC = 8709;
                break;
              case PILOT_PP6:
                C_DATA = 13288;
                N_FC = 13064;
                C_FC = 11801;
                break;
              case PILOT_PP7:
                C_DATA = 13416;
                N_FC = 13064;
                C_FC = 11170;
                break;
              case PILOT_PP8:
                C_DATA = 13406;
                N_FC = 0;
                C_FC = 0;
                break;
            }
          }
          else {
            switch (pilotpattern) {
              case PILOT_PP1:
                C_DATA = 12678;
                N_FC = 9280;
                C_FC = 6573;
                break;
              case PILOT_PP2:
                C_DATA = 12698;
                N_FC = 11600;
                C_FC = 10697;
                break;
              case PILOT_PP3:
                C_DATA = 13262;
                N_FC = 11600;
                C_FC = 8011;
                break;
              case PILOT_PP4:
                C_DATA = 13276;
                N_FC = 12760;
                C_FC = 11563;
                break;
              case PILOT_PP5:
                C_DATA = 13552;
                N_FC = 12760;
                C_FC = 8893;
                break;
              case PILOT_PP6:
                C_DATA = 13568;
                N_FC = 13340;
                C_FC = 12051;
                break;
              case PILOT_PP7:
                C_DATA = 13698;
                N_FC = 13340;
                C_FC = 11406;
                break;
              case PILOT_PP8:
                C_DATA = 13688;
                N_FC = 0;
                C_FC = 0;
                break;
            }
          }
          if (paprmode == PAPR_TR || paprmode == PAPR_BOTH) {
            if (C_DATA != 0) {
              C_DATA -= 144;
            }
            if (N_FC != 0) {
              N_FC -= 144;
            }
            if (C_FC != 0) {
              C_FC -= 144;
            }
          }
          break;
        case FFTSIZE_32K:
        case FFTSIZE_32K_T2GI:
          if (carriermode == CARRIERS_NORMAL) {
            switch (pilotpattern) {
              case PILOT_PP1:
                C_DATA = 0;
                N_FC = 0;
                C_FC = 0;
                break;
              case PILOT_PP2:
                C_DATA = 24886;
                N_FC = 22720;
                C_FC = 20952;
                break;
              case PILOT_PP3:
                C_DATA = 0;
                N_FC = 0;
                C_FC = 0;
                break;
              case PILOT_PP4:
                C_DATA = 26022;
                N_FC = 24992;
                C_FC = 22649;
                break;
              case PILOT_PP5:
                C_DATA = 0;
                N_FC = 0;
                C_FC = 0;
                break;
              case PILOT_PP6:
                C_DATA = 26592;
                N_FC = 26128;
                C_FC = 23603;
                break;
              case PILOT_PP7:
                C_DATA = 26836;
                N_FC = 0;
                C_FC = 0;
                break;
              case PILOT_PP8:
                C_DATA = 26812;
                N_FC = 0;
                C_FC = 0;
                break;
            }
          }
          else {
            switch (pilotpattern) {
              case PILOT_PP1:
                C_DATA = 0;
                N_FC = 0;
                C_FC = 0;
                break;
              case PILOT_PP2:
                C_DATA = 25412;
                N_FC = 23200;
                C_FC = 21395;
                break;
              case PILOT_PP3:
                C_DATA = 0;
                N_FC = 0;
                C_FC = 0;
                break;
              case PILOT_PP4:
                C_DATA = 26572;
                N_FC = 25520;
                C_FC = 23127;
                break;
              case PILOT_PP5:
                C_DATA = 0;
                N_FC = 0;
                C_FC = 0;
                break;
              case PILOT_PP6:
                C_DATA = 27152;
                N_FC = 26680;
                C_FC = 24102;
                break;
              case PILOT_PP7:
                C_DATA = 27404;
                N_FC = 0;
                C_FC = 0;
                break;
              case PILOT_PP8:
                C_DATA = 27376;
                N_FC = 0;
                C_FC = 0;
                break;
            }
          }
          if (paprmode == PAPR_TR || paprmode == PAPR_BOTH) {
            if (C_DATA != 0) {
              C_DATA -= 288;
            }
            if (N_FC != 0) {
              N_FC -= 288;
            }
            if (C_FC != 0) {
              C_FC -= 288;
            }
          }
          break;
      }
      if (miso == false) {
        if (guardinterval == GI_1_128 && pilotpattern == PILOT_PP7) {
          N_FC = 0;
          C_FC = 0;
        }
        if (guardinterval == GI_1_32 && pilotpattern == PILOT_PP4) {
          N_FC = 0;
          C_FC = 0;
        }
        if (guardinterval == GI_1_16 && pilotpattern == PILOT_PP2) {
          N_FC = 0;
          C_FC = 0;
        }
        if (guardinterval == GI_19_256 && pilotpattern == PILOT_PP2) {
          N_FC = 0;
          C_FC = 0;
        }
      }
      init_prbs();
      for (int i = 0; i < C_PS; i++) {
        p2_carrier_map[i] = DATA_CARRIER;
      }
      if ((dvbt2_fftsize == FFTSIZE_32K || dvbt2_fftsize == FFTSIZE_32K_T2GI) && (miso == false)) {
        step = 6;
      }
      else {
        step = 3;
      }
      for (int i = 0; i < C_PS; i += step) {
        if (miso == true && miso_group == MISO_TX2) {
          if (((i / 3) % 2) && (i % 3 == 0)) {
            p2_carrier_map[i] = P2PILOT_CARRIER_INVERTED;
          }
          else {
            p2_carrier_map[i] = P2PILOT_CARRIER;
          }
        }
        else {
          p2_carrier_map[i] = P2PILOT_CARRIER;
        }
      }
      if (carriermode == CARRIERS_EXTENDED) {
        for (int i = 0; i < K_EXT; i++) {
          if (miso == true && miso_group == MISO_TX2) {
            if (((i / 3) % 2) && (i % 3 == 0)) {
              p2_carrier_map[i] = P2PILOT_CARRIER_INVERTED;
            }
            else {
              p2_carrier_map[i] = P2PILOT_CARRIER;
            }
            if ((((i + (C_PS - K_EXT)) / 3) % 2) && ((i + (C_PS - K_EXT)) % 3 == 0)) {
              p2_carrier_map[i + (C_PS - K_EXT)] = P2PILOT_CARRIER_INVERTED;
            }
            else {
              p2_carrier_map[i + (C_PS - K_EXT)] = P2PILOT_CARRIER;
            }
          }
          else {
            p2_carrier_map[i] = P2PILOT_CARRIER;
            p2_carrier_map[i + (C_PS - K_EXT)] = P2PILOT_CARRIER;
          }
        }
      }
      if (miso == true) {
        p2_carrier_map[K_EXT + 1] = P2PILOT_CARRIER;
        p2_carrier_map[K_EXT + 2] = P2PILOT_CARRIER;
        p2_carrier_map[C_PS - K_EXT - 2] = P2PILOT_CARRIER;
        p2_carrier_map[C_PS - K_EXT - 3] = P2PILOT_CARRIER;
      }
      switch (dvbt2_fftsize) {
        case FFTSIZE_1K:
          for (int i = 0; i < 10; i++) {
            p2_carrier_map[p2_papr_map_1k[i]] = P2PAPR_CARRIER;
          }
          if (miso == true) {
            for (int i = 0; i < 10; i++) {
              ki = p2_papr_map_1k[i] + K_EXT;
              if (i < 9) {
                if (((ki % 3) == 1) && ((ki + 1) != (p2_papr_map_1k[i + 1] + K_EXT))) {
                  p2_carrier_map[ki + 1] = P2PILOT_CARRIER;
                }
              }
              else {
                if ((ki % 3) == 1) {
                  p2_carrier_map[ki + 1] = P2PILOT_CARRIER;
                }
              }
              if (i > 0) {
                if (((ki % 3) == 2) && ((ki - 1) != (p2_papr_map_1k[i - 1] + K_EXT))) {
                  p2_carrier_map[ki - 1] = P2PILOT_CARRIER;
                }
              }
              else {
                if ((ki % 3) == 2) {
                  p2_carrier_map[ki - 1] = P2PILOT_CARRIER;
                }
              }
            }
          }
          cp_bpsk[0] = gr_complex(4.0 / 3.0, 0.0);
          cp_bpsk[1] = gr_complex(-4.0 / 3.0, 0.0);
          cp_bpsk_inverted[0] = gr_complex(-4.0 / 3.0, 0.0);
          cp_bpsk_inverted[1] = gr_complex(4.0 / 3.0, 0.0);
          break;
        case FFTSIZE_2K:
          for (int i = 0; i < 18; i++) {
            p2_carrier_map[p2_papr_map_2k[i]] = P2PAPR_CARRIER;
          }
          if (miso == true) {
            for (int i = 0; i < 18; i++) {
              ki = p2_papr_map_2k[i] + K_EXT;
              if (i < 17) {
                if (((ki % 3) == 1) && ((ki + 1) != (p2_papr_map_2k[i + 1] + K_EXT))) {
                  p2_carrier_map[ki + 1] = P2PILOT_CARRIER;
                }
              }
              else {
                if ((ki % 3) == 1) {
                  p2_carrier_map[ki + 1] = P2PILOT_CARRIER;
                }
              }
              if (i > 0) {
                if (((ki % 3) == 2) && ((ki - 1) != (p2_papr_map_2k[i - 1] + K_EXT))) {
                  p2_carrier_map[ki - 1] = P2PILOT_CARRIER;
                }
              }
              else {
                if ((ki % 3) == 2) {
                  p2_carrier_map[ki - 1] = P2PILOT_CARRIER;
                }
              }
            }
          }
          cp_bpsk[0] = gr_complex(4.0 / 3.0, 0.0);
          cp_bpsk[1] = gr_complex(-4.0 / 3.0, 0.0);
          cp_bpsk_inverted[0] = gr_complex(-4.0 / 3.0, 0.0);
          cp_bpsk_inverted[1] = gr_complex(4.0 / 3.0, 0.0);
          break;
        case FFTSIZE_4K:
          for (int i = 0; i < 36; i++) {
            p2_carrier_map[p2_papr_map_4k[i]] = P2PAPR_CARRIER;
          }
          if (miso == true) {
            for (int i = 0; i < 36; i++) {
              ki = p2_papr_map_4k[i] + K_EXT;
              if (i < 35) {
                if (((ki % 3) == 1) && ((ki + 1) != (p2_papr_map_4k[i + 1] + K_EXT))) {
                  p2_carrier_map[ki + 1] = P2PILOT_CARRIER;
                }
              }
              else {
                if ((ki % 3) == 1) {
                  p2_carrier_map[ki + 1] = P2PILOT_CARRIER;
                }
              }
              if (i > 0) {
                if (((ki % 3) == 2) && ((ki - 1) != (p2_papr_map_4k[i - 1] + K_EXT))) {
                  p2_carrier_map[ki - 1] = P2PILOT_CARRIER;
                }
              }
              else {
                if ((ki % 3) == 2) {
                  p2_carrier_map[ki - 1] = P2PILOT_CARRIER;
                }
              }
            }
          }
          cp_bpsk[0] = gr_complex((4.0 * std::sqrt(2.0)) / 3.0, 0.0);
          cp_bpsk[1] = gr_complex(-(4.0 * std::sqrt(2.0)) / 3.0, 0.0);
          cp_bpsk_inverted[0] = gr_complex(-(4.0 * std::sqrt(2.0)) / 3.0, 0.0);
          cp_bpsk_inverted[1] = gr_complex((4.0 * std::sqrt(2.0)) / 3.0, 0.0);
          break;
        case FFTSIZE_8K:
        case FFTSIZE_8K_T2GI:
          for (int i = 0; i < 72; i++) {
            p2_carrier_map[p2_papr_map_8k[i] + K_EXT] = P2PAPR_CARRIER;
          }
          if (miso == true) {
            for (int i = 0; i < 72; i++) {
              ki = p2_papr_map_8k[i] + K_EXT;
              if (i < 71) {
                if (((ki % 3) == 1) && ((ki + 1) != (p2_papr_map_8k[i + 1] + K_EXT))) {
                  p2_carrier_map[ki + 1] = P2PILOT_CARRIER;
                }
              }
              else {
                if ((ki % 3) == 1) {
                  p2_carrier_map[ki + 1] = P2PILOT_CARRIER;
                }
              }
              if (i > 0) {
                if (((ki % 3) == 2) && ((ki - 1) != (p2_papr_map_8k[i - 1] + K_EXT))) {
                  p2_carrier_map[ki - 1] = P2PILOT_CARRIER;
                }
              }
              else {
                if ((ki % 3) == 2) {
                  p2_carrier_map[ki - 1] = P2PILOT_CARRIER;
                }
              }
            }
          }
          cp_bpsk[0] = gr_complex(8.0 / 3.0, 0.0);
          cp_bpsk[1] = gr_complex(-8.0 / 3.0, 0.0);
          cp_bpsk_inverted[0] = gr_complex(-8.0 / 3.0, 0.0);
          cp_bpsk_inverted[1] = gr_complex(8.0 / 3.0, 0.0);
          break;
        case FFTSIZE_16K:
        case FFTSIZE_16K_T2GI:
          for (int i = 0; i < 144; i++) {
            p2_carrier_map[p2_papr_map_16k[i] + K_EXT] = P2PAPR_CARRIER;
          }
          if (miso == true) {
            for (int i = 0; i < 144; i++) {
              ki = p2_papr_map_16k[i] + K_EXT;
              if (i < 143) {
                if (((ki % 3) == 1) && ((ki + 1) != (p2_papr_map_16k[i + 1] + K_EXT))) {
                  p2_carrier_map[ki + 1] = P2PILOT_CARRIER;
                }
              }
              else {
                if ((ki % 3) == 1) {
                  p2_carrier_map[ki + 1] = P2PILOT_CARRIER;
                }
              }
              if (i > 0) {
                if (((ki % 3) == 2) && ((ki - 1) != (p2_papr_map_16k[i - 1] + K_EXT))) {
                  p2_carrier_map[ki - 1] = P2PILOT_CARRIER;
                }
              }
              else {
                if ((ki % 3) == 2) {
                  p2_carrier_map[ki - 1] = P2PILOT_CARRIER;
                }
              }
            }
          }
          cp_bpsk[0] = gr_complex(8.0 / 3.0, 0.0);
          cp_bpsk[1] = gr_complex(-8.0 / 3.0, 0.0);
          cp_bpsk_inverted[0] = gr_complex(-8.0 / 3.0, 0.0);
          cp_bpsk_inverted[1] = gr_complex(8.0 / 3.0, 0.0);
          break;
        case FFTSIZE_32K:
        case FFTSIZE_32K_T2GI:
          for (int i = 0; i < 288; i++) {
            p2_carrier_map[p2_papr_map_32k[i] + K_EXT] = P2PAPR_CARRIER;
          }
          if (miso == true) {
            for (int i = 0; i < 288; i++) {
              ki = p2_papr_map_32k[i] + K_EXT;
              if (i < 287) {
                if (((ki % 3) == 1) && ((ki + 1) != (p2_papr_map_32k[i + 1] + K_EXT))) {
                  p2_carrier_map[ki + 1] = P2PILOT_CARRIER;
                }
              }
              else {
                if ((ki % 3) == 1) {
                  p2_carrier_map[ki + 1] = P2PILOT_CARRIER;
                }
              }
              if (i > 0) {
                if (((ki % 3) == 2) && ((ki - 1) != (p2_papr_map_32k[i - 1] + K_EXT))) {
                  p2_carrier_map[ki - 1] = P2PILOT_CARRIER;
                }
              }
              else {
                if ((ki % 3) == 2) {
                  p2_carrier_map[ki - 1] = P2PILOT_CARRIER;
                }
              }
            }
          }
          cp_bpsk[0] = gr_complex(8.0 / 3.0, 0.0);
          cp_bpsk[1] = gr_complex(-8.0 / 3.0, 0.0);
          cp_bpsk_inverted[0] = gr_complex(-8.0 / 3.0, 0.0);
          cp_bpsk_inverted[1] = gr_complex(8.0 / 3.0, 0.0);
          break;
      }
      switch (pilotpattern) {
        case PILOT_PP1:
          sp_bpsk[0] = gr_complex(4.0 / 3.0, 0.0);
          sp_bpsk[1] = gr_complex(-4.0 / 3.0, 0.0);
          sp_bpsk_inverted[0] = gr_complex(-4.0 / 3.0, 0.0);
          sp_bpsk_inverted[1] = gr_complex(4.0 / 3.0, 0.0);
          dx = 3;
          dy = 4;
          break;
        case PILOT_PP2:
          sp_bpsk[0] = gr_complex(4.0 / 3.0, 0.0);
          sp_bpsk[1] = gr_complex(-4.0 / 3.0, 0.0);
          sp_bpsk_inverted[0] = gr_complex(-4.0 / 3.0, 0.0);
          sp_bpsk_inverted[1] = gr_complex(4.0 / 3.0, 0.0);
          dx = 6;
          dy = 2;
          break;
        case PILOT_PP3:
          sp_bpsk[0] = gr_complex(7.0 / 4.0, 0.0);
          sp_bpsk[1] = gr_complex(-7.0 / 4.0, 0.0);
          sp_bpsk_inverted[0] = gr_complex(-7.0 / 4.0, 0.0);
          sp_bpsk_inverted[1] = gr_complex(7.0 / 4.0, 0.0);
          dx = 6;
          dy = 4;
          break;
        case PILOT_PP4:
          sp_bpsk[0] = gr_complex(7.0 / 4.0, 0.0);
          sp_bpsk[1] = gr_complex(-7.0 / 4.0, 0.0);
          sp_bpsk_inverted[0] = gr_complex(-7.0 / 4.0, 0.0);
          sp_bpsk_inverted[1] = gr_complex(7.0 / 4.0, 0.0);
          dx = 12;
          dy = 2;
          break;
        case PILOT_PP5:
          sp_bpsk[0] = gr_complex(7.0 / 3.0, 0.0);
          sp_bpsk[1] = gr_complex(-7.0 / 3.0, 0.0);
          sp_bpsk_inverted[0] = gr_complex(-7.0 / 3.0, 0.0);
          sp_bpsk_inverted[1] = gr_complex(7.0 / 3.0, 0.0);
          dx = 12;
          dy = 4;
          break;
        case PILOT_PP6:
          sp_bpsk[0] = gr_complex(7.0 / 3.0, 0.0);
          sp_bpsk[1] = gr_complex(-7.0 / 3.0, 0.0);
          sp_bpsk_inverted[0] = gr_complex(-7.0 / 3.0, 0.0);
          sp_bpsk_inverted[1] = gr_complex(7.0 / 3.0, 0.0);
          dx = 24;
          dy = 2;
          break;
        case PILOT_PP7:
          sp_bpsk[0] = gr_complex(7.0 / 3.0, 0.0);
          sp_bpsk[1] = gr_complex(-7.0 / 3.0, 0.0);
          sp_bpsk_inverted[0] = gr_complex(-7.0 / 3.0, 0.0);
          sp_bpsk_inverted[1] = gr_complex(7.0 / 3.0, 0.0);
          dx = 24;
          dy = 4;
          break;
        case PILOT_PP8:
          sp_bpsk[0] = gr_complex(7.0 / 3.0, 0.0);
          sp_bpsk[1] = gr_complex(-7.0 / 3.0, 0.0);
          sp_bpsk_inverted[0] = gr_complex(-7.0 / 3.0, 0.0);
          sp_bpsk_inverted[1] = gr_complex(7.0 / 3.0, 0.0);
          dx = 6;
          dy = 16;
          break;
      }
      for (int i = 0; i < C_PS; i++) {
        fc_carrier_map[i] = DATA_CARRIER;
      }
      for (int i = 0; i < C_PS; i++) {
        if (i % dx == 0) {
          if (miso == true && miso_group == MISO_TX2) {
            if ((i / dx) % 2) {
              fc_carrier_map[i] = SCATTERED_CARRIER_INVERTED;
            }
            else {
              fc_carrier_map[i] = SCATTERED_CARRIER;
            }
          }
          else {
            fc_carrier_map[i] = SCATTERED_CARRIER;
          }
        }
      }
      if (dvbt2_fftsize == FFTSIZE_1K && pilotpattern == PILOT_PP4) {
        fc_carrier_map[C_PS - 2] = SCATTERED_CARRIER;
      }
      else if (dvbt2_fftsize == FFTSIZE_1K && pilotpattern == PILOT_PP5) {
        fc_carrier_map[C_PS - 2] = SCATTERED_CARRIER;
      }
      else if (dvbt2_fftsize == FFTSIZE_2K && pilotpattern == PILOT_PP7) {
        fc_carrier_map[C_PS - 2] = SCATTERED_CARRIER;
      }
      if (miso == true && miso_group == MISO_TX2) {
        if ((numdatasyms + N_P2 - 1) % 2) {
          fc_carrier_map[0] = SCATTERED_CARRIER_INVERTED;
          fc_carrier_map[C_PS - 1] = SCATTERED_CARRIER_INVERTED;
        }
        else {
          fc_carrier_map[0] = SCATTERED_CARRIER;
          fc_carrier_map[C_PS - 1] = SCATTERED_CARRIER;
        }
      }
      else {
        fc_carrier_map[0] = SCATTERED_CARRIER;
        fc_carrier_map[C_PS - 1] = SCATTERED_CARRIER;
      }
      if (paprmode == PAPR_TR || paprmode == PAPR_BOTH) {
        switch (dvbt2_fftsize) {
          case FFTSIZE_1K:
            for (int i = 0; i < 10; i++) {
              fc_carrier_map[p2_papr_map_1k[i]] = TRPAPR_CARRIER;
            }
            break;
          case FFTSIZE_2K:
            for (int i = 0; i < 18; i++) {
              fc_carrier_map[p2_papr_map_2k[i]] = TRPAPR_CARRIER;
            }
            break;
          case FFTSIZE_4K:
            for (int i = 0; i < 36; i++) {
              fc_carrier_map[p2_papr_map_4k[i]] = TRPAPR_CARRIER;
            }
            break;
          case FFTSIZE_8K:
          case FFTSIZE_8K_T2GI:
            for (int i = 0; i < 72; i++) {
              fc_carrier_map[p2_papr_map_8k[i] + K_EXT] = TRPAPR_CARRIER;
            }
            break;
          case FFTSIZE_16K:
          case FFTSIZE_16K_T2GI:
            for (int i = 0; i < 144; i++) {
              fc_carrier_map[p2_papr_map_16k[i] + K_EXT] = TRPAPR_CARRIER;
            }
            break;
          case FFTSIZE_32K:
          case FFTSIZE_32K_T2GI:
            for (int i = 0; i < 288; i++) {
              fc_carrier_map[p2_papr_map_32k[i] + K_EXT] = TRPAPR_CARRIER;
            }
            break;
        }
      }



      if ((dvbt2_fftsize == FFTSIZE_32K || dvbt2_fftsize == FFTSIZE_32K_T2GI) && (miso == false)) {
        p2_bpsk[0] = gr_complex(std::sqrt(37.0) / 5.0, 0.0);
        p2_bpsk[1] = gr_complex(-(std::sqrt(37.0) / 5.0), 0.0);
        p2_bpsk_inverted[0] = gr_complex(-(std::sqrt(37.0) / 5.0), 0.0);
        p2_bpsk_inverted[1] = gr_complex(std::sqrt(37.0) / 5.0, 0.0);
      }
      else {
        p2_bpsk[0] = gr_complex(std::sqrt(31.0) / 5.0, 0.0);
        p2_bpsk[1] = gr_complex(-(std::sqrt(31.0) / 5.0), 0.0);
        p2_bpsk_inverted[0] = gr_complex(-(std::sqrt(31.0) / 5.0), 0.0);
        p2_bpsk_inverted[1] = gr_complex(std::sqrt(31.0) / 5.0, 0.0);
      }



      int symb=0;
      for ( ; symb<N_P2; ++symb )
      {
         for ( int k=0; k<C_PS; ++k )
         { carrier_map[symb][k] = p2_carrier_map[k]; }
      }
      for ( ; symb<N_P2+numdatasyms-( N_FC != 0 ? 1 : 0 ); ++symb )
      {
         init_pilots( carrier_map[symb], symb );
      }
      if ( N_FC != 0 )
      { 
         for ( int k=0; k<C_PS; ++k )
         { carrier_map[symb][k] = fc_carrier_map[k]; }
      }



   }


   void
   pilot_remove_renorm_impl::update_H(void)
   {
      int max_states, xor_size, pn_mask, result;
      int q_even = 0;
      int q_odd = 0;
      int q_evenP2 = 0;
      int q_oddP2 = 0;
      int q_evenFC = 0;
      int q_oddFC = 0;
      int lfsr = 0;
      int logic1k[2] = {0, 4};
      int logic2k[2] = {0, 3};
      int logic4k[2] = {0, 2};
      int logic8k[4] = {0, 1, 4, 6};
      int logic16k[6] = {0, 1, 4, 5, 9, 11};
      int logic32k[4] = {0, 1, 2, 12};
      int *logic;
      const int *bitpermeven, *bitpermodd;
      int pn_degree, even, odd;


      switch (dvbt2_fftsize) {
        case FFTSIZE_1K:
          pn_degree = 9;
          pn_mask = 0x1ff;
          max_states = 1024;
          logic = &logic1k[0];
          xor_size = 2;
          bitpermeven = &bitperm1keven[0];
          bitpermodd = &bitperm1kodd[0];
          break;
        case FFTSIZE_2K:
          pn_degree = 10;
          pn_mask = 0x3ff;
          max_states = 2048;
          logic = &logic2k[0];
          xor_size = 2;
          bitpermeven = &bitperm2keven[0];
          bitpermodd = &bitperm2kodd[0];
          break;
        case FFTSIZE_4K:
          pn_degree = 11;
          pn_mask = 0x7ff;
          max_states = 4096;
          logic = &logic4k[0];
          xor_size = 2;
          bitpermeven = &bitperm4keven[0];
          bitpermodd = &bitperm4kodd[0];
          break;
        case FFTSIZE_8K:
        case FFTSIZE_8K_T2GI:
          pn_degree = 12;
          pn_mask = 0xfff;
          max_states = 8192;
          logic = &logic8k[0];
          xor_size = 4;
          bitpermeven = &bitperm8keven[0];
          bitpermodd = &bitperm8kodd[0];
          break;
        case FFTSIZE_16K:
        case FFTSIZE_16K_T2GI:
          pn_degree = 13;
          pn_mask = 0x1fff;
          max_states = 16384;
          logic = &logic16k[0];
          xor_size = 6;
          bitpermeven = &bitperm16keven[0];
          bitpermodd = &bitperm16kodd[0];
          break;
        case FFTSIZE_32K:
        case FFTSIZE_32K_T2GI:
          pn_degree = 14;
          pn_mask = 0x3fff;
          max_states = 32768;
          logic = &logic32k[0];
          xor_size = 4;
          bitpermeven = &bitperm32k[0];
          bitpermodd = &bitperm32k[0];
          break;
        default:
          pn_degree = 0;
          pn_mask = 0;
          max_states = 0;
          logic = &logic1k[0];
          xor_size = 0;
          break;
      }


      for (int i = 0; i < max_states; i++) {
        if (i == 0 || i == 1) {
          lfsr = 0;
        }
        else if (i == 2) {
          lfsr = 1;
        }
        else {
          result = 0;
          for (int k = 0; k < xor_size; k++) {
            result ^= (lfsr >> logic[k]) & 1;
          }
          lfsr &= pn_mask;
          lfsr >>= 1;
          lfsr |= result << (pn_degree - 1);
        }
        even = 0;
        odd = 0;
        for (int n = 0; n < pn_degree; n++) {
          even |= ((lfsr >> n) & 0x1) << bitpermeven[n];
        }
        for (int n = 0; n < pn_degree; n++) {
          odd |= ((lfsr >> n) & 0x1) << bitpermodd[n];
        }
        even = even + ((i % 2) * (max_states / 2));
        odd = odd + ((i % 2) * (max_states / 2));
        if (even < C_DATA) {
          Heven[q_even++] = even;
        }
        if (odd < C_DATA) {
          Hodd[q_odd++] = odd;
        }
        if (even < C_P2) {
          HevenP2[q_evenP2++] = even;
        }
        if (odd < C_P2) {
          HoddP2[q_oddP2++] = odd;
        }
        if (even < N_FC) {
          HevenFC[q_evenFC++] = even;
        }
        if (odd < N_FC) {
          HoddFC[q_oddFC++] = odd;
        }
      }
      if (dvbt2_fftsize == FFTSIZE_32K || dvbt2_fftsize == FFTSIZE_32K_T2GI) {
        for (int j = 0; j < q_odd; j++) {
          int a;
          a = Hodd[j];
          Heven[a] = j;
        }
        for (int j = 0; j < q_oddP2; j++) {
          int a;
          a = HoddP2[j];
          HevenP2[a] = j;
        }
        for (int j = 0; j < q_oddFC; j++) {
          int a;
          a = HoddFC[j];
          HevenFC[a] = j;
        }
      }


   }


   void
   pilot_remove_renorm_impl::forecast (int noutput_items, gr_vector_int &ninput_items_required)
   {
      ninput_items_required[0] = ceil( forecast_f * noutput_items + 1 );
   }


    int
    pilot_remove_renorm_impl::general_work (int noutput_items,
                       gr_vector_int &ninput_items,
                       gr_vector_const_void_star &input_items,
                       gr_vector_void_star &output_items)
    {
      const gr_complex *in = (const gr_complex *) input_items[0];
      gr_complex *out = (gr_complex *) output_items[0];

      int nofnnk;
      int nnk[MAX_FFT_SIZE];

      static bool synced = false;
      static int symbol = 0; // symbol #
      static uint64_t framestart = 0;


// handle frame start tag
      std::vector<tag_t> tags;
      std::vector<tag_t>::iterator t;
      get_tags_in_range( tags, 0,
                         nitems_read(0), nitems_read(0) + ninput_items[0] );

      for ( t = tags.begin(); (!synced) && (t != tags.end()); ++t )
      {
         if ( t->key == pmt::intern("L1-Pre") )
         {
            tag_t &tagL1 = *t;
            synced = true;
            symbol = 0;
            framestart = tagL1.offset;

            pmt::pmt_t L1TagValue_t = tagL1.value;
            if ( L1TagValue != L1TagValue_t )
            {
               setup_L1pre( l1pre, L1TagValue_t );

///l1pre_out( l1pre );

               dvbt2_fftsize = l1pre.s2>>1;
               switch ( dvbt2_fftsize )
               {
                  case FFTSIZE_2K:  { fftsize= 2*1024; break; }
                  case FFTSIZE_8K:  { fftsize= 8*1024; break; }
                  case FFTSIZE_4K:  { fftsize= 4*1024; break; }
                  case FFTSIZE_1K:  { fftsize= 1*1024; break; }
                  case FFTSIZE_16K: { fftsize=16*1024; break; }
                  case FFTSIZE_32K: { fftsize=32*1024; break; }
                  case FFTSIZE_8K_T2GI:  { fftsize= 8*1024; break; }
                  case FFTSIZE_32K_T2GI: { fftsize=32*1024; break; }
                  case FFTSIZE_16K_T2GI: { fftsize=16*1024; break; }
               }

               update_cmap();
               update_H();

               for ( int k=0; k<C_PS; ++k )
               {
                  int ks = k - (Kmax-Kmin)/2;
                  k2idx[k] = ( ks < 0 ? ks+fftsize : ks );
               }

               L1TagValue = L1TagValue_t;

               forecast_f = (fftsize*(N_P2 + l1pre.num_data_symbols))/(double)(C_P2*N_P2 + C_DATA*l1pre.num_data_symbols);

            }

            break;
         }

      }


      // Do <+signal processing+>
      // Tell runtime system how many input items we consumed on
      // each input stream.

// not enough to start frame?
      if ( ! synced )
      {
         consume_each(ninput_items[0]);
         return(0);
      }
      else if ( symbol == 0 )
      {
         int maxNin = ninput_items[0] - fftsize;
         if ( maxNin < 0 )
         {
            consume_each(0);
            return(0);
         }

         uint64_t uld = framestart - nitems_read(0); // modulo difference
         if ( uld > 0 )
         {
            consume_each( uld );
            return(0);
         }
// => framestart = nitems_read(0);
      }


      nofnnk = 0;
// update normalization
      for ( int k=0; k<C_PS; ++k )
      {
         int &idx = k2idx[k];

         int bpsk = ( prbs[k + K_OFFSET ] ^ pn_sequence[symbol] ? 1 : 0 );

         switch ( carrier_map[symbol][k] )
         {
            case P2PILOT_CARRIER:
            {
               c_norm[idx] = p2_bpsk[ bpsk ] / in[idx];
               nnk[ nofnnk++ ] = k;
               break;
            }
            case SCATTERED_CARRIER:
            {
               c_norm[idx] = sp_bpsk[ bpsk ] / in[idx];
               nnk[ nofnnk++ ] = k;
               break;
            }
            case CONTINUAL_CARRIER:
            {
               c_norm[idx] = cp_bpsk[ bpsk ] / in[idx];
               nnk[ nofnnk++ ] = k;
               break;
            }
            case P2PILOT_CARRIER_INVERTED:
            {
               c_norm[idx] = p2_bpsk_inverted[ bpsk ] / in[idx];
               nnk[ nofnnk++ ] = k;
               break;
            }
            case SCATTERED_CARRIER_INVERTED:
            {
               c_norm[idx] = sp_bpsk_inverted[ bpsk ] / in[idx];
               nnk[ nofnnk++ ] = k;
               break;
            }
            case CONTINUAL_CARRIER_INVERTED:
            {
               c_norm[idx] = cp_bpsk_inverted[ bpsk ] / in[idx];
               nnk[ nofnnk++ ] = k;
               break;
            }
            case P2PAPR_CARRIER:
            case TRPAPR_CARRIER:
            default:
            {
               c_norm[idx] = 0;
               break;
            }
         }
      }

//
// c_norm filter
      for ( int sec=1; sec<nofnnk; ++sec )
      {
         int k0 = nnk[sec-1];
         int k1 = nnk[sec  ];
         float D = 1.0f / (float)(k1-k0);

         gr_complex &cn0 = c_norm[ k2idx[k0] ];
         gr_complex &cn1 = c_norm[ k2idx[k1] ];

         for ( int k=k0+1; k<k1; ++k )
         {
            int &idx = k2idx[k];
            c_norm[idx] = ( cn1*(float)(k-k0) + cn0*(float)(k1-k) )*D;
         }
      }


// output norm
// freq deinterleave
      int dcarriers = 0;
      int *H;

      if ( symbol < N_P2 )
      {
         if ((symbol % 2) == 0) { H = HevenP2; }
         else                   { H = HoddP2;  }
      }
      else if ( symbol < l1pre.num_data_symbols+N_P2-( N_FC!=0 ? 1 : 0 ) )
      {
         if ((symbol % 2) == 0) { H = Heven; }         
         else                   { H = Hodd;  }
      }
      else
      {
         if ((symbol % 2) == 0) { H = HevenFC; }
         else                   { H = HoddFC;  }
      }

      for ( int k=0; k<C_PS; ++k )
      {
         int &idx = k2idx[k];

         if ( carrier_map[symbol][k] == DATA_CARRIER )
         {
            out[ H[ dcarriers++ ] ] = in[idx] * c_norm[idx];
         }
      }

      if ( symbol == 0 )
      {
         uint64_t tpos = nitems_written(0);
         add_item_tag( 0, tpos, pmt::intern("L1-Pre"), L1TagValue );
      }

      if ( symbol < l1pre.num_data_symbols+N_P2-1 )
      { ++symbol; }
      else
      { synced = false; }

      consume_each(fftsize);

      // Tell runtime system how many output items we produced.
      return( dcarriers );
    }

  } /* namespace DVBparts */
} /* namespace gr */

