/* -*- c++ -*- */
/* 
 * Copyright 2018 Ingo D. Rullhusen <d01c@freeshell.de>.
 * 
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gnuradio/io_signature.h>
#include "time_deinterleaver_impl.h"


using namespace std;


namespace gr {
  namespace DVBparts {

    time_deinterleaver::sptr
    time_deinterleaver::make()
    {
      return gnuradio::get_initial_sptr
        (new time_deinterleaver_impl());
    }

    /*
     * The private constructor
     */
    time_deinterleaver_impl::time_deinterleaver_impl()
      : gr::block("time_deinterleaver",
              gr::io_signature::make(1,1, sizeof(gr_complex)),
              gr::io_signature::make(1,1, sizeof(gr_complex)))
    {
       ti_buffer_offset = 0;
       bzero(&PLPsl,sizeof(PLPsl));
       set_tag_propagation_policy(TPP_DONT);
    }

    /*
     * Our virtual destructor.
     */
    time_deinterleaver_impl::~time_deinterleaver_impl()
    {}

    void
    time_deinterleaver_impl::forecast (int noutput_items, gr_vector_int &ninput_items_required)
    { ninput_items_required[0] = noutput_items; }

    int
    time_deinterleaver_impl::general_work (int noutput_items,
                       gr_vector_int &ninput_items,
                       gr_vector_const_void_star &input_items,
                       gr_vector_void_star &output_items)
    {
       const gr_complex *in = (const gr_complex *) input_items[0];
       gr_complex *out = (gr_complex *) output_items[0];

       static enum { Wait, Read, Write, Copy } state = Wait;


       static int PI = 1;
       static int NTI = 1;
       static int IJUMP = 1;
       static int N_BLOCKS_IF = 0;
       static int s = 1;


       static int Nc = 0;
       static int Nr = 0;
       static int ti_c = 0;
       static int ti_r = 0;
       if ( state == Write )
       {
          int nout = 0;

          for ( ; ti_c<Nc; ++ti_c )
          {
             for ( ; ti_r<Nr; ++ti_r )
             {
                if ( nout < noutput_items )
                {
                   out[nout++] = ti_buffer[ Nc*ti_r + ti_c ];
                }
                else
                { goto NcNrExit; }
             }
             ti_r = 0;
          }
          state = Wait;
NcNrExit:
          consume_each(0);
          return(nout);
       }


       static int items_to_read = 0;
       if ( state == Read )
       {
          int nin = items_to_read;
          if ( nin > ninput_items[0] ) { nin = ninput_items[0]; }
          memcpy( &ti_buffer[ti_buffer_offset], in, nin*sizeof(gr_complex) );
          ti_buffer_offset += nin;
          items_to_read -= nin;

          if ( items_to_read < 1 ) { state = Write; }

          consume_each(nin);
          return(0);
       }

       if ( state == Copy )
       {
          int nin = items_to_read;
          if ( nin > ninput_items[0] ) { nin = ninput_items[0]; }
          memcpy( out, in, nin*sizeof(gr_complex) );
          items_to_read -= nin;   

          if ( items_to_read < 1 ) { state = Wait; }

          consume_each(nin);
          return(nin);
       }



// next TI frame
       if ( s < NTI )
       {
          int N_FEC_TI = N_BLOCKS_IF;
          if ( NTI > 1 )
          {
             N_FEC_TI = floor( N_BLOCKS_IF / NTI );
             if ( s >= NTI - (N_BLOCKS_IF % NTI) ) { ++N_FEC_TI; }
          }

          Nc = 5*N_FEC_TI;
          ti_c = 0;
          ti_r = 0;

          state = Read;
          ti_buffer_offset = 0;
          items_to_read = Nr * Nc;
if ( TI_BUFFER_SIZE < items_to_read ) { cerr << "TI: Illegal FEC block count!" << endl; }

          uint64_t tpos = nitems_written(0);
          add_item_tag( 0, tpos, pmt::intern("PLPti"),
                           pmt::make_tuple(
                              pmt::from_long(items_to_read),
                              pmt::from_long(N_FEC_TI),
                              pmt::from_long(s),
                              PLPTagValue
                                          )
                      );

          ++s;

          consume_each(0);
          return(0);
       }
       else if ( s == 0 && NTI == 0 )
       {
          int N_FEC_TI = N_BLOCKS_IF;
          items_to_read = PLPsl.length * PLPsl.Nsubslices * PI;

          Nc = 1;
          ti_c = 0;
          ti_r = 0;

          state = Copy;

          uint64_t tpos = nitems_written(0);
          add_item_tag( 0, tpos, pmt::intern("PLPti"),
                           pmt::make_tuple(
                              pmt::from_long(items_to_read),
                              pmt::from_long(N_FEC_TI),
                              pmt::from_long(s),
                              PLPTagValue
                                          )
                      );

          ++s;

          consume_each(0);
          return(0);
       }


// handle frame start tag
       uint64_t uld = ninput_items[0];

       std::vector<tag_t> tags;
       get_tags_in_range( tags, 0,
                          nitems_read(0), nitems_read(0) + ninput_items[0],
                          pmt::intern("PLPslice") );

       if ( tags.size() > 0 )
       {
          tag_t &tagSLICE = tags[0];

          setup_PLPslice( PLPsl, tagSLICE.value );
          const L1Pplp &P = PLPsl.l1plp;

          PI  = ( P.time_il_type ? P.time_il_length : 1 );
          IJUMP = P.frame_interval;

// first
          if ( ( ( PLPsl.frame_idx % (PI*IJUMP) ) == 0 ) &&
               (              PLPsl.sub_slice_idx == 0 )    )
          {
             PLPTagValue = pmt::tuple_ref(tagSLICE.value,5);

             const int NR[2][4] = { {  8100/5,  4050/5,  2700/5, 2025/5 },
                                    { 32400/5, 16200/5, 10800/5, 8100/5 }  };

             Nr = NR[ P.plp_fec_type ][ P.plp_mod ];
                
             N_BLOCKS_IF = PLPsl.N_BLOCKS_IF;

             NTI = ( P.time_il_type ? 1 : P.time_il_length );
             s = 0;  // start TI frame

             uint64_t framestart = tagSLICE.offset;
             uld = framestart - nitems_read(0); // modulo difference
          }

       }

       consume_each(uld);
       return(0);
    }

  } /* namespace DVBparts */
} /* namespace gr */

