/* -*- c++ -*- */
/*
 * Copyright 2021 Ingo D. Rullhusen <d01c@freeshell.de>.
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gnuradio/io_signature.h>
#include "xcopy_impl.h"

namespace gr {
  namespace DVBparts {

    xcopy::sptr
    xcopy::make(bool enable)
    {
      return gnuradio::get_initial_sptr
        (new xcopy_impl(enable));
    }


    /*
     * The private constructor
     */
    xcopy_impl::xcopy_impl(bool enable)
      : gr::block("xcopy",
        gr::io_signature::make(1,1, sizeof(gr_complex)),
        gr::io_signature::make(1,1, sizeof(gr_complex))),
        d_enabled(enable)
    { set_tag_propagation_policy(TPP_CUSTOM); }

    /*
     * Our virtual destructor.
     */
    xcopy_impl::~xcopy_impl()
    {}


    void xcopy_impl::set_enabled(bool enable)
    {
       d_enabled = enable;
    }


    void
    xcopy_impl::forecast (int noutput_items, gr_vector_int &ninput_items_required)
    {
      /* <+forecast+> e.g. ninput_items_required[0] = noutput_items */
      ninput_items_required[0] = noutput_items;
    }

    int
    xcopy_impl::general_work (int noutput_items,
                       gr_vector_int &ninput_items,
                       gr_vector_const_void_star &input_items,
                       gr_vector_void_star &output_items)
    {
      const gr_complex *in = (const gr_complex *) input_items[0];
      gr_complex *out = (gr_complex *) output_items[0];

      int n = 0;
      if (d_enabled)
      {

// fix tag offset
         std::vector<tag_t> tags;
         std::vector<tag_t>::iterator t;
         get_tags_in_range( tags, 0,
                            nitems_read(0), nitems_read(0) + noutput_items );

         for ( t = tags.begin(); t != tags.end(); ++t )
         {
            uint64_t ipos = t->offset;
            uint64_t uld = ipos - nitems_read(0); // modulo difference
            uint64_t opos = nitems_written(0) + uld;
            add_item_tag( 0, opos, t->key, t->value );
         }

         memcpy( out, in, noutput_items * sizeof(gr_complex) );
         n = noutput_items;

      }

      consume_each(noutput_items);
      return n;
    }

  } /* namespace DVBparts */
} /* namespace gr */

