#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2021 Ingo D. Rullhusen <d01c@freeshell.de>.
#
# This is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this software; see the file COPYING.  If not, write to
# the Free Software Foundation, Inc., 51 Franklin Street,
# Boston, MA 02110-1301, USA.
#

from gnuradio import gr, gr_unittest
from gnuradio import blocks, dtv
import DVBparts_swig as DVBparts

import pmt

import random

class qa_ldpc_decoder(gr_unittest.TestCase):

    def setUp(self):
        self.tb = gr.top_block()
        self.nchunks = 4
        self.nframes = self.nchunks * 4 - 1
        
        self.ref_tb = gr.top_block()

    def tearDown(self):
        self.tb = None
        
        self.ref_tb = None

    def test_001_RB(self):

        # set up data

        PLPTagValue = [ 0,1,3,0,0,0,0,0,0,0,1,33,1,2,0,1,0,2,1,0 ]
        #  plp.plp_id = 0
        #  plp.plp_type = 1
        #  plp.plp_payload_type = 3
        #  plp.ff_flag = 0
        #  plp.first_rf_idx = 0
        #  plp.first_frame_index = 0
        #  plp.plp_group_id = 0
        #  plp.plp_cod = 0      : 1/2 -> DVB_S2_TABLE_B4 -> M = 360, N = 64800, K = 32400
        N = 64800
        K = 32400
        #  plp.plp_mod = 0
        #  plp.plp_rotation = 0
        #  plp.plp_fec_type = 1 : 64K
        #  plp.plp_num_blocks_max = 33
        #  plp.frame_interval = 1
        #  plp.time_il_length = 2
        #  plp.time_il_type = 0
        #  plp.in_band_a_flag = 1
        #  plp.in_band_b_flag = 0
        #  plp.plp_mode = 2
        #  plp.static_flag = 1
        #  plp.static_padding_flag = 0


        # data from reference encoder
        ref_data = [ random.getrandbits(1) for i in range(self.nframes * K) ]

        ref_src = blocks.vector_source_b( ref_data )
        ref_op = dtv_dvb_ldpc_bb_0 = dtv.dvb_ldpc_bb(
            dtv.STANDARD_DVBT2,
            dtv.FECFRAME_NORMAL,
            dtv.C1_2,
            dtv.MOD_OTHER
                                                    )
        ref_dst = blocks.vector_sink_b()
        
        self.ref_tb.connect(ref_src, ref_op)
        self.ref_tb.connect(ref_op, ref_dst)

        self.ref_tb.run()

        # generate error
        src_data = list( ref_dst.data() )
        ebcnt = round( 1e-4 * len(src_data) )
        eboffset = len(src_data) // ebcnt
        for eb in range(ebcnt):
           src_data[ eb * eboffset ] ^= True

        # padding according to forecast
        max_nin = round( (54000 * self.nchunks) * N / K )
        tail = max_nin - N*( self.nframes % self.nchunks )
        src_data += [0]*( tail )


        tags = []
        for frno in range( 0, self.nframes ):
           tag = gr.tag_t()
           tag.offset = frno * N
           tag.key = pmt.string_to_symbol( "PLPfec" )
           tag.value = pmt.make_tuple(
                          pmt.from_long( N ), 
                          pmt.init_u32vector( len(PLPTagValue), PLPTagValue )
                                     )
           tags.append( tag )


        # run
        
        src = blocks.vector_source_b( src_data, tags=tuple(tags) )

        op = DVBparts.ldpc_decoder()

        dst = blocks.vector_sink_b()

        self.tb.connect(src, op)
        self.tb.connect(op, dst)

        self.tb.run()

        result_data = list( dst.data() )
        result_tags = dst.tags()


        # check data

##        for tag in result_tags:
##           print( "result_tag at", tag.offset )

        self.assertEqual( len(tags), len(result_tags) )
        self.assertTrue( ref_data == result_data )


if __name__ == '__main__':
    gr_unittest.run( qa_ldpc_decoder )

