/* -*- c++ -*- */
/* 
 * Copyright 2021 Ingo D. Rullhusen <d01c@freeshell.de>.
 * 
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gnuradio/io_signature.h>
#include "BB2payload_impl.h"

#include <iostream>


using namespace std;


namespace gr {
  namespace DVBparts {

    BB2payload::sptr
    BB2payload::make()
    {
      return gnuradio::get_initial_sptr
        (new BB2payload_impl());
    }

    /*
     * The private constructor
     */
    BB2payload_impl::BB2payload_impl()
      : gr::block("BB2payload",
              gr::io_signature::make(1,1, sizeof(unsigned char)),
              gr::io_signature::make(1,1, sizeof(unsigned char)))
    {
       forecast_f = 1.0;

       d_msg_port = pmt::mp("msg");
       message_port_register_out(d_msg_port);

       set_tag_propagation_policy(TPP_DONT);
       set_output_multiple(4*80);
       set_fixed_rate(false);
    }

    /*
     * Our virtual destructor.
     */
    BB2payload_impl::~BB2payload_impl()
    {
    }

///    void
///    BB2payload_impl::forecast (int noutput_items, gr_vector_int &ninput_items_required)
///    { ninput_items_required[0] = ceil( forecast_f * noutput_items + 1 ); }

#define CRC_POLY 0x1d5

   bool BB2payload_impl::setup_header( const unsigned char *in )
   {
      int bit = 0;

      MATYPE1 &mt = bbheader.matype.matype_1;

      bzero(&mt,sizeof(MATYPE1));
      if( in[bit++] ) { mt.ts_gs |= 0x2; }
      if( in[bit++] ) { mt.ts_gs |= 0x1; }

      if( in[bit++] ) { mt.sis_mis |= 0x1; }

      if( in[bit++] ) { mt.ccm_acm |= 0x1; }

      if( in[bit++] ) { mt.issyi |= 0x1; }

      if( in[bit++] ) { mt.npd |= 0x1; }

      if( in[bit++] ) { mt.ext |= 0x2; }
      if( in[bit++] ) { mt.ext |= 0x1; }

      unsigned char &mt2 = bbheader.matype.matype_2;

      mt2 = 0;
      for ( unsigned int mask=0x80; mask>0; mask>>=1 ) { if(in[bit++]) mt2 |= mask; }


      bbheader.upl = 0;
      for ( unsigned int mask=0x8000; mask>0; mask>>=1 ) { if(in[bit++]) bbheader.upl |= mask; }

      bbheader.dfl = 0;
      for ( unsigned int mask=0x8000; mask>0; mask>>=1 ) { if(in[bit++]) bbheader.dfl |= mask; }

      bbheader.sync = 0;
      for ( unsigned int mask=0x80; mask>0; mask>>=1 ) { if(in[bit++]) bbheader.sync |= mask; }

      bbheader.syncd = 0;
      for ( unsigned int mask=0x8000; mask>0; mask>>=1 ) { if(in[bit++]) bbheader.syncd |= mask; }

      bbheader.crc8mode = 0;
      for ( unsigned int mask=0x80; mask>0; mask>>=1 ) { if(in[bit++]) bbheader.crc8mode |= mask; }


      bit = 0;
      unsigned int crc = 0x0;
      unsigned int b;
      for (int n = 0; n < 72; n++) 
      {
         b = (in[bit++] & 0x01) ^ ((crc >> 7) & 0x01);
         crc <<= 1;
         if (b) { crc ^= CRC_POLY; } 
      }
      bbheader.MODE = bbheader.crc8mode ^ crc;

      bzero(bbheader.SYNC,8);
      bbheader.O_UPL_nsb = bbheader.UPL = 0;
      if ( bbheader.MODE == 0 ) // NM
      {
         for ( unsigned char mask=0x80, b=0; mask>0; mask>>=1, ++b )
         { bbheader.SYNC[b] = ( bbheader.sync & mask ? 1 : 0 ); }

         bbheader.UPL = bbheader.upl;
         bbheader.O_UPL_nsb = 188*8 - 8/*sync*/; // TS only!
      }
      if ( bbheader.MODE == 1 ) // HEM
      {
         for ( unsigned char mask=0x80, b=0; mask>0; mask>>=1, ++b )
         { bbheader.SYNC[b] = ( 0x47 & mask ? 1 : 0 ); }

         bbheader.O_UPL_nsb = 188*8 - 8/*sync*/;
         bbheader.UPL = bbheader.O_UPL_nsb + ( mt.npd == 1 ? 8 : 0 );
      }

      return( bbheader.MODE==0 || bbheader.MODE==1 );
   }

   void BB2payload_impl::bbheader_out() const
   {
      clog << "BB Header:" << endl;
      clog << "MATYPE-1" << endl;

      const MATYPE1 &mt = bbheader.matype.matype_1;
      clog << " TS/GS " << hex << "0x" << (int)mt.ts_gs << dec << endl;
      clog << " SIS/MIS " << (int)mt.sis_mis << endl;
      clog << " CCM/ACM " << (int)mt.ccm_acm << endl;
      clog << " ISSYI " << (int)mt.issyi << endl;
      clog << " NPD " << (int)mt.npd << endl;
      clog << " EXT " << hex << "0x" << (int)mt.ext << dec << endl;

      clog << "MATYPE-2 " << hex << "0x" << (int)bbheader.matype.matype_2 << dec << endl;

      clog << "UPL " << bbheader.upl << endl;
      clog << "DFL " << bbheader.dfl << endl;

      clog << "SYNC " << hex << "0x" << (int)bbheader.sync << dec << endl;

      clog << "SYNCD " << bbheader.syncd << endl;
      clog << "CRC-8 MODE " << hex << "0x" << (int)bbheader.crc8mode << dec << endl;
      clog << "[MODE " << (int)bbheader.MODE << " ]" << endl;

   }

    int
    BB2payload_impl::general_work (int noutput_items,
                       gr_vector_int &ninput_items,
                       gr_vector_const_void_star &input_items,
                       gr_vector_void_star &output_items)
    {
       const pmt::pmt_t PLPfecKey( pmt::intern("PLPfec") );
       const unsigned char *in = (const unsigned char *) input_items[0];
       unsigned char *out = (unsigned char *) output_items[0];

       static bool synced = false;
       static int items_to_read = 0;
       static uint64_t framestart = 0;
       static uint64_t DFstart = 0;
       static unsigned int nUP = 0x10000;

       std::vector<tag_t> tags;
       get_tags_in_window( tags, 0, 0, noutput_items, PLPfecKey );

       if ( (!synced) && tags.size() > 0 )
       {
          tag_t &tagPLP = tags.front();

          framestart = tagPLP.offset;

          if ( setup_header(&in[framestart - nitems_read(0)]) )
          {
             synced = true;
             consume_each( (framestart - nitems_read(0)) + 80ul );
             items_to_read = bbheader.dfl;
             DFstart = framestart + 80ul;

             int fecdatasize = pmt::to_long( pmt::tuple_ref(tagPLP.value,0) );
             forecast_f = fecdatasize / (double)bbheader.dfl;

             if ( pmt::length( message_ports_in() ) )
             {
                const MATYPE1 &mt = bbheader.matype.matype_1;
                pmt::pmt_t bbnuml = pmt::make_u32vector(6,0);
                pmt::u32vector_set( bbnuml, 0, mt.ts_gs );
                pmt::u32vector_set( bbnuml, 1, mt.sis_mis );
                pmt::u32vector_set( bbnuml, 2, mt.ccm_acm );
                pmt::u32vector_set( bbnuml, 3, mt.issyi );
                pmt::u32vector_set( bbnuml, 4, mt.npd );
                pmt::u32vector_set( bbnuml, 5, mt.ext );

                pmt::pmt_t
                msg = pmt::make_tuple( pmt::mp( "BBheader" ), bbnuml );

                message_port_pub(d_msg_port, msg);
             }

///bbheader_out();
          }
          else
          {
             synced = false;
             items_to_read = 0;
             consume_each( (framestart - nitems_read(0)) + 1ul );
          }

          return(0);
       }


       if ( ! synced )
       {
          consume_each( ninput_items[0] > noutput_items
                        ? noutput_items
                        : ninput_items[0] );
          return(0);
       }


       int nin = 0;
       int nout = 0;

       int nin_max = noutput_items;
       if ( nin_max > ninput_items[0]-8 ) { nin_max = ninput_items[0]-8; }
       if ( nin_max > items_to_read   ) { nin_max = items_to_read;   }

       while ( nin<nin_max && nout<noutput_items )
       {
          if ( ( (uint64_t)nin+nitems_read(0)-DFstart == bbheader.syncd )
               || ( (bbheader.MODE == 0 ? nUP+8 : nUP ) == bbheader.UPL ) )
          {
// zero padding to block
             int bpad = bbheader.O_UPL_nsb - nUP;
             if ( bpad > 0 )
             {
                if ( noutput_items-nout-8 < bpad )
                {
                   bpad = noutput_items-nout-8;
                   bzero( &out[nout], bpad );
                   nout += bpad;
                   nUP += bpad;
                   break;
                }
                else
                {
                   bzero( &out[nout], bpad );
                   nout += bpad;
                   nUP += bpad;
                }
             }

             if ( noutput_items-9 < nout ) { break; }

             memcpy( &out[nout], bbheader.SYNC, 8 );
             nout += 8;

             nUP = 0;

             if ( bbheader.MODE == 0 ) // skip CRC-8 in NM
             {
                nin += 8;
             }

          }

          if ( nUP < bbheader.O_UPL_nsb && nout < noutput_items )
          {
             out[ nout++ ] = in[ nin ];
          }
          ++nin;
          if ( nUP < 0x10000 ) { ++nUP; }

       }

       items_to_read -= nin;


       if ( items_to_read < 1 ) { synced = false; }

       consume_each(nin);
       return(nout);
    }

  } /* namespace DVBparts */
} /* namespace gr */

